from .. import devo
from ..evo import Member, evoutils



class MorphologyControllerDevelopmentCreature(Member):
    """A creature with a genotype with three chromosomes: morphology, controller, and development.

    The genotype is a list of values between 0 and 1, with three contiguous sections for, in order,
    morphology, controller, and development.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.morphology_size  = len(self.morphology_limits(self.params))
        self.controller_size  = len(self.controller_limits(self.params))
        self.development_size = len(self.development_limits(self.params))

        self.motor_noise = self.params.get('motor_noise', 0)
        self.motor_noise_kind = self.params.get('motor_noise_kind', 'uniform')

        self.dev_vars = {}
        for name, dev_params in self.params.get('dev_vars', {}).items():
            self.dev_vars[name] = devo.DtTrajectoryGenerator(dev_params)


    ## Genotypes

    @property
    def morphology_genotype(self):
        return self.genotype[:self.morphology_size]

    @morphology_genotype.setter
    def morphology_genotype(self, v):
        """Use responsibly."""
        assert len(v) == self.morphology_size
        for i, v_i in enumerate(v):
            self.genotype[i] = v_i

    @property
    def controller_genotype(self):
        return self.genotype[self.morphology_size:self.morphology_size+self.controller_size]

    @controller_genotype.setter
    def controller_genotype(self, v):
        """Use responsibly."""
        assert len(v) == self.controller_size
        for i, v_i in enumerate(v):
            self.genotype[i + self.morphology_size] = v_i

    @property
    def development_genotype(self):
        return self.genotype[self.morphology_size+self.controller_size:]

    @development_genotype.setter
    def development_genotype(self, v):
        """Use responsibly."""
        assert len(v) == self.development_size
        for i, v_i in enumerate(v):
            self.genotype[i + self.morphology_size + self.controller_size] = v_i


    # Random genotypes

    @classmethod
    def random_genotype(cls, params, random_gen):
        """Return a random genotype"""
        return (cls.random_morphology_genotype(params, random_gen) +
                cls.random_controller_genotype(params, random_gen) +
                cls.random_development_genotype(params, random_gen)  )

    @classmethod
    def random_morphology_genotype(cls, params, random_gen):
        """Return a random morphology"""
        if params.get('morphology_genotype', None) is None:
            return [random_gen.random() for _ in range(len(cls.morphology_limits(params)))]
        else:
            return params['morphology_genotype']

    @classmethod
    def random_controller_genotype(cls, params, random_gen):
        """Return a random controller"""
        if params.get('controller_genotype', None) is None:
            return [random_gen.random() for _ in range(len(cls.controller_limits(params)))]
        else:
            return params['controller_genotype']

    @classmethod
    def random_development_genotype(cls, params, random_gen):
        """Return a random development"""
        if params.get('development_genotype', None) is None:
            return [random_gen.random() for _ in range(len(cls.development_limits(params)))]
        else:
            return params['development_genotype']


    # Mutating genotypes

    def mutate_morphology(self, n_mutation=None, sigma=None):
        """Create a new genotype by mutating the morphology of the creature"""
        if n_mutation is None:
            n_mutation = self.params['mutation']['morphology']['n_mutation']
        if sigma is None:
            sigma = self.params['mutation']['morphology']['sigma']

        if n_mutation > 0 and len(self.morphology_genotype) > 0:
            new_chromosome = list(self.morphology_genotype)
            m_index = self.random_gen.sample(range(len(new_chromosome)), n_mutation)
            for i in m_index:
                new_chromosome[i] = self._disturb(new_chromosome[i], sigma=sigma)
            return new_chromosome
        else:
            return self.morphology_genotype

    def mutate_controller(self, n_mutation=None, sigma=None):
        """Create a new genotype by mutating the controller of the creature"""
        if n_mutation is None:
            n_mutation = self.params['mutation']['controller']['n_mutation']
        if sigma is None:
            sigma = self.params['mutation']['controller']['sigma']

        if n_mutation > 0 and len(self.controller_genotype) > 0:
            new_chromosome = list(self.controller_genotype)
            m_index = self.random_gen.sample(range(len(new_chromosome)), n_mutation)
            for i in m_index:
                new_chromosome[i] = self._disturb(new_chromosome[i], sigma=sigma)
            return new_chromosome
        else:
            return self.controller_genotype

    def mutate_development(self, n_mutation=None, sigma=None):
        """Create a new genotype by mutating the development of the creature"""
        if n_mutation is None:
            n_mutation = self.params['mutation']['development']['n_mutation']
        if sigma is None:
            sigma = self.params['mutation']['development']['sigma']

        if n_mutation > 0 and len(self.development_genotype) > 0:
            new_chromosome = list(self.development_genotype)
            m_index = self.random_gen.sample(range(len(new_chromosome)), n_mutation)
            for i in m_index:
                new_chromosome[i] = self._disturb(new_chromosome[i], sigma=sigma)
            return new_chromosome
        else:
            return self.development_genotype

    def mutate_genotype(self, n_mutation=None, sigma=None):
        """Create a new genotype by mutating the one of this creature

        :param n_mutation:  number of global mutations. If `None`, taken from the config. Note that
                            morphology/controller/development number of mutation will be added to
                            this number.
        """
        # distributing global (anywhere in the genotype mutations
        if n_mutation is None:
            n_mutation = self.params['mutation']['global']['n_mutation']
        mutated_chromosomes, mutation_distrib, weights = [], {}, []
        weight_vector = (self.morphology_size, self.controller_size, self.development_size)
        for i, name in enumerate(['morphology_genotype', 'controller_genotype', 'development_genotype']):
            mutation_distrib[i] = 0
            if self.params.get(name, None) is None: # only mutating parts of the genotype that
                weights.append(weight_vector[i])    # are not fixed in the parameters.
                mutated_chromosomes.append(i)       # FIXME: too brittle to change in code/config

        for i in self.random_gen.choices(mutated_chromosomes, k=n_mutation, weights=weights):
            mutation_distrib[i] += 1

        new_morphology  = self.mutate_morphology( mutation_distrib[0] + self.params['mutation']['morphology']['n_mutation'],  sigma=sigma)
        new_controller  = self.mutate_controller( mutation_distrib[1] + self.params['mutation']['controller']['n_mutation'],  sigma=sigma)
        new_development = self.mutate_development(mutation_distrib[2] + self.params['mutation']['development']['n_mutation'], sigma=sigma)
        return new_morphology + new_controller + new_development


    ## Morphology phenotype

    @property
    def morphology(self):
        return [a[0] + (a[1]-a[0]) * g_i for a, g_i in
                zip(self.morphology_limits(self.params), self.morphology_genotype)]

    @classmethod
    def morphology_limits(cls, params):
        """Return morphology limits"""
        raise NotImplementedError


    ## Controller phenotype

    @property
    def controller(self):
        return [a[0] + (a[1]-a[0]) * g_i for a, g_i in
                zip(self.controller_limits(self.params), self.controller_genotype)]

    @classmethod
    def controller_limits(cls, params):
        """Return controller limits"""
        raise NotImplementedError


    ## Development phenotype

    @property
    def development(self):
        return [a[0] + (a[1]-a[0]) * g_i for a, g_i in
                zip(self.development_limits(self.params), self.development_genotype)]

    @classmethod
    def development_limits(cls, params):
        """Developmental genotype limits"""
        limits = []
        for name, dev_params in sorted(params.get('dev_genotype', {}).items()):
            limits += dev_params['dev_ranges']
        return limits


    ## Updates during simulations

    def bootstrap_updates(self, space, robot):
        self.motor_interval = round(0.1/space.dt)  # log the position every second #FIXME: params
        self.dev_interval   = round(0.2/space.dt)  # update development 5 times per second #FIXME: params
        self.pos_history, self.sensor_values = [], []
        self.log_interval    = round(self.params.get('log_interval', 1.0)/space.dt)  # log the position every second
        self.sensor_interval = round(self.params.get('sensor_interval', 0.1)/space.dt)  # log the sensors values at 10Hz

        # computing ticks at which the position needs to be logged for fitness computation
        self._fitness_milestones = {}   # ticks -> t
        for start, end in self.params['environment']['fitness_segments']:
            self._fitness_milestones[round(start/space.dt)] = start
            self._fitness_milestones[round(end/space.dt)]   = end
        self._fitness_ticks = set(self._fitness_milestones.keys())
        self.fitness_logs = {0.0: robot.eval_state()}   # t -> pos


    def update(self, robot, space, random_gen=None):
        """
        :param starfish:  physics' engine springs starfish instance.
        """
        self.update_log(robot, space)
        self.update_fitness(robot, space)
        self.update_motors(robot, space, random_gen=random_gen)
        self.update_development(robot, space, random_gen=random_gen)

    def update_log(self, robot, space):
        """
        :param starfish:  physics engine springs starfish instance.
        """
        if self.params.get('log_position', False) and space.ticks % self.log_interval == 0:
            self.pos_history.append((space.t, robot.eval_state()))
        if self.params.get('sensors', None) is not None and space.ticks % self.sensor_interval == 0:
            self.sensor_values.append((space.t, space.sensor_values()))


    def update_fitness(self, robot, space):
        """
        :param starfish:  physics engine springs starfish instance.
        """
        if space.ticks in self._fitness_ticks:
            self.fitness_logs[self._fitness_milestones[space.ticks]] = robot.eval_state()

    def update_motors(self, robot, space, random_gen=None):
        pass

    def update_development(self, robot, space, random_gen=None):
        pass


    ## Developmental mechanisms

    def _express_dev_genotype(self):
        """If developmental genotype is present, instanciate it a dev_vars trajectories,

        TODO: allow more complex trajectories with 3 or more milestones.
        """
        offset = 0
        for name, devovable_params in sorted(self.params.get('dev_genotype', {}).items(), reverse=True):
            assert name not in self.dev_vars  # detecting mistakes
            milestones = []
            for i, t in enumerate(devovable_params['t_milestones']):
                min_val, max_val = devovable_params['dev_ranges'][i]
                val = min_val + (max_val - min_val) * self.development_genotype[offset]
                offset += 1
                milestones.append([t, val])

            dev_params = {'function': devovable_params['function'], 'milestones': milestones}
            self.dev_vars[name] = devo.DtTrajectoryGenerator(dev_params)



class MCPosCreature(MorphologyControllerDevelopmentCreature):

    state_keys = Member.state_keys + ('pos_history',)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # position history
        self.pos_history = []

    def copy_exp(self, other_member):
        """Copy all experience from one member to another.

        Useful when a member is evaluated in another thread."""
        super().copy_exp(other_member)
        self.pos_history = other_member.pos_history

    # State

    @property
    def state(self):
        """Return the current state of the creature."""
        s = super().state
        return s + (self.pos_history,)

    @classmethod
    def from_state(cls, state, repository, random_gen):
        """Create a creature from a state (not a structured state)"""
        c = super().from_state(state, repository, random_gen)
        c.pos_history = state[6]
        return c
