import math
import statistics

from .. import devo
from ..evo import evoutils, Member
from .creature import MCPosCreature


class SinusController:

    def __init__(self, control_groups):
        self.control_groups = control_groups

    def __call__(self, starfish, space, random_gen):
        m_signal = [amp * math.sin(space.t + phase) for amp, phase in self.control_groups]
        starfish.muscle_interface.actuate(m_signal)


class UniformNoisySinusController:

    def __init__(self, control_groups, noise):
        self.control_groups = control_groups
        self.noise = noise

    def __call__(self, starfish, space, random_gen):
        m_signal = [amp * (math.sin(space.t + phase) + random_gen.uniform(-self.noise, self.noise))
                    for amp, phase in self.control_groups]
        starfish.muscle_interface.actuate(m_signal)


class GaussNoisySinusController:

    def __init__(self, control_groups, noise):
        self.control_groups = control_groups
        self.noise = noise

    def __call__(self, starfish, space, random_gen):
        m_signal = [amp * (math.sin(space.t + phase) + random_gen.gauss(0, self.noise))
                    for amp, phase in self.control_groups]
        starfish.muscle_interface.actuate(m_signal)


class StarfishCreature(MCPosCreature):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._update_control_groups()
        self.morphology_size = len(self.morphology_limits(self.params))
        self.controller_size = len(self.controller_limits(self.params))
        self._express_dev_genotype()

        self.motor_delay = self.params.get('motor_delay', 0.0)

    def _update_control_groups(self): # FIXME: merge with _create controller ?
        if self.params.get('controller', 'sinus'):
            k = len(self.params['control_limits'])
            self.control_groups = [self.controller[k*i:k*(i+1)]
                                   for i in range(len(self.controller) // k)]

    @MCPosCreature.controller_genotype.setter
    def controller_genotype(self, v):
        """Use responsibly."""
        super().controller_genotype = v
        self._update_control_groups()


    # State

    def arm_dims(self):
        n_arm, n_section = self.params['n_arm'], self.params['n_section']
        dims = []
        for i in range(n_arm):
            arm_dim = [(self.morphology[(n_section + 1) * i + j], self.params['arm_width']) for j in range(n_section)]
            arm_dim.append((self.morphology[n_section * i + n_section], 0))
            dims.append(arm_dim)
        return dims


    # Phenotypic limits

    @classmethod
    def morphology_limits(cls, params):
        """Morphology genotype limits"""
        n_arm, n_section = params['n_arm'], params['n_section']
        limits = n_arm * (n_section + 1) * params['morpho_limits']
        return limits

    @classmethod
    def controller_limits(cls, params):
        """Controller genotype limits"""
        control_name = params.get('controller', 'sinus') # TODO: sinus 3d with period?
        if params.get('controller', 'sinus') == 'sinus':
            n_arm, n_muscle_groups = params['n_arm'], params['n_muscle_groups']
            return n_arm * n_muscle_groups * params['control_limits']
        elif control_name == 'ffnn':
            n_arm, n_muscle_groups = params['n_arm'], params['n_muscle_groups']
            n_muscle_interface = n_arm * n_muscle_groups
            n_sensor_values = cls._n_sensor_values(params)
            nn_dims = [n_sensor_values] + params['nn_hidden_layer_size'] + [n_muscle_interface]
            n_weights = nn.Network.weight_lengths(nn_dims)
            return n_weights * [[-1.0, 1.0]]

    @classmethod
    def _n_sensor_values(cls, params):
        cfg = params.get('sensors', None)
        if cfg is None:
            return 0

        n_arm, n_muscle_groups = params['n_arm'], params['n_muscle_groups']
        n = 0
        # touch
        if cfg['touch'] == 'muscle_group_sides':
            n += 2 * n_arm * n_muscle_groups
        elif cfg['touch'] == 'none':
            pass
        else:
            raise NotImplementedError
        # angle
        factor = 2 if cfg['angular_velocity'] else 1
        if cfg['angle'] == 'muscle_group':
            n += factor  # center sensors
            n += factor * n_arm * n_muscle_groups  # tentacle sensors
        elif cfg['angle'] == 'none':
            if cfg['center_angle']:
                n += factor
        else:
            raise NotImplementedError
        return n

    # Updates during simulation

    def bootstrap_updates(self, space, starfish):
        super().bootstrap_updates(space, starfish)
        self._control_fun = self._create_controller(starfish, space)

    def _create_controller(self, starfish, space):
        control_name = self.params.get('controller', 'sinus')
        if control_name == 'sinus':
            assert len(self.control_groups) == len(starfish.muscle_interface)
            if self.motor_noise == 0:
                return SinusController(self.control_groups)
            elif self.motor_noise_kind == 'uniform':
                return UniformNoisySinusController(self.control_groups, self.motor_noise)
            elif self.motor_noise_kind == 'gauss':
                return GaussNoisySinusController(self.control_groups, self.motor_noise)
            else:
                raise ValueError('controller kind {} not known'.format(self.motor_noise_kind))
        elif control_name == 'ffnn':
            nn_dims = [len(space.sensor_values())] + self.params['nn_hidden_layer_size'] + [len(starfish.muscle_interface)]
            network = nn.Network(nn_dims, weights=self.controller)
            amp = self.params['control_limits'][0][1]
            return FFNNController(nn_dims, amp, weights=self.controller)
        else:
            raise ValueError

    def update_motors(self, starfish, space, random_gen=None):
        """
        :param starfish:    physics' engine springs starfish instance.
        :param random_gen:  random generator, usually provided by the environment to allow
                            repeatable execution of the evaluation independently of the random
                            generator creating new creatures.  #FIXME: remove
        """
        if random_gen is None:
            random_gen = self.random_gen
        if space.t >= self.motor_delay and space.ticks % self.motor_interval == 0:
            self._control_fun(starfish, space, random_gen=random_gen)

    def update_development(self, starfish, space, random_gen=None):
        """
        :param starfish:    physics' engine springs starfish instance.
        :param random_gen:  random generator, usually provided by the environment to allow
                            repeatable execution of the evaluation independently of the random
                            generator creating new creatures.
        """
        if space.ticks % self.dev_interval == 0:
            for name, traj_gen in self.dev_vars.items():
                update_needed, value = traj_gen.generate(space)
                if update_needed:
                    if name == 'birth_size':
                        starfish.change_height_dev_factor(value)
                    elif name == 'node_mass':
                        starfish.change_mass(value)
                    else:
                        raise NotImplementedError('The developmental `{}` is not known.'.format(name))


    # Developmental values

    def dev_value(self, name, t=0):
        """Return the current value of a parameter, be it fixed or developping"""
        if name in self.dev_vars:
            return self.dev_vars[name](t)
        # elif name == 'materials':

        else:
            return self.params[name]

    def body_size(self, t):
        """Body size at time t"""
        arm_dims = self.arm_dims()
        current_size = self.dev_value('birth_size', t=t)
        avg_tentacle_length = current_size * statistics.mean([sum(list(zip(*arm_dim))[0])
                                                              for arm_dim in arm_dims])
        center_size = self.dev_value('center_size', t=t)
        center_apothem = center_size / (2 * math.tan(math.pi / len(arm_dims)))

        return 2 * (center_apothem + avg_tentacle_length)

    @classmethod
    def develop_params(cls, params, name, new_val):
        if name == 'muscle_freq':
            params['materials']['link_sec_side']['frequency'] = new_val
        elif name == 'muscle_stiff':
            params['materials']['link_sec_side']['stiffness'] = new_val
        elif name == 'node_mass':
            params['materials']['node_center']['mass']        = new_val
            params['materials']['node_section']['mass']       = new_val
            params['materials']['node_tip']['mass']           = new_val
        elif name == 'birth_size':
            params['birth_size']                              = new_val
        elif name in ['gravity', 'gravity_angle', 'ceiling']:
            params['environment'][name]                       = new_val
        elif name in ['step_sizes', 'ground_heights']:
            params['environment']['ground']['heights']        = new_val
        elif name == 'ground_widths':
            params['environment']['ground']['widths']         = new_val
        else:
            raise NotImplementedError
        return params
