# TODO: simple poly
# TODO: constant
import math

from .. import utils

# Linear case
liner_template = """
muscle_strength:
    function: "linear"
    milestone: [(   0,  50),   # before 0, y=50; after y = (125 - 50)/1000 * t + 50
                (1000, 125),   # after 1000, y = (500 - 125)/500 * (t - 1000) + 125
                (1500, 500)]   # after 2000, y=500
"""

class TrajectoryGenerator:
    """Trajectory generator for developmental variables.

    Developmental variables, like morphological size, muscle strength, mass,
    will change at regular intervals. This class handle computing the trajectory
    of those variables.
    """

    def __init__(self, params):
        self.params = params
        self._construct()


    def _construct(self):
        funmap = {'step'              : self._init_step,
                  'linear'            : self._init_linear,
                  'linear_indexed'    : self._init_linear_indexed,
                  'quadratic_indexed' : self._init_quadratic_indexed,
                  'cubic_indexed'     : self._init_cubic_indexed,
                  'logscale'          : self._init_logscale,
                  'decay_factor'      : self._init_decayfactor,
                  'multi'             : self._init_multi,
                  'sum'               : self._init_sum,
                  'mult'              : self._init_mult,
                  'bounded'           : self._init_bounded,
                  'sinus'             : self._init_sinus}
        try:
            init_fun = funmap[self.params['function']]
        except KeyError:
            raise NotImplementedError('<{}> generator is unknown'.format(self.params['function']))
        init_fun()

    def __getstate__(self):
        return self.params

    def __setstate__(self, params):
        self.params = params
        self._construct()

    def __call__(self, t):
        return self.traj_fun(t)

    # Step generator

    def _init_step(self):
        milestones = self.params['milestones']

        # verifying that time milestones are chronological
        for t in range(len(milestones) - 1):
            assert milestones[t][0] < milestones[t+1][0]

        if len(milestones) == 1:  # constant case
            def _generator(t): return milestones[0][1]

        else:
            def _generator(t):
                # excess temporal bounds: constant
                if t <= milestones[0][0]:
                    return milestones[0][1]
                elif milestones[-1][0] < t:
                    return milestones[-1][1]
                # otherwise
                else:
                    for m_index, ml in enumerate(milestones[1:]):
                        if t == ml[0]:
                            return ml[1]
                        elif t < ml[0]:
                            prev_ml = milestones[m_index]  # previous milestone
                            return prev_ml[1]

        self.traj_fun = _generator



    # Linear generator

    def _init_linear(self):
        milestones = self.params['milestones']

        # verifying that time milestones are chronological
        for t in range(len(milestones) - 1):
            assert milestones[t][0] < milestones[t+1][0]

        if len(milestones) == 1:  # constant case
            def _generator(t): return milestones[0][1]

        else:
            def _generator(t):
                # excess temporal bounds: constant
                if t <= milestones[0][0]:
                    return milestones[0][1]
                elif milestones[-1][0] < t:
                    return milestones[-1][1]
                # otherwise
                else:
                    for m_index, ml in enumerate(milestones[1:]):
                        if t <= ml[0]:
                            prev_ml = milestones[m_index]  # previous milestone
                            return prev_ml[1] + (ml[1] - prev_ml[1])/(ml[0] - prev_ml[0]) * (t - prev_ml[0])

        self.traj_fun = _generator


    # Linear generator, indexed on a subgenerator

    def _init_linear_indexed(self):
        ref_point, ref_index = self.params['ref_point']
        index = self.params.get('index', None)
        index_generator = TrajectoryGenerator(index)

        point_1 = ref_point / (ref_index)
        def _generator(t):
            index_t = index_generator(t)
            return point_1 * index_t

        self.traj_fun = _generator


    # Quadratic generator, indexed on a subgenerator

    def _init_quadratic_indexed(self):
        ref_point, ref_index = self.params['ref_point']
        index = self.params.get('index', None)
        index_generator = TrajectoryGenerator(index)

        point_1 = ref_point / (ref_index * ref_index)
        def _generator(t):
            index_t = index_generator(t)
            return point_1 * index_t * index_t

        self.traj_fun = _generator


    # Cubic generator, indexed on a subgenerator

    def _init_cubic_indexed(self):
        ref_point, ref_index = self.params['ref_point']
        index = self.params.get('index', None)
        index_generator = TrajectoryGenerator(index)

        point_1 = ref_point / (ref_index * ref_index * ref_index)
        def _generator(t):
            index_t = index_generator(t)
            return point_1 * index_t * index_t * index_t

        self.traj_fun = _generator


    # Logscale generator

    def _init_logscale(self):
        milestones = self.params['milestones']

        # verifying that time milestones are chronological
        for t in range(len(milestones) - 1):
            assert milestones[t][0] < milestones[t+1][0]

        if len(milestones) == 1:  # constant case
            def _generator(t): return milestones[0][1]

        else:
            def _generator(t):
                # excess temporal bounds: constant
                if t <= milestones[0][0]:
                    return milestones[0][1]
                elif milestones[-1][0] < t:
                    return milestones[-1][1]
                # otherwise
                else:
                    for m_index, ml in enumerate(milestones[1:]):
                        if t <= ml[0]:
                            t_start, v_start = milestones[m_index]  # previous milestone
                            t_end, v_end = ml
                            progress = (t - t_start) / (t_end - t_start)
                            return utils.log_transform(progress, v_start, v_end)

        self.traj_fun = _generator


    # Decay factor generator

    def _init_decayfactor(self):

        start_val, min_val = self.params['start'], self.params['min']
        decay_factor = self.params['decay_factor']
        assert 0.0 <= decay_factor <= 1.0


        def _generator(t):
            return max(min_val, start_val * decay_factor ** t)

        self.traj_fun = _generator


    # Multi-dimensional generator

    def _init_multi(self):
        self.traj_fun = self._gen_multi

        self._sub_generators = []
        for subparams in self.params['trajectories']:
            self._sub_generators.append(TrajectoryGenerator(subparams))

    def _gen_multi(self, t):
        return [subgen(t) for subgen in self._sub_generators]


    # Sum of generators

    def _init_sum(self):
        self.traj_fun = self._gen_sum

        self._sub_generators = []
        for subparams in self.params['subgenerators']:
            self._sub_generators.append(TrajectoryGenerator(subparams))

    def _gen_sum(self, t):
        return sum(subgen(t) for subgen in self._sub_generators)


    # Multiplication of generators

    def _init_mult(self):
        self.traj_fun = self._gen_mult

        self._sub_generators = []
        for subparams in self.params['subgenerators']:
            self._sub_generators.append(TrajectoryGenerator(subparams))

    def _gen_mult(self, t):
        result = 1.0
        for subgen in self._sub_generators:
            result *= subgen(t)
        return result


    # Boundary generator (limits another generator between bounds)

    def _init_bounded(self):
        v_min, v_max = self.params['min'], self.params['max']
        sub_generator = TrajectoryGenerator(self.params['subgenerator'])
        def _generator(t):
            return min(v_max, max(v_min, sub_generator(t)))

        self.traj_fun = _generator

    # Sinus generator

    def _init_sinus(self):
        a, p, ph = self.params['amplitude'], self.params['period'], self.params['phase']
        def _generator(t):
            return a * math.sin(2*math.pi*(t/p + ph))

        self.traj_fun = _generator


class DtTrajectoryGenerator(TrajectoryGenerator):
    """Trajectory generator for developmental variables during simulation

    Take into consideration period of constant values, relative to the dt parameter of the
    simulation, and also returns if the update is needed or not.
    """

    def __init__(self, params):
        super().__init__(params)
        self.tick_period = None

    def __setstate__(self, params):
        super().__setstate__(params)
        self.tick_period = None

    def generate(self, space, force=False):
        if self.tick_period is None:
            if self.params.get('period', None) is not None:
                self.tick_period = round(max(1, self.params['period'] / space.dt))
            else:
                self.tick_period = 1

        update_needed = force or (space.ticks % self.tick_period == 0)
        if update_needed:
            return True, self.traj_fun(space.t)
        else:
            return False, None
