import copy
import math
import random
import statistics

import numpy as np

import springs
from springs.creatures import Starfish, DevStarfish, SectionOneMuscle, FusedSquares, Centipede

import evodevo


class MetaEnvironment:

    def __init__(self, params, random_gen):
        self.params, self.random_gen = params, random_gen

    def run(self, member):
        subrun = evodevo.LightRun(member.params['run'], self.random_gen)
        subrun.run_generations()
        member.fitness = self._compute_fitness(subrun)
        return member

    def _compute_fitness(self, run):
        # compute fitness
        run.population.repository.archive_all()
        fitness = float('-inf')
        for uids in run.population.generations[-self.params['max_fitness_window']:]:
            for uid in uids:
                member_state = run.population.repository.archives[uid]
                fitness = max(fitness, member_state[evodevo.Member.fitness_index])
        return fitness


class MetaMorphologyEnvironment(MetaEnvironment):

    def run(self, member):
        meta_member_params = copy.deepcopy(member.params)
        meta_member_params.pop('run')
        meta_member_params.update(member.params['run']['member'])
        member.params['run']['member'] = meta_member_params
        member.params['run']['member']['morphology_genotype'] = member.morphology_genotype
        return super().run(member)


class StarfishEnvironment:

    springs_class = DevStarfish

    def __init__(self, params, random_gen):
        """
        :param params:  dictionary of parameters. See the section 'environment' of the  file
                        `conf_template.yaml` for details.
        """
        self.params, self.random_gen = params, random_gen
        self.dt, self.T = self.params['dt'], self.params['T']

    def _make_ground(self, space):
        gp = self.params.get('ground', {'type': 'flat', 'height': 0, 'restitution': 0.2})  # ground params
        restitution = gp['restitution']
        if gp['type'] == 'flat':
            space.add_rect(-1000000, 1000000,  gp['height']-1000, gp['height'], restitution=restitution)
        elif gp['type'] in ['steps', 'saw']:
            assert len(gp['heights']) == len(gp['widths']) > 0
            y_offset = 0
            sum_sizes, sum_widths = sum(gp['heights']), sum(gp['widths'])
            max_size = max(gp['heights'])
            k_steps = int(20000/sum_widths)
            for i in range(-int(round(k_steps/10)), k_steps):
                x_offset = sum(gp['widths']) * i
                for size, width in zip(gp['heights'], gp['widths']):
                    y_offset = sum_sizes * i
                    if gp['type'] == 'steps':
                        #print(y_offset + size)
                        space.add_rect(x_offset, x_offset + width*1.01, -1000, y_offset + size,
                                       restitution=restitution)
                    elif gp['type'] == 'saw':
                        extra = math.tan(math.atan2(size, width/2))*0.1
                        space.add_triangle(x_offset - width*0.1, -max_size-extra, x_offset + width*1.1, -max_size-extra, x_offset + width/2, size-max_size,
                                           restitution=restitution)
                    x_offset += width
                    y_offset += size
            space.add_rect(-1000000, 1000000, -max_size-1000, -max_size, restitution=restitution)
        else:
            raise NotImplementedError

    def _make_ceiling(self, space):
        ceiling = self.params.get('ceiling', None)
        if ceiling is not None:
            space.add_rect(-2000, 500, ceiling + 800, ceiling + 900, restitution=self.params['restitution'])
            space.add_rect(500, 40000, ceiling, ceiling + 900, restitution=self.params['restitution'])

    def _create_space(self, creature):
        gravity_x = self.params['gravity'] * math.sin(math.radians(self.params.get('gravity_angle', 0)))
        gravity_y = self.params['gravity'] * math.cos(math.radians(self.params.get('gravity_angle', 0)))
        space = springs.create_space(dt=self.dt, gravity=(gravity_x, gravity_y), n_substep=self.params['n_substep'],
                                     engine=self.params.get('engine', 'cpp'))

        materials = creature.dev_value('materials', t=0)
        for m in materials.values():
            if 'adjusted_stiffness' in m:
                m['stiffness'] = m.pop('adjusted_stiffness')
        # print(materials['link_sec_side'])
        starfish = self.springs_class(space, creature.arm_dims(), (0, 300),
                                      creature.dev_value('center_size', t=0),
                                      # section_cls=springs.CentralBoneSection,
                                      materials=materials,
                                      height_dev_factor=creature.dev_value('birth_size', t=0),
                                      sensor_cfg=creature.params.get('sensors', None))
        starfish.create_muscle_interface(n_group=creature.params['n_muscle_groups'],
                                         muscle_cls=SectionOneMuscle)

        min_y = min(node.y for node in starfish.nodes)
        starfish.translate(0, -min_y)
        space.add_entity(starfish)

        return space, starfish

    def check_stability(self, space):
        events = {}
        for i, link in enumerate(space.links):
            if link.max_length > 10 * link.relax_length:
                name = 'link{}'.format(i)
                events[name] = {'desc': 'unstable_link',
                                'max_length': link.max_length,
                                'relax_length': link.relax_length,
                                'stiffness': link.stiffness,
                                'damping_ratio': link.damping_ratio,
                                'class': link.__class__.__name__,
                               }
        return events

    #@profile
    def run(self, creature, headless=None):
        # creating the physic engine.
        space, starfish = self._create_space(creature)

        # ground & ceiling
        self._make_ground(space)
        self._make_ceiling(space)

        # register control, log, development callbacks
        creature.bootstrap_updates(space, starfish)
        starfish.add_controller(creature.update, self.random_gen)

        # run the simulation
        n_steps = int(self.T/self.dt) + 1
        headless = self.params.get('headless', True) if headless is None else headless
        if headless:
            for tick in range(n_steps):
                space.step()
        else:
            springs.create_display(space, steps=n_steps, spf=self.params['spf'],
                                   zoom=self.params['zoom'],
                                   size_factor=self.params.get('size_factor', 1.0),
                                   width =self.params['window_size'][0],
                                   height=self.params['window_size'][1],
                                   screenshot_folder=self.params.get('screenshot_folder', None),
                                   paused=self.params.get('start_paused', False),
                                   pause_at_end=self.params.get('pause_at_end', True),
                                   start_skip=self.params.get('start_skip', 0.0),
                                   message=self.params.get('message', None)).run()

        # compute fitness
        fitness_function = self.params.get('fitness_function', 'body_lengths')
        creature.fitness = 0
        for start, end in self.params['fitness_segments']:
            pos_start = creature.fitness_logs[start]
            pos_end   = creature.fitness_logs[end]
            if fitness_function == 'body_lengths':
                avg_size = (creature.body_size(start) + creature.body_size(end)) / 2.0
                creature.fitness += (pos_end[0] - pos_start[0]) / avg_size
            elif fitness_function == 'distance':
                creature.fitness += (pos_end[0] - pos_start[0])
            else:
                raise ValueError('Unknown fitness function')

        assert not np.isnan(creature.fitness)

        creature.sensor_values = []
        if len(space.sensors) > 0:
            creature.sensor_values = [sensor.value for sensor in space.sensors]
        # print(creature.sensor_values)

        return creature


class CentipedeEnvironment(StarfishEnvironment):

    springs_class = Centipede


class SquaresEnvironment:

    def __init__(self, params, random_gen):
        """
        :param params:  dictionary of parameters. See the section 'environment' of the  file
                        `conf_square_template.yaml` for details.
        """
        self.params, self.random_gen = params, random_gen
        self.dt, self.T = self.params['dt'], self.params['T']

    def run(self, creature, headless=None):
        # creating the physic engine.
        creature.pos_history = []

        space = Space(dt=self.dt, gravity=self.params['gravity'], n_substep=self.params['n_substep'])
        squares = FusedSquares(space, creature.square_size(), creature.square_stiffness(),
                               amp_limit=creature.params['amp_limit'],
                               amp_factor=creature.params['amp_factor'],
                               α_freq=creature.params['α_freq'],
                               β_freq=creature.params['β_freq'],
                               γ_stif=creature.params['γ_stif'],
                               origin=(100, 120))

        log_interval = round(1.0/space.dt)  # log the position every second
        def update(space):
            if space.ticks % log_interval == 0:
                creature.pos_history.append((space.t, squares.avg_center))
            m_signal = [amp*math.sin(space.t + math.pi * phase) for amp, phase in zip(creature.amplitudes(), creature.phases())]
            squares.actuate(m_signal)

        space.add_update_function(update)
        space.add_ground(y=100, restitution=self.params['restitution'])

        if headless is None:
            headless = self.params['meta']['headless']
        if headless:
            for t in range(int(self.T/self.dt) + 1):
                space.step()
        else:
            import springs.display
            springs.display.Springs(space, duration=self.T, spf=self.params['spf'],
                                    zoom=self.params['zoom'], width =self.params['window_size'][0],
                                    height=self.params['window_size'][1]).run()

        creature.fitness = creature.pos_history[-1][1][0] - creature.pos_history[0][1][0]
        return creature
