# Analysis
import sys
import numpy as np

from .member import Member


def genealogy(population, member):
    """Return the list of creatures forming the given creature genealogy"""
    gen_list = [member.uid]
    puid = member.puid
    while puid is not None:
        gen_list.append(puid)
        parent = population.repository.member(puid)
        try:
            puid = parent.puid
        except AttributeError:
            puid = parent['puid'][0]
            if puid == -1:
                puid = None
    return gen_list

def population_genealogy(population, window=None):
    """Create a population genealogy

    :return:  a list of member states and a list of the list index of each parent.
    """
    member_states = []
    parents       = []
    parent_map    = {}

    gens = population.generations
    if window is not None:
        gens = gens[window[0]:window[1]]

    for t_gen, generation in enumerate(gens):
        for uid in generation:
            m_state = population.repository.member_state(uid)
            member_states.append(m_state)
            parent_map[m_state[Member.uid_index]] = len(member_states) - 1
            puid = m_state[Member.puid_index]
            if puid is not None and t_gen > 0:
                parents.append(parent_map[puid])
            else:
                parents.append(None)

    assert min([p for p in parents if p is not None]) >= 0
    assert len(member_states) == len(parents)
    return member_states, parents


def genealogy_distance(population, uids=None, window=None):
    """Compute the genealogy distance of a population.

    The genealogy distance is the sum of the distance between each pair of sequential ancestor
    from the root to a given member.

    :param uids:    if not `None`, consider only the genealogy with those uids as leaves.
    :param window:  if `window` is not None, all the edges outside the time window have distance
                    zero. To compute the distances only over a portion of the search process.
    """
    if uids is None:
        member_states, parents = population_genealogy(population)
    else:
        member_states, parents = population_genealogy_subset(population, uids)
    members = [population.repository.member(ms[Member.uid_index]) for ms in member_states]
    distances = []
    for m, p_i in zip(members, parents):
        if p_i is None:
            distances.append(0.0)
        else:
            m_p = members[p_i]
            # if the member is outside the window, it's distance to its parent is zero.
            if window is not None and (m.generation < window[0] or window[1] < m.generation):
                distances.append(distances[p_i])
            else:
                distances.append(genotype_distance(m, m_p) + distances[p_i])
    return members, distances


def population_genealogy_subset(population, uids, window=None):
    sys.setrecursionlimit(10000)
    """Create a population genealogy whoses leaves are restricted to the given uids"""
    member_states, parents = population_genealogy(population, window=window)
    selected = len(member_states) * [False]

    def _mark_selected(index):
        if selected[index]:
            return
        else:
            selected[index] = True
            if parents[index] is not None:
                _mark_selected(parents[index])

    uids = set(uids)
    for i, m_state in enumerate(member_states):
        uid = m_state[Member.uid_index]
        if uid in uids:
            _mark_selected(i)

    subset_member_states, subset_parents, offsets = [], [], []
    for m_state, parent, s_i in zip(member_states, parents, selected):
        if s_i:
            subset_member_states.append(m_state)
            subset_parents.append(None if parent is None else parent + offsets[parent])
            if len(offsets) == 0:
                offsets.append(0)
            else:
                offsets.append(offsets[-1])
        else:
            if len(offsets) == 0:
                offsets.append(-1)
            else:
                offsets.append(offsets[-1]-1)

    assert min([s for s in subset_parents if s is not None]) >= 0
    return subset_member_states, subset_parents


def select_elite_history(population, period=100, width=10, milestones=None):
    if milestones is None:
        n_gen = len(population.generations)
        milestones = list(range(0, n_gen, period))
        if (n_gen - 1) % period != 0:
            milestones.append(n_gen - 1)

    elites = []
    for t in milestones:
        members = []
        for t_i in range(max(t - width + 1, 0), t + 1):
            gen = population.generations[t_i]
            for uid in gen:
                member = population.repository.member(uid)
                if member.generation == t_i:
                    members.append(member)
        if len(members) > 0:
            member = sorted(members, key=lambda m: m.fitness)[-1]
            elites.append(member)

    return elites


def sorted_creatures(population, uids):
    repo = population.repository
    all_perfs = [(repo.member_state(uid)[Member.fitness_index], uid) for uid in uids]
    bests = sorted(all_perfs, reverse=True)
    return [repo.member(uid) for _, uid in bests]


def select_best(population, n=1, window=None):
    repo = population.repository

    if window is None:
        generations = population.generations
    else:
        if window[1] is None:
            generations = population.generations[window[0]:]
        generations = population.generations[window[0]:window[1]]

    uids = []
    for generation in generations:
        uids.extend(generation)

    all_perfs = [(repo.member_state(uid)[Member.fitness_index], uid) for uid in uids]
    bests = sorted(all_perfs)[-n:]
    return [repo.member(uid) for _, uid in bests]


def noise_fitness(run, member, n, noise=None):
    if noise is not None:
        original_noise = member.motor_noise
        member.motor_noise = noise
    fitnesses = []
    for i in range(n):
        run.population.evaluate_member(member, eval_seed=i)
        fitnesses.append(member.fitness)
    if noise is not None:
        member.motor_noise = original_noise
    return fitnesses


def performances(population, fitness_function=None):
    perfs = []
    for gen in population.generations:
        perfs.append([])
        for uid in gen:
            m_fitness = fitness(population, uid, fitness_function)
            perfs[-1].append(m_fitness)
    return perfs


def performance_age(population, fitness_function=None):
    """Return the performance of each member, as well as the age of the most
        ancient ancestor and if it's new this generation"""
    perfs, age_map = [], {}
    for t, gen in enumerate(population.generations):
        new_age_map = {}
        perfs.append([])
        for uid in gen:
            m_fitness = fitness(population, uid, fitness_function)

            if uid in age_map:
                perfs[-1].append((m_fitness, age_map[uid], False))
                new_age_map[uid] = age_map[uid]
            else:
                state = population.repository.member_state(uid)
                m_puid = state[3]
                if m_puid is None:
                    perfs[-1].append((m_fitness, t, True))
                    new_age_map[uid] = t
                else:
                    try:
                        perfs[-1].append((m_fitness, age_map[m_puid], True))
                        new_age_map[uid] = age_map[m_puid]
                    except KeyError:
                        perfs[-1].append((m_fitness, new_age_map[m_puid], True))
                        new_age_map[uid] = new_age_map[m_puid]
                        # FIXME, FRAGILE: relies on copies being first in generation list.

        age_map = new_age_map
    return perfs


def fitness(population, uid, fitness_function=None):
    """Return the fitness of one member of the population, specified by uid"""
    m_fitness = population.repository.member_state(uid)[5]
    if fitness_function is None or fitness_function == population.params['member']['environment']['fitness_function']:
        return m_fitness

    else:
        member = population.repository.member(uid)
        member_ff = member.params['environment']['fitness_function']
        if fitness_function == 'distance' and member_ff == 'body_lengths':
            assert (len(member.params['environment']['fitness_segments']) == 1
                    and member.params['environment']['fitness_segments'][0][0] == 0)
            end = member.params['environment']['T']
            return m_fitness * (member.body_size(0) + member.body_size(end)) / 2

        elif fitness_function == 'body_lengths' and member_ff == 'distance':
            assert (len(member.params['environment']['fitness_segments']) == 1
                    and member.params['environment']['fitness_segments'][0][0] == 0)
            end = member.params['environment']['T']
            return m_fitness / ((member.body_size(0) + member.body_size(end)) / 2)

        else:
            raise NotImplementedError


def genotype_distance(m1, m2, average=False):
    """Compute the genotype distance between two members.

    The genotype distance is the mean (i.e. divided by the number mutable of dimensions) distance
    of a the genotype of a member to the genotype its oldest ancestor (the root).
    """
    if m1.params['mutation']['global']['n_mutation'] > 0:
        m1_genotype, m2_genotype = m1.genotype, m2.genotype
    else:
        m1_genotype, m2_genotype = [], []
        if m1.params['mutation']['morphology']['n_mutation'] > 0:
            m1_genotype += m1.morphology_genotype
            m2_genotype += m2.morphology_genotype
        if m1.params['mutation']['controller']['n_mutation'] > 0:
            m1_genotype += m1.controller_genotype
            m2_genotype += m2.controller_genotype
        if m1.params['mutation']['morphology']['n_mutation'] > 0:
            m1_genotype += m1.development_genotype
            m2_genotype += m2.development_genotype
    m1_genotype, m2_genotype = np.array(m1_genotype), np.array(m2_genotype)
    if average:
        return np.linalg.norm(m1_genotype - m2_genotype) / len(m1_genotype)
    else:
        return np.linalg.norm(m1_genotype - m2_genotype) #/len(m1_genotype)

def genotype_distance_pop(population, window=None):
    """Compute the genotype distance of all members of the population

    :param root_gen:  root generation. The distances are calculated from this generation and only
                      for the descendants of that generation.
    """
    root_gen, end_gen = 0, len(population.generations)
    if window is not None:
        root_gen, end_gen = window

    roots, families = {}, {}
    uids, dists = [], []
    for t, gen in enumerate(population.generations[root_gen:end_gen]):
        dists.append([])
        for uid in gen:
            state = population.repository.member_state(uid)
            m_puid = state[Member.puid_index]
            uids.append(state[Member.uid_index])
            m_generation = state[Member.generation_index]
            if m_puid is None or m_generation <= root_gen:
                families[uid] = len(families)
                roots[uid] = population.repository.member(uid)
                dists[-1].append((families[uid], 0.0))
            else:
                root = roots[m_puid]
                roots[uid] = root
                d = genotype_distance(root, population.repository.member(uid))
                dists[-1].append((families[root.uid], d))

    if all(len(d_i) == len(dists[0]) for d_i in dists):
        dists = np.array(dists)
    else:
        dists = np.array(dists, dtype=object)
    return population.generations[root_gen:end_gen], dists


def random_walk(run, member, n, generation=None, sigma=None):
    walk = [member]
    pop = run.population
    if generation is None:
        generation = len(pop.generations) - 1
    else:
        raise NotImplementedError
    for i in range(n-1):
        m = walk[-1].mutate(params=member.params, sigma=sigma)
        m.uid = member.uid + 1 + i
        walk.append(m)

    return walk

def noisy_member(run, member, n, generation=None, sigma=None):
    members = [member]
    pop = run.population
    if generation is None:
        generation = len(pop.generations) - 1
    else:
        raise NotImplementedError
    for i in range(n-1):
        m = member.mutate(params=member.params, sigma=sigma)
        m.uid = member.uid + 1 + i
        members.append(m)

    return members

def gradient(members):
    assert all(m.fitness is not None for m in members)

    def _gradient(m1, m2):
        assert m1.genotype != m2.genotype
        return (m2.fitness - m1.fitness)/genotype_distance(m1, m2)
    gs = [abs(_gradient(members[i], members[i+1])) for i in range(len(members)-1)]
    return np.mean(gs)
