import numpy as np
from .. import utils


def pareto_set(fitnesses):
    """Return the index of the pareto set (non-dominated points), where higher is better.

    :param fitnesses:  a np.array of size n_points x n_fitnesses
    """
    p_set = np.ones(fitnesses.shape[0], dtype=bool)
    for i, fitness in enumerate(fitnesses):
        if p_set[i]:
            p_set[p_set] = np.any(fitnesses[p_set] > fitness, axis=1)
            p_set[i] = True
    return p_set


def pareto_sets(fitnesses):
    """Return the ordering of all fitnesses, grouped by pareto sets"""
    results = []
    fitnesses = np.array(fitnesses)
    indexes = np.array(range(fitnesses.shape[0]))
    while(len(indexes) > 0):
        pareto_indexes = pareto_set(fitnesses)
        results.append(list(indexes[pareto_indexes]))
        fitnesses = fitnesses[~pareto_indexes]
        indexes = indexes[~pareto_indexes]
    return results


def _mirror(v, v_min, v_max):
    """Recursively constraint a value inside a range, by mirroring excesses against boundaries."""
    if v > v_max:
        v -= 2 * (v - v_max)
        return _mirror(v, v_min, v_max)
    elif v < v_min:
        v += 2*(v_min - v)
        return _mirror(v, v_min, v_max)
    return v

def gaussian_mutation(v, limits, sigma, random_gen):
    """Gaussian mutations, with mirror effect when going over the limit.

    If the upper limit is 1 and the new value is 1.07, then the returned mirror value is 0.93.
    """
    v_min, v_max = limits
    new_v = v + (v_max - v_min) * random_gen.gauss(0, sigma)
    return _mirror(new_v, v_min, v_max)


def parent_selection(members, n_parent, random_gen, pareto_age=True):
    """Select the members to become parents of the next generation

    :param members:     the members to select from.
    :param n_parent:   the number of parents to select.
    :param pareto_age:  if true, consider the age of the members as well as the fitness and compute
                        the pareto fronts to select the surviving members
    :return:            a pair of list of the surviving and dead members, respectively.
    """

    assert not any(np.isnan([member.fitness for member in members]))

    if not pareto_age:
        members = sorted(members, key=lambda x: x.fitness, reverse=True)
        return members[:n_parent], members[n_parent:]

    else:
        fitnesses = [[member.fitness, member.generation] for member in members]
        order  = pareto_sets(fitnesses)

        index_survived, index_dead = [], []
        while len(index_survived) < n_parent:
            first_order = order.pop(0)
            overflow = len(index_survived) + len(first_order) - n_parent
            if overflow >= 0:
                random_gen.shuffle(first_order)
                index_survived.extend(first_order[:len(first_order) - overflow])
                index_dead = first_order[len(first_order) - overflow:]
            else:
                index_survived.extend(first_order)
        index_dead.extend(utils.flatten_once(order))

        return [members[index] for index in index_survived], [members[index] for index in index_dead]
