import copy
from . import evoutils


class Member:

    state_keys = ('genotype', 'generation', 'uid', 'puid', 'param_uid', 'fitness')
    genotype_index, generation_index, uid_index, puid_index, param_uid_index, fitness_index = 0, 1, 2, 3, 4, 5

    def __init__(self, genotype, params, generation=None, uid=None, puid=None, random_gen=None):
        self.genotype = genotype
        self.params = params  # non-evolvable parameters, but possibly developable
        self.random_gen = random_gen
        self.eval_seed_offset = 0

        # context params
        self.generation = generation
        self.uid = uid
        self.puid = puid

        # fitness
        self.fitness = None

        # any event happening during simulation worth reporting
        self.events = {}


    @classmethod
    def genotype_length(cls, params):
        raise NotImplementedError

    @classmethod
    def random(cls, params, generation, uid, puid=None, random_gen=None):
        genotype = cls.random_genotype(params, random_gen)
        return cls(genotype, params, generation=generation, uid=uid, puid=puid, random_gen=random_gen)

    @classmethod
    def random_genotype(cls, params, random_gen):
        """Return a random---normalized, i.e. list of value in [0, 1]---genotype"""
        return [random_gen.random() for _ in range(cls.genotype_length(params))]

    @property
    def key_state(self):
        """Return the current state of the member as a dictionary."""
        return {'genotype'    : self.genotype,
                'generation'  : self.generation,
                'uid'         : self.uid,
                'puid'        : self.puid,
                'params'      : self.params,  # be careful about not changing parameters
                'fitness'     : self.fitness}

    @property
    def state(self):
        """Return the current state of the member as a tuple.

        The description of each value can be accessed through the Member.state_keys class variable.
        Useful to archive results. MemberRepository.member can reinstanciate a state into a Member.
        """
        return (self.genotype, self.generation, self.uid, self.puid,
                self.params['_param_uid'], self.fitness)

    @classmethod
    def from_state(cls, state, repository, random_gen):
        """Create a creature from a tuple state"""
        genotype, generation, uid, puid, param_uid, fitness = state[:6]
        params = repository.member_parameters(param_uid)
        c = cls(genotype, params, generation=generation, uid=uid, puid=puid, random_gen=random_gen)
        c.fitness = fitness
        return c

    def copy(self, generation=None, uid=None, params=None):
        """Create a new creature by copying this one.

        :params:  the new members parameters are usually inherited verbatim from the parent
                  member, but we may change them by providing this argument with a non-None value.
        """
        params = params if params is not None else self.params
        return self.__class__(self.genotype, params, generation, uid, self.uid, self.random_gen)

    def mutate(self, generation=None, uid=None, params=None, sigma=None):
        """Create a new creature by mutating this one

        :params:  the new members parameters are usually inherited verbatim from the parent
                  member, but we may change them by providing this argument with a non-None value.
        """
        new_genotype = self.mutate_genotype(sigma=sigma)
        params = params if params is not None else self.params
        return self.__class__(new_genotype, params, generation, uid, self.uid, self.random_gen)

    def mutate_genotype(self, n_mutation=None, sigma=None):
        if n_mutation is None:
            n_mutation = self.params['mutation']['global']['n_mutation']
        if sigma is None:
            sigma = self.params['mutation']['global']['sigma']
        if n_mutation > 0:
            new_genotype = list(self.genotype)
            m_index = self.random_gen.sample(range(len(new_genotype)), n_mutation)
            for i in m_index:
                new_genotype[i] = self._disturb(new_genotype[i], sigma=sigma)
            return new_genotype
        else:
            return self.genotype

    def _disturb(self, v, sigma):
        return evoutils.gaussian_mutation(v, (0, 1), sigma, self.random_gen)

    def copy_exp(self, other_member):
        """Copy all experience from one member to another.

        Useful when a member is evaluated in another thread."""
        self.fitness = other_member.fitness
