import os
import shutil
from pathlib import Path

import reproducible

from .. import utils


class ResultPackage:
    """Handles saving and loading results

    A result package is composed of:
    - `seed1/manifest.yaml`: the manifest file, listing other files of the package.
    - `seed1/params.yaml`  : the parameter file, specifying every parameter values to
        relaunch the same experiment if needed.
    - `seed1.repro.yaml`   : the reproducible, describing the computational context
        in which the computation happened. Currently, multiple reproducible file
        are not supported for resumed (or continued) computations. #TODO
    - `seed1/chunk_0001.msgpack.xz`: a chunk file, containing the result of the
        computation for a range of generation. The data in the file is serialized using
        the msgpack format (msgpack.org), and then compressed using LZMA ('.xz' extension).
        A result package can contain several chunks, not necessarily of the same length.
    - `seed1/tmpchunk.msgpack.xz`: a partial chunk, using during computation to save
        the result frequently and enable fast resume in the event the computation
        is brutally terminated. Usually not present in finished computations.
    - `seed1/done`: an empty file created when the computation is finished. If absent,
        the simulation is running or need to be resumed.
    - `seed1/lock`: an empty file, whose presence prevent modification of the manifest,
        or creation of new files.

    :param run:     The `Run` instance to save result from
    :param resume:  if True, will attempt to load the manifest file if it exists and load the
                    necessary data to continue the computation.
    :param frozen:  if True, won't change existing file or add new ones under any circumstances.
                    Will also attempt to load all existing data. Overrides the `resume` option.
    """

    """
    state() # provide the current state as (valid_range, state).
    free(start, end) # remove memory about the start-end range of generation (except the last generation)
    load() # load a (maybe partial) state.
    """

    # TODO: recover gracefully from corruption.
    # FIXME: duplicate repro_0 files in manifest

    manifest_filename = 'manifest.yaml'
    params_filename   = 'params.yaml'
    done_filename     = 'done'
    lock_filename     = 'lock'
    repro_filepattern = 'repro_{}.yaml'
    chunk_filepattern = 'chunk_{:04d}.msgpack.xz'
    tmpchunk_filename = 'tmpchunk.msgpack.xz'

    @classmethod
    def load_params(self, manifest_filepath):
        """Only load the parameter settings without loading the package."""
        if not str(manifest_filepath).endswith(self.manifest_filename):
            manifest_filepath = os.path.join(manifest_filepath, self.manifest_filename)
        manifest = utils.load_yaml(manifest_filepath)

        params_filepath = os.path.join(os.path.dirname(manifest_filepath),
                                       manifest['params']['filename'])
        return utils.load_config(params_filepath)

    @classmethod
    def verify(cls, manifest_filepath, tmpchunk=True, tmpchunk_only=False):
        """Verify the package integrity without loading it

        :param tmpchunk:      if False, does not verify the integrity of the tmpchunk.
        :param tmpchunk_only: if True, only verify the tmpchunk.
        """
        if tmpchunk_only:
            tmpchunk = True

        manifest = utils.load_yaml(manifest_filepath)
        if manifest is None:
            raise ValueError('manifest {} seems empty'.format(manifest_filepath))
        package_dir = Path(manifest_filepath).parent

        corrupted = []
        for kind, file_desc in cls._file_descs(manifest):
            if kind == 'tmpchunk':
                if not tmpchunk: continue
            else:
                if tmpchunk_only: continue

            if 'sha256' in file_desc:
                actual_sha256 = reproducible.sha256(Path(package_dir, file_desc['filename']))
                if actual_sha256 != file_desc['sha256']:
                    corrupted.append(file_desc['filename'])

        if len(corrupted) > 0:
            raise ValueError("Some files' integrity ({}, in {}) is compromised. Aborting load.".format(
                             ', '.join(corrupted), str(package_dir)))
        return True


    def __init__(self, run, manifest_filepath=None, resume=True, frozen=False,
                       verbose_save=None, verbose_load=None):
        """
        :param verbose_save:  display a message when saving data
                              if None, the value of param['meta']['verbose_save'] is used.
                              (if itself not found, defaults to False)
        :param verbose_load:  display a message when loading data
                              if None, the value of param['meta']['verbose_load'] is used.
                              (if itself not found, defaults to True)
        """
        self.run, self.frozen = run, frozen
        self.verbose_save, self.verbose_load = verbose_save, verbose_load
        if self.verbose_save is None:
            self.verbose_save = self.run.params['meta'].get('verbose_save', False)
        if self.verbose_load is None:
            self.verbose_load = self.run.params['meta'].get('verbose_load', True)

        self.manifest = {'params': None, 'reproducibles': [], 'done': None,
                         'chunks': [], 'tmpchunk': None, 'generation_count': 0}
        self._find_package_dir(manifest_filepath)
        self.locked = Path(self._fullpath(self.lock_filename)).exists() if self.active else False

        loaded = False
        if resume or self.frozen or self.locked:
            loaded = self.load(manifest_filepath)
        if not loaded:
            self.from_scratch(manifest_filepath)

        self.backup_period  = self.run.params['meta'].get('backup_period', None)
        self.save_period    = self.run.params['meta'].get('save_period', None)

    @property
    def active(self):
        return self.package_dir is not None

    def lock(self):
        if self.active:
            Path(self._fullpath(self.lock_filename)).touch()
        self.locked = True

    def unlock(self):
        if self.active:
            os.remove(self._fullpath(self.lock_filename))
        self.locked = False

    def update_manifest(self):
        """Robustly update manifest. Leaves a .bak backup copy"""
        manifest_filepath = self._fullpath(self.manifest_filename)
        utils.export_yaml(self.manifest, manifest_filepath + '.bak')
        shutil.copyfile(manifest_filepath + '.bak', manifest_filepath)

    def from_scratch(self, manifest_filepath=None):
        """Create a repository package without resuming from existing files.

        Note that since the manifest keeps track of which files belong to the packages
        and their sha256, we care little about
        """
        assert not self.frozen and not self.locked # defensive

        # reset the manifest
        self.manifest = {'params': None, 'reproducibles': [], 'done': None,
                         'chunks': [], 'tmpchunk': None, 'generation_count': 0}

        if manifest_filepath is None:
            self.package_dir = self.run.params['meta'].get('save_filepath', None)
        else:
            if str(manifest_filepath).endswith(self.manifest_filename):
                self.package_dir = os.path.dirname(manifest_filepath)
            else:
                self.package_dir = manifest_filepath

        if self.active:
            os.makedirs(self.package_dir, exist_ok=True)

            # create parameter file
            utils.export_yaml(self.run.params, self._fullpath(self.params_filename))
            sha256 = self.run.context.add_file(self._fullpath(self.params_filename),
                                               category='input', already=False)
            self.manifest['params'] = {'filename': self.params_filename, 'sha256': sha256}

            # create reproducible file
            assert len(self.run.contexts) == 1 and len(self.manifest['reproducibles']) == 0
            self.add_new_reproducible_file(manifest_update=False)

            # writing manifest file
            self.update_manifest()

    def _find_package_dir(self, manifest_filepath=None):
        if manifest_filepath is None:
            self.package_dir = self.run.params['meta'].get('save_filepath', None)
        else:
            if str(manifest_filepath).endswith(self.manifest_filename):
                self.package_dir = os.path.dirname(manifest_filepath)
            else:
                self.package_dir = manifest_filepath


    def load(self, manifest_filepath=None, verbose=None):
        """Load a package from existing file

        :return:   True if data was loaded, else False.
        """
        if not self.active:
            return False

        if not os.path.isfile(self._fullpath(self.manifest_filename)):
            if self.frozen or self.locked or (manifest_filepath is not None):
                err_str = ("the manifest file {} was not found"
                           .format(self._fullpath(self.manifest_filename)))
                raise FileNotFoundError(err_str)
            else:
                return False

        self.verify(self._fullpath(self.manifest_filename), tmpchunk=False)
        self.manifest = utils.load_yaml(self._fullpath(self.manifest_filename))

        # loading data
        loaded = False
        if self.frozen:
            for chunk_desc in self.manifest['chunks']:
                chunk_data = utils.load_msgpack_xz(self._fullpath(chunk_desc['filename']))
                self.run.load_chunk(chunk_data)
                loaded = True
            if self.manifest['tmpchunk'] is not None:
                tmpchunk_filepath = self._fullpath(self.manifest['tmpchunk']['filename'])
                tmpchunk_data = utils.load_msgpack_xz(tmpchunk_filepath)
                self.run.load_chunk(tmpchunk_data)
                loaded = True
        else:  # only load the last chunk
            if self.manifest['tmpchunk'] is not None:
                try:
                    tmpchunk_filepath = self._fullpath(self.manifest['tmpchunk']['filename'])
                    self.verify(self._fullpath(self.manifest_filename), tmpchunk_only=True)
                    tmpchunk_data = utils.load_msgpack_xz(tmpchunk_filepath)
                    self.run.load_chunk(tmpchunk_data)
                    loaded = True
                except ValueError: # tmpchunk is corrupted. happens.
                    if len(self.manifest['chunks']) == 0:
                        return False  # nothing to load
                    else:
                        chunk_filename = self.manifest['chunks'][-1]['filename']
                        chunk_data = utils.load_msgpack_xz(self._fullpath(chunk_filename))
                        self.run.load_chunk(chunk_data)
                        # restore manifest to previous state
                        self.manifest['tmpchunk'] = None
                        self.manifest['generation_count'] = self.run.population.generation_count
                        # remove tmpchunk
                        self.run.context.untrack_file(tmpchunk_filepath, category='output', notfound_ok=True)
                        os.remove(tmpchunk_filepath)
                        self.update_manifest()
                        loaded = True
            else:
                if (not loaded) and len(self.manifest['chunks']) > 0:
                    chunk_filename = self.manifest['chunks'][-1]['filename']
                    chunk_data = utils.load_msgpack_xz(self._fullpath(chunk_filename))
                    self.run.load_chunk(chunk_data)
                    loaded = True

        # add a new reproducible file
        if not (self.frozen or self.locked) and loaded and self.manifest['done'] is None:
            self.add_new_reproducible_file(manifest_update=True)

        gen_count = self.run.population.generation_count
        self.manifest['generation_count'] = gen_count

        if verbose is None:
            verbose = self.verbose_load
        if verbose and loaded:
            done_str = ', finished' if self.manifest['done'] is not None else ''
            if self.frozen:
                print('loaded all data from {} (gen. {}{}, frozen)'.format(self.package_dir,
                                                                           gen_count, done_str))
            else:
                print(('loaded the last checkpoint from {} (gen. {}); '
                       'ready to resume computation{}.').format(self.package_dir,
                                                                gen_count, done_str))

        return loaded


    def add_new_reproducible_file(self, manifest_update=True):
        # add previous tracked files to this reproducibility trace?
        # for i in range(len(self.manifest['reproducibles']) - 1):
        #     repro_filename = self.repro_filepattern.format(i)
        #     repro_filepath = self._fullpath(repro_filename)
        #     previous_context = utils.load_yaml(repro_filepath)
        #     if 'files' in previous_context:
        #         self.run.context.data.setdefault('files', {})
        #         if 'input' in previous_context['files']:
        #             self.run.context.data['files'].setdefault('input', {})
        #             self.run.context.data['files']['input'].update(previous_context['files']['input'])
        #         if 'output' in previous_context['files']:
        #             self.run.context.data['files'].setdefault('output', {})
        #             self.run.context.data['files']['output'].update(previous_context['files']['output'])

        repro_filename = self.repro_filepattern.format(len(self.manifest['reproducibles']))
        repro_filepath = self._fullpath(repro_filename)
        self.run.context.export_yaml(repro_filepath)
        repro_desc = {'filename': repro_filename, 'sha256': reproducible.sha256(repro_filepath)}
        self.manifest['reproducibles'].append(repro_desc)

        # updating manifest file
        if manifest_update:
            self.update_manifest()

    def update_reproducible_file(self, manifest_update=True):
        assert len(self.manifest['reproducibles']) > 0
        repro_filename = self.repro_filepattern.format(len(self.manifest['reproducibles']) - 1)
        repro_filepath = self._fullpath(repro_filename)
        self.run.contexts[-1].export_yaml(repro_filepath)
        repro_desc = {'filename': repro_filename, 'sha256': reproducible.sha256(repro_filepath)}
        self.manifest['reproducibles'][-1] = repro_desc

        # updating manifest file
        if manifest_update:
            self.update_manifest()


    # Saving

    def maybe_save(self):
        """Save or backup if scheduled by the `backup_period` or `save_period` parameters"""
        if self.active:
            gen_count = self.run.population.generation_count
            if   self.save_period   is not None and gen_count % self.save_period == 0:
                self.save()
            elif self.backup_period is not None and gen_count % self.backup_period == 0:
                self.backup()
            # else, nothing to do

    def save(self, verbose=None):
        """Save the current run, and free unecessary data from memory.

        :param verbose:  if None, default to the value of `self.verbose_save`.
        """
        self.save_without_flushing(verbose=verbose)
        self.run.free()

    def save_without_flushing(self, verbose=None):
        """Save all the data to disk, without flushing.

        :param verbose:  if None, default to the value of `self.verbose_save`.
        """
        assert not (self.frozen or self.locked)  # defensive

        if self.manifest['done'] is not None:
            return

        if self.active:
            state = self.run.state()
            chunk_filename = self.chunk_filepattern.format(len(self.manifest['chunks']))
            chunk_filepath = self._fullpath(chunk_filename)
            msgpack_filepath = utils.save_msgpack_xz(state, chunk_filepath)
            assert os.path.realpath(chunk_filepath) == os.path.realpath(msgpack_filepath)
            sha256 = self.run.context.add_file(chunk_filepath, category='output', already=False)
            chunk_desc = {'filename': chunk_filename, 'sha256': sha256}
            self.manifest['chunks'].append(chunk_desc)

            # remove tmpchunk file, containing redundant data.
            tmpchunk_filepath = self._fullpath(self.tmpchunk_filename)
            if os.path.isfile(tmpchunk_filepath):
                self.run.context.untrack_file(tmpchunk_filepath, category='output', notfound_ok=True)
                os.remove(tmpchunk_filepath)
                self.manifest['tmpchunk'] = None

            self.manifest['generation_count'] = self.run.population.generation_count
            is_done = self.run.params['population']['n_generation'] <= self.manifest['generation_count']

            if is_done:
                self.manifest['done'] = {'filename': self.done_filename}

            # update reproducible file and manifest
            self.update_reproducible_file(manifest_update=True)
            self.verify(self._fullpath(self.manifest_filename))

            # only touching the done file after the manifest has been updated
            if is_done:
                Path(self._fullpath(self.done_filename)).touch()

        if verbose is None:
            verbose = self.verbose_save
        if verbose:
            print('saved results in {} at generation {}.'.format(os.path.dirname(self._fullpath(self.manifest_filename)),
                                                                 self.run.population.generation_count))

    def backup(self):
        """Save the current run in a temporary file.

        Useful in the event of a crash for a fast resume.
        """
        assert not (self.frozen or self.locked)  # defensive

        if self.run.population.generation_count == self.manifest['generation_count']:
            # we just saved, don't backup empty changes
            return

        if self.active:
            state = self.run.state()
            tmpchunk_filepath = self._fullpath(self.tmpchunk_filename)
            bak_filepath = tmpchunk_filepath.replace('.msgpack.xz', '.bak.msgpack.xz')
            msgpack_filepath = utils.save_msgpack_xz(state, bak_filepath)
            # print(bak_filepath, msgpack_filepath)
            assert os.path.realpath(bak_filepath) == os.path.realpath(msgpack_filepath)
            os.rename(bak_filepath, tmpchunk_filepath)

            sha256 = self.run.context.add_file(tmpchunk_filepath, category='output', already=True)
            self.manifest['tmpchunk'] = {'filename': self.tmpchunk_filename, 'sha256': sha256}
            self.manifest['generation_count'] = self.run.population.generation_count

            # update reproducible file and manifest
            self.update_reproducible_file(manifest_update=True)
            self.verify(self._fullpath(self.manifest_filename))

        if self.verbose_save:
            print('temporary backup at generation {}.'.format(self.run.population.generation_count))


    # Auxiliary functions

    def _fullpath(self, filename):
        return os.path.join(self.package_dir, filename)

    @classmethod
    def _file_descs(cls, manifest_dict):
        """Iterate over all file descriptions dictionary"""
        # if os.path.isfile(self._fullpath(self.manifest_filename)):
        #     yield {'filename': self.manifest_filename}
        if manifest_dict['params'] is not None:
            yield 'params', manifest_dict['params']
        for repro_desc in manifest_dict['reproducibles']:
            yield 'reproducible', repro_desc
        for chunk_desc in manifest_dict['chunks']:
            yield 'chunk', chunk_desc
        if manifest_dict['tmpchunk'] is not None:
            yield 'tmpchunk', manifest_dict['tmpchunk']
        if manifest_dict['done'] is not None:
            yield 'done', manifest_dict['done']
