"""Evolution with multiprocessing support"""

import time
import multiprocessing as mp

from .. import devo
from .. import utils
from .population_serial import PopulationSerial, evaluate_one_member



class Population(PopulationSerial):
    """
    :param params:  dictionary of parameters. Aside from population-specific
                    parameters, this should also contain the member and
                    environment parameters. See the file entire
                    `conf_template.yaml` for details.
    """

    def evaluate_generation(self, members, gen_count_increase=True, eval_seed=None):
        """Evaluate a list of members of the population

        Already evaluated members (i.e. with a non-None fitness) will not be evaluated again.
        Note that other functions expect the returned members in the same order as input.

        :param eval_seed:  if None, a seed based on each member uid will be used.
                           if not None, the same seed will be used for all members.
        """
        to_evaluate = [m for m in members if m.fitness is None]

        if self.params['meta']['multiprocessing'] and utils.n_available_cores() > 1:
            if self.pool is None:
                pp = self.params['population']
                n = pp['n_mutant'] + pp['n_random']
                if pp['reevaluate']:
                    n += pp['n_elites']
                self.n_proc = min(utils.n_available_cores(), n) # FIXME: if in slurm, always use the number of cores provided
                self.pool = mp.Pool(self.n_proc)
                utils.print_info('Using {} processes'.format(self.n_proc))

            to_evaluate_map = {m.uid: m for m in to_evaluate}
            evaluated = self.pool.map(evaluate_one_member, to_evaluate)
            for m_e in evaluated:
                to_evaluate_map[m_e.uid].copy_exp(m_e) # copy the experience.
                if len(m_e.events) > 0:
                    self.events[m_e.uid] = m_e.events

        else:  # serial execution
            for m in to_evaluate:
                evaluate_one_member(m, eval_seed=eval_seed)
                if len(m.events) > 0:
                    self.events[m.uid] = m.events

        if gen_count_increase:
            self.generation_count += 1
            assert len(self.generations) == self.generation_count, '{} != {}'.format(len(self.generations), self.generation_count)

    def close(self):
        """Clean-up multiprocessing pool before exiting python.

        Multiprocessing has issues. See https://bugs.python.org/issue29759
        """
        if self.pool is not None:  # being quite defensive here
            self.pool.close()
            self.pool.join()
            self.pool.terminate()
            self.pool = None



class DevelopingPopulation(Population):

    def __init__(self, params, *args, **kwargs):
        self.dev_vars = {}
        for name, dev_params in params['population'].get('dev_vars', {}).items():
            self.dev_vars[name] = devo.TrajectoryGenerator(dev_params)

        super().__init__(params, *args, **kwargs)


    def pre_gen(self):
        """Update the member parameters according to the developmental trajectories"""

        if len(self.dev_vars) > 0:
            new_params = self.repository.new_parameters(self.repository.last_member_params)
            # `new_params` now *is* `self.repository.last_member_params`

        for name, trajgen in self.dev_vars.items():
            new_val = trajgen(self.generation_count)
            if name == 'n_parent':

                self.params['population']['n_parent'] = int(round(new_val))
                print(self.params['population']['n_parent'])
            elif name == 'n_elites':
                self.params['population']['n_elites'] = int(round(new_val))
            elif name == 'n_mutant':
                self.params['population']['n_mutant'] = int(round(new_val))
            elif name == 'mutation_sigma':
                new_params['mutation']['global']['sigma']      = new_val
                new_params['mutation']['morphology']['sigma']  = new_val
                new_params['mutation']['controller']['sigma']  = new_val
                new_params['mutation']['development']['sigma'] = new_val
            else:
                new_params = self.member_class.develop_params(new_params, name, new_val)


class ControllerDevelopingPopulation(DevelopingPopulation) :
    """Initialize a random morphology genotype (useful for a StarfishCreature)"""

    def __init__(self, params, random_gen, *args, **kwargs):
        if params['member'].get('morphology_genotype', None) is None:
            member_class = utils.import_str(params['member']['class'])
            params['member']['morphology_genotype'] = member_class.random_morphology_genotype(params['member'], random_gen)
        super().__init__(params, random_gen, *args, **kwargs)
