import os
import time
import copy
import random
from pathlib import Path

import psutil
import tqdm
import numpy as np

from . import evoutils
from .repo import MemberRepository
from .. import utils
from .. import paths



def evaluate_one_member(member, eval_seed=None, screenshot_folder=None):
    """Evaluate one member

    Kept outside of the Population class to avoid pickle issues during multiprocessing.
    """
    env_class = utils.import_str(member.params['environment']['class'])
    eval_random_gen = random.Random()  # independent random generator
    if eval_seed is None:
        eval_seed = member.eval_seed_offset + member.uid
    eval_random_gen.seed(eval_seed)
    env_params = copy.deepcopy(member.params['environment'])
    env_params['screenshot_folder'] = screenshot_folder
    env = env_class(env_params, eval_random_gen)
    result = env.run(member)
    return result


class PopulationSerial:
    """
    :param params:  dictionary of parameters. Aside from population-specific
                    parameters, this should also contain the member and
                    environment parameters. See the file entire
                    `conf_template.yaml` for details.
    """
    def __init__(self, params, random_gen, eval_seed_offset=None, frozen=False):
        self.params, self.random_gen, self.frozen = params, random_gen, frozen
        # creating an offset to compute evaluation seeds out of
        self.eval_seed_offset = params['meta']['seed'] * 10e10

        self.member_class = utils.import_str(self.params['member']['class'])
        self.repository  = MemberRepository(self.params['member'], self.random_gen)
        self.generations = []
        self.generation_count = 0

        self.pool = None  # initialized if necessary at first evaluate_generation() invocation.
        self.events = {}

        self.n_proc = 1
        self.stiffness_table = None

    def from_state(self, state):
        self.generations = state['generations']
        self.repository.from_state(state['repository'])
        self.generation_count = state['generation_count']

    @property
    def n_parent(self):
        try:
            return self.params['population']['n_parent']
        except KeyError:
            return self.params['population']['n_survivor']

    @property
    def n_candidates(self):
        """Number of candidate members per generation"""
        pp = self.params['population']
        n_pop = pp['n_mutant'] + pp['n_random']
        if pp['parents_survive']:
            n_pop += self.n_parent
        return n_pop


    def _first_generation(self):
        self.pre_gen()
        self.generations.append([])
        pp = self.params['population']

        n_pop = self.n_candidates
        if isinstance(pp.get('initial_pop', None), int):
            n_pop = pp['initial_pop']

        elif pp.get('initial_pop', None) is not None:  # assume it's a path to file with genotypes
            with open(os.path.join(paths.DATA_DIR, pp['initial_pop']), 'rb') as fd:
                genotypes = [[float(f) for f in genotype] for genotype in np.load(fd)]
            new_gen = []
            for genotype in genotypes:
                c = self.repository.manual_member(genotype,
                                                  params=self.repository.last_member_params,
                                                  generation=self.generation_count)
                self.generations[-1].append(c.uid)
                new_gen.append(c)
            self.post_gen()
            return new_gen

        new_gen = []
        for _ in range(n_pop):
            c = self.repository.random_member(params=self.repository.last_member_params,
                                              generation=self.generation_count)
            self.generations[-1].append(c.uid)
            new_gen.append(c)
        self.post_gen()
        return new_gen


    # Running generations

    def one_generation(self):
        """Run one generation"""
        assert not self.frozen
        if self.generation_count == 0:
            new_gen = self._first_generation()
        else:
            previous_gen = [self.repository.member(uid) for uid in self.generations[-1]]
            parents      = self.select_parents(previous_gen)
            new_gen      = self.new_generation(parents, previous_gen)
        self.evaluate_generation(new_gen)


    def evaluate_member(self, member, eval_seed=None, screenshot_folder=None):
        return evaluate_one_member(member, eval_seed=eval_seed, screenshot_folder=screenshot_folder)

    def evaluate_generation(self, members):
        """Evaluate a list of members of the population

        Already evaluated members (i.e. with a non-None fitness) will not be evaluated again.
        Note that other functions expect the returned members in the same order as input.
        """
        to_evaluate = [m for m in members if m.fitness is None]

        for m in to_evaluate:
            evaluate_one_member(m)
            if len(m.events) > 0:
                self.events[m.uid] = m.events

        self.generation_count += 1
        assert len(self.generations) == self.generation_count


    def select_parents(self, members):
        """Select the members to survive the next generation"""
        parents, dead = evoutils.parent_selection(members, self.n_parent, self.random_gen,
                                                  pareto_age=self.params['population'].get('pareto_age', True))
        return parents


    def new_generation(self, parents, previous_gen):
        """Create a new generation from the parents members"""
        # any development goes here
        self.pre_gen()
        self.generations.append([])

        # select elites
        elites = []
        if self.params['population'].get('n_elites', 0) > 0:
            sorted_previous_gen = sorted(previous_gen, key=lambda x: x.fitness, reverse=True)
            elites = sorted_previous_gen[:self.params['population'].get('n_elites', 0)]
        if self.params['population']['parents_survive']:
            # avoid doublons in parents and elites
            elites = [member for member in elites if member not in parents]
        if self.params['population']['reevaluate']:
            # if reevaluating elites, create copies.
            elites  = [self.repository.copy(m, params=self.repository.last_member_params,
                                            generation=self.generation_count)
                       for m in elites]
        self.generations[-1].extend([m.uid for m in elites])

        # keep or copy the parents?
        if self.params['population']['parents_survive']:
            # if reevaluating parents, create copies.
            if self.params['population']['reevaluate']:
                parents = [self.repository.copy(m, params=self.repository.last_member_params,
                                                generation=self.generation_count)
                           for m in parents]
            self.generations[-1].extend([m.uid for m in parents])

        # create mutants out of the parents
        for k in range(self.params['population']['n_mutant']):
            member = parents[k % len(parents)]
            m = self.repository.mutate(member, params=self.repository.last_member_params,
                                       generation=self.generation_count)
            self.generations[-1].append(m.uid)

        # add new random members
        for _ in range(self.params['population']['n_random']):
            m = self.repository.random_member(params=self.repository.last_member_params,
                                              generation=self.generation_count)
            self.generations[-1].append(m.uid)

        # archive dead members to save memory.
        alive_uids = set(self.generations[-1])
        for member in previous_gen:
            if member.uid not in alive_uids:
                self.repository.archive(member.uid)

        self.post_gen()

        return [self.repository.member(uid) for uid in self.generations[-1]]


    def pre_gen(self):
        """Anything to be done before the new generation members are created

        This is the perfect place to create new `self.repository.last_member_params`.
        """
        pass

    def post_gen(self):
        """Anything to be done after the new generation members are created"""
        if self.params['member'].get('adjusted_stiffness', False):
            if self.stiffness_table is None:
                mp = self.params['meta']
                path = Path(paths.DATA_DIR, 'stiffnesses', mp['exp_name'],
                            'v{:03d}.{:03d}.pickle'.format(mp['major'], mp['minor']))
                self.stiffness_table = utils.load_data(path)
            for uid in self.generations[-1]:
                m = self.repository.member(uid)
                mass  = m.params['materials']['node_section']['mass']
                stiff = m.params['materials']['link_sec_side']['stiffness']
                m.params['materials']['link_sec_side']['adjusted_stiffness'] = self.stiffness_table[(stiff, mass)]

    # Manual Generation

    def manual_generation(self):
        """Create a new empty generation"""
        self.pre_gen()
        self.generations.append([])

    def manual_member(self, genotype=None, params=None, puid=None):
        """Add a new creature with a given genotype to the current generation

        :param genotype:  if None, creates a random genotype.
        :param params:    if None, defaults to the base member parameters of the population.
        """
        if genotype is None:
            new_m = self.repository.random_member(params, generation=self.generation_count)
        else:
            new_m = self.repository.manual_member(genotype, params, generation=self.generation_count,
                                                  puid=puid)
        self.generations[-1].append(new_m.uid)
        return new_m

    def evaluate_all(self, force=False, all_generations=False):
        """Evaluate all non-evaluated member

        :param force:  if True, reevaluate even already-evaluated every members.
        :param all_generations:  if True, reevaluate previous generation as well.
        """
        if all_generations:
            raise NotImplementedError
        to_evaluate = [m for m in self.last_generation() if force or m.fitness is None]
        self.evaluate_generation(to_evaluate)

    def last_generation(self):
        for uid in self.generations[-1]:
            yield self.repository.member(uid)


    # State, Load, free

    def state(self):
        return {'generations'        : self.generations,
                'generation_count'   : self.generation_count,
                'repository_state'   : self.repository.state()}

    def load_chunk(self, state):
        self.repository.load_chunk(state['repository_state'])
        self.generation_count    = state['generation_count']
        self.generations.extend(state['generations'][len(self.generations):])
        # temporary code
        for i, gen in enumerate(self.generations):
            if gen is None: assert state['generations'][i] is None
        # end temp

    def free(self):
        self.repository.free(self.generations[-1])
        for i in range(len(self.generations)-1):
            self.generations[i] = None
