import copy

from .. import utils


class MemberRepository:
    """Hold description of the member of the population.

    Member parameter hold any parameters not in the genotype. They are usually shared by
    the member of a same generation, and sometimes by all members of all generations. To
    memory and file size, the repository holds these parameters separately from the members.
    """


    def __init__(self, params, random_gen, eval_seed_offset=0):
        self.random_gen = random_gen
        self.member_class = utils.import_str(params['class'])
        self.eval_seed_offset = eval_seed_offset

        self.parameter_count = 0
        self.member_count    = 0

        self.members    = {}  # uid -> Member
        self.parameters = {}  # uid -> parameter dictionary
        self.archives   = {}  # uid -> Member state tuple

        self.new_parameters(params)

    @property
    def uids(self):
        """list of all uids"""
        return [uid for uid in self.archives.keys()] + [uid for uid in self.members.keys()]

    def member(self, uid):
        """Return a member (reinstanciate and dearchive if archived) from a uid"""
        try:
            return self.members[uid]
        except KeyError:
            m_state = self.archives.pop(uid)
            m = self.member_class.from_state(m_state, self, self.random_gen)
            m.eval_seed_offset = self.eval_seed_offset
            self.members[uid] = m
            return m

    def member_state(self, uid):
        """Return a member state (directly from archive or generated just-in-time from a uid.

        Does not archive member if not archived.
        """
        try:
            return self.archives[uid]
        except KeyError:
            return self.members[uid].state

    def new_parameters(self, params):
        """Return a fresh deep copy of the parameters, ready to be modified.

        Parameters should never be changed after being used on evaluated members without using
        this function.
        """
        new_params = copy.deepcopy(params)
        new_params['_param_uid'] = self.parameter_count
        self.parameters[self.parameter_count] = new_params
        self.parameter_count += 1
        self.last_member_params = new_params
        return new_params

    def member_parameters(self, param_uid=None):
        """Return the member parameters corresponding to the given uid"""
        if param_uid is None:
            param_uid = self.parameter_count - 1
        return self.parameters[param_uid]


    def copy(self, member, params=None, generation=None):
        """Return a copy of a member with a new uid.

        Useful to change the parameters but not the genotype.
        :member:  the member to be copied.
        :params:  new member parameters, or copied for the original member is None.
        """
        new_m = member.copy(generation=generation, uid=self.member_count, params=params)
        self.members[new_m.uid] = new_m
        self.member_count += 1
        new_m.eval_seed_offset = self.eval_seed_offset
        return new_m

    def mutate(self, member, params=None, generation=None, sigma=None):
        """Return a new mutated member from an existing one.

        :member:  the parent of the new mutated member.
        :params:  the new members parameters are usually inherited verbatim from the parent
                  member, but we may change them by providing this argument with a non-None value.
        """
        new_m = member.mutate(generation=generation, uid=self.member_count, params=params,
                              sigma=sigma)
        self.members[new_m.uid] = new_m
        self.member_count += 1
        new_m.eval_seed_offset = self.eval_seed_offset
        return new_m

    # FIXME: remove params as arguments?
    def random_member(self, params=None, generation=None):
        """Return a new randomly created member"""
        if params is None:
            params = self.member_parameters()

        new_m = self.member_class.random(params, generation=generation, uid=self.member_count,
                                         random_gen=self.random_gen)
        self.members[new_m.uid] = new_m
        self.member_count += 1
        new_m.eval_seed_offset = self.eval_seed_offset
        return new_m

    def manual_member(self, genotype, params=None, generation=None, puid=None):
        if params is None:
            params = self.member_parameters()
        new_m = self.member_class(genotype, params, generation=generation, uid=self.member_count,
                                  puid=puid, random_gen=self.random_gen)
        self.members[new_m.uid] = new_m
        self.member_count += 1
        new_m.eval_seed_offset = self.eval_seed_offset
        return new_m

    def archive(self, uid):
        """Archive a (probably dead or past) member to reduce the memory footprint

        :raise AssertionError:  if member does not have a fitness, or can't be found in active or
                                archived members.
        """
        try:
            m = self.members[uid]
            assert m.fitness is not None
            self.archives[uid] = m.state
            self.members.pop(uid)
        except KeyError:
            assert uid in self.archives

    def archive_all(self):
        for uid in list(self.members.keys()):  # `list()` because we're modifying `self.members`.
            self.archive(uid)


    # State, Load, Free

    def state(self):
        """Return a copy of the state of the repository, useful for saving to disk."""
        archives = copy.copy(self.archives)
        for m in self.members.values():
            archives[m.uid] = m.state
        return {'archives': archives,
                'member_count'   : self.member_count,
                'parameters'     : self.parameters,
                'parameter_count': self.parameter_count}

    def load_chunk(self, state):
        assert self.member_count <= state['member_count']
        self.archives.update(state['archives'])
        self.member_count = state['member_count']
        self.parameters.update(state['parameters'])
        self.parameter_count = state['parameter_count']

    def free(self, protected_uids):
        """Remove all members from the repository except the given uids"""
        self.members = {uid: self.member(uid) for uid in protected_uids}
        self.archives = {}
        protected_param_uids = [m.params['_param_uid'] for m in self.members.values()]
        self.paramters = {param_uid: self.parameters[param_uid]
                          for param_uid in protected_param_uids}
