import os
import time
import random

import tqdm
import reproducible

from .. import utils
from . import population
from . import repo
from . import package



class LightRun:

    def __init__(self, params, random_gen):
        self.params, self.random_gen = params, random_gen
        self.population_class = utils.import_str(self.params['population']['class'])
        self.population = self.population_class(self.params, self.random_gen)

    def run_generations(self, target_gen=None, progress_bar=True):
        """Run generations up until a generation number is reached

        :param target_gen:    oldest generation to compute. If `None`, the params value is used.
        :param progress_bar:  if True, displays a progress_bar
        """
        target_gen = self.params['population']['n_generation'] if target_gen is None else target_gen

        for t in range(self.population.generation_count, target_gen):
            self.population.one_generation()

    def state(self):
        return {'random_state': self.random_gen.getstate(), 'params': self.params,
                'generations': self.population.generations,
                'repository': self.population.repository.state()}


class Run:

    @classmethod
    def from_results(cls, filepath, resume=True, frozen=True, verbose=None, **kwargs):
        params = package.ResultPackage.load_params(filepath)
        return cls(params, resume=resume, frozen=frozen, manifest_filepath=filepath,
                   verbose_load=verbose, **kwargs)

    @classmethod
    def from_paramfile(cls, param_filepath, resume=False, frozen=False, **kwargs):
        params = utils.load_config(param_filepath)
        return cls(params, resume=resume, frozen=frozen, **kwargs)

    @classmethod
    def verify(cls, manifest_filepath):
        return package.ResultPackage.verify(manifest_filepath)

    @property
    def context(self):
        return self.contexts[-1]

    def __init__(self, params, random_gen=None, resume=True, frozen=False, manifest_filepath=None,
                       verbose_load=None, verbose_save=None, run_seed=None):
        """
        :param params:      parameters of the run.
        :param random_gen:  random generator. If None, one will be created.
        :param frozen:      if True, the run cannot compute new generations. Avoids creating
                            a context.
        """
        self.params = params
        self.random_gen = random_gen
        self.frozen = frozen

        self.contexts = []
        self.generation_contexts = []  # for each generation, the index of the context is was
                                       # computed in.
        if not self.frozen:
            self.contexts.append(reproducible.Context())
            self.context.add_editable_repos(allow_dirty=True)
            self.context.add_repo(allow_dirty=True)
        if not self.frozen or run_seed is not None:
            self.random_init(run_seed)

        self.population_class = utils.import_str(self.params['population']['class'])
        self.population = self.population_class(self.params, self.random_gen, frozen=self.frozen)

        # will load data if necessary
        self.results = package.ResultPackage(self, resume=resume, frozen=self.frozen,
                                                   manifest_filepath=manifest_filepath,
                                                   verbose_load=verbose_load,
                                                   verbose_save=verbose_save)

    def __eq__(self, other):
        """Equivalence comparison between run.

        Does not compare contexts, only uids, random state, fitness, and genotypes.
        Comparing uid is necessary as they determine evaluation random generators.
        """
        equal  = self.population.generations == other.population.generations
        if self.random_gen is not None and other.random_gen is not None:
            equal &= self.random_gen.getstate()  == other.random_gen.getstate()
        repo_self, repo_other = self.population.repository, other.population.repository
        if not equal: return False
        equal &= set(repo_self.uids) == set(repo_other.uids)
        for uid in repo_self.uids:
            ms_self, ms_other = repo_self.member_state(uid), repo_other.member_state(uid)
            equal &= (ms_self[repo_self.member_class.genotype_index] ==
                      ms_other[repo_other.member_class.genotype_index])
            equal &= (ms_self[repo_self.member_class.fitness_index] ==
                      ms_other[repo_other.member_class.fitness_index])
        return equal

    def random_init(self, seed=None):
        if self.random_gen is None:
            self.random_gen = random.Random()
            if seed is None:
                seed = self.params['meta'].get('seed', None)
            if seed is None:
                seed = self.random_gen.randint(0, 10e10)
            if not self.frozen:
                self.context.add_data('seed', seed)
            self.random_gen.seed(seed)

    def lock(self):
        return self.results.lock()

    def unlock(self):
        return self.results.unlock()


    def run_generations(self, target_gen=None, progress_bar=None):
        """Run generations up until a generation number is reached

        :param target_gen:    oldest generation to compute. If `None`, the params value is used.
        :param progress_bar:  if True, displays a progress_bar. if None
        """
        if self.results.locked:
            raise RuntimeError('trying to compute new generations on a locked run.')
        if self.frozen:
            raise RuntimeError('trying to compute new generations on a frozen or locked run.')

        if progress_bar is None:
            progress_bar = self.params['meta'].get('progress_bar', True)
        if target_gen is None:
            target_gen = self.params['population']['n_generation']
        if self.population.generation_count >= target_gen:
            return  # nothing more to do

        range_gen = range(self.population.generation_count, target_gen)
        generations = tqdm.tqdm(range_gen) if progress_bar else range_gen

        try:
            for t in generations:
                if progress_bar:
                    generations.set_description('{} processes, mem: {:.1f} Mb, '.format(
                                                 self.population.n_proc, utils.max_memory()))
                self.population.one_generation()
                self.generation_contexts.append(len(self.contexts) - 1)
                self.results.maybe_save()
        finally:
            self.population.close()


    # save, state, load, and free

    def save(self, verbose=None):
        self.results.save(verbose)

    def state(self):
        """Provides the current state of the run"""
        contexts = self.contexts
        if isinstance(contexts[-1], reproducible.Context):
            contexts = contexts[:-1] + [contexts[-1].data]

        state_ = {'params'             : self.params,
                  'contexts'           : contexts,
                  'generation_contexts': self.generation_contexts,
                  'random_state'       : self.random_gen.getstate(),
                  'pop_state'          : self.population.state()}
        return state_

    def load_chunk(self, state):
        """Load a previously saved state chunk"""
        self.params = state['params']
        if not self.frozen:
            self.random_gen.setstate(state['random_state'])
            self.contexts = state['contexts'] + [self.context]
        else:
            self.contexts = state['contexts'][:]
        self.generation_contexts.extend(state['generation_contexts'][len(self.generation_contexts):])

        self.population.load_chunk(state['pop_state'])

    def free(self):
        """Free any unneeded data to continue the computation"""
        self.population.free()
        for i in range(len(self.generation_contexts) - 1):
            self.generation_contexts[i] = None
