import numpy as np

from . import devo
from . import utils


_example = {'morphology': ['w', 'N'],
            'genes': {'w': {'kind': 'valuetraj',
                            'length': 2000,
                            'morphology': [{'n_genes': 3, 'function': 'logscale',
                                            'preprocess': [('logscale', {'limits': [0.00001, 1.0]})]},
                                           {'n_genes': 3, 'function': 'logscale',
                                            'preprocess': [('logscale', {'limits': [0.00001, 1.0]})]},
                                           {'n_genes': 3, 'function': 'logscale',
                                            'preprocess': [('logscale', {'limits': [0.00001, 1.0]})]}
                                          ]
                           },
                      'N': {'kind': 'fixed',
                            'morphology': [{'n_genes': 3,
                                            'preprocess': [('intbudget', {'sum': 150, 'min': 10})]}]}
                    }}


class GenotypeDecoder:

    def __init__(self, params):
        self.params = params

    def genotype_length(self):
        length = 0
        for name in self.params['morphology']:
            gene_desc = self.params['genes'][name]
            length += self.gene_length(gene_desc)
        return length

    def gene_length(self, sub_genotype):
        length = 0
        for sub_element in sub_genotype['morphology']:
            length += sub_element['n_genes']
        return length

    def express(self, genes):
        assert len(genes) == self.genotype_length()

        phenotype = {}
        offset = 0
        for name in self.params['morphology']:
            gene_desc = self.params['genes'][name]
            n_gene = self.gene_length(gene_desc)
            gene = genes[offset: offset + n_gene]
            offset += n_gene

            if gene_desc['kind'] == 'valuetraj':
                phenotype[name] = self._express_valuetraj(gene, gene_desc)
            elif gene_desc['kind'] == 'fixed':
                phenotype[name] = self._express_fixed(gene, gene_desc)
            else:
                raise NotImplementedError('{} is not known'.format(gene_desc['kind']))

        return phenotype

    def _express_valuetraj(self, gene, sub_genotype):
        trajectories = []
        offset = 0
        for traj in sub_genotype['morphology']:
            t_milestones = np.linspace(0, sub_genotype['length'], traj['n_genes'])
            v_milestones = gene[offset: offset + traj['n_genes']]
            offset += traj['n_genes']

            if 'preprocess' in traj:
                v_milestones = Processor(traj['preprocess'])(v_milestones)

            trajectories.append({'function': traj['function'],
                                 'milestones': list(zip(t_milestones, v_milestones))})

        params = {'function': 'multi',
                  'trajectories': trajectories}
        return devo.TrajectoryGenerator(params)

    def _express_fixed(self, gene, sub_genotype):
        values = []
        offset = 0
        for sub_element in sub_genotype['morphology']:
            sub_values = gene[offset: offset + sub_element['n_genes']]
            offset += sub_element['n_genes']

            if 'preprocess' in sub_element:
                sub_values = Processor(sub_element['preprocess'])(sub_values)
            values.extend(sub_values)
        return values



class Processor:
    """Pre- or post-process genotype values."""

    def __init__(self, params):
        self.params = params
        self.transforms = []

        transform_map = {'logscale' : self._init_logscale,
                         'linear'   : self._init_linear,
                         'normalize': self._init_normalize,
                         'intbudget': self._init_intbudget}

        for step in params:
            name, stepparams = step
            self.transforms.append(transform_map[name](stepparams))

    def __call__(self, values):
        for transform in self.transforms:
            values = transform(values)
        return values

    def _init_linear(self, params):
        v_min, v_max = params['limits']

        def linear_transform(values):
            return [v_min + (v_max - v_min) * v for v in values]

        return linear_transform

    def _init_logscale(self, params):
        v_min, v_max = params['limits']

        def logscale_transform(values):
            return [utils.log_transform(v, v_min, v_max) for v in values]

        return logscale_transform

    def _init_normalize(self, params):
        target_sum = params['sum']

        def normalize_transform(values):
            sum_v = sum(values)
            assert(all(v >= 0 for v in values) and sum_v > 0)
            ratio = target_sum/sum(values)
            return [ratio*v for v in values]

        return logscale_transform

    def _init_intbudget(self, params):
        target_sum = params['sum']
        v_min = params['min']

        def intbudget_transform(values):
            sum_v = sum(values)
            assert all(v >= 0 for v in values) and sum_v > 0

            target_sum_min = target_sum - len(values) * v_min
            assert target_sum_min >= 0
            ratio = target_sum_min/sum_v
            return [round(ratio*v + v_min) for v in values]

        return intbudget_transform
