"""MPI-Enabled classes"""
import time

import tqdm
from mpi4py import MPI

from .evo import Population
from . import utils

comm = MPI.COMM_WORLD

def mpi_print(*args, root=0, **kwargs):
    if comm.rank == root:
        print(*args, **kwargs)


class MPIPopulation(Population):

    def __init__(self, *args, **kwargs):
        self.is_root = comm.rank == 0
        super().__init__(*args, **kwargs)

    def run_generation(self, headless=True):
        """Run a generation, dispatching evaluation among MPI nodes"""
        # 1. root decides the creatures to evaluate
        if self.is_root:
            creatures = [self.creatures[uid] for uid in self.generations[-1]]
            old_creatures = [c for c in creatures if c.fitness is not None]
            new_creatures = [c for c in creatures if c.fitness is None]
            genotypes = [creature.genotype for creature in new_creatures]
            work_packages = utils.even_divide_list(genotypes, comm.size)
        else:
            work_packages = None

        # 2. creatures to evaluate are dispached
        work_package = comm.scatter(work_packages, root=0)
        # print(comm.rank, work_packages, work_package)
        to_evaluate = [self.creature_factory.creature_from_genotype(genotype)
                       for genotype in work_package]

        # 3. evaluation by nodes
        # headless = headless is self.is_root else False
        evaluated = self.evaluate(to_evaluate, headless=headless)
        raise Error # need to send uid too
        results = [(creature.fitness, creature.pos_history) for creature in evaluated]

        # 4. fitnesses, positions are gathered
        full_results = comm.gather(results)

        # 5. selection, new generation by root
        if self.is_root:
            full_results = utils.flatten_once(full_results)
            # results are set into the creatures
            for creature, result in zip(new_creatures, full_results):  # MPI conserves order.
                fitness, pos_history = result
                creature.fitness = fitness
                creature.pos_history = pos_history

            creatures = old_creatures + new_creatures
            survived = self.selection(creatures)
            self.new_generation(survived)

    def save(self, *args, **kwargs):
        """Avoid saving on MPI nodes other than root"""
        if self.is_root:
            super().save(*args, **kwargs)
