import os
import sys
import math
import random
import subprocess
from pathlib import Path

import numpy as np
from scipy.stats import pearsonr, spearmanr

from ... import evo
from ... import utils
from ... import paths
from ..  import loadlib
from ..  import metrics
from .   import color
from .   import graph
from ..postprocessing import post_analysis


def prepare_boxplot_data_noise(run, confidence):
    data = []
    pop = run.population
    for generation in pop.generations:
        fitnesses = []
        for uid in generation:
            member = pop.repository.member(uid)
            fitnesses.append(member.fitness)
        data_metrics = metrics.metrics_acute(fitnesses[1:], include_data=True,
                                             confidence_median=confidence)
        exp_name = run.params['meta'].get('exp_name', None)
        if exp_name is None:
            exp_name = os.path.dirname(os.path.dirname(run.params['meta']['filename_prefix']))
        data.append({'color': color.control, 'offset': 0.0, 'metrics': data_metrics,
                     'exp': (exp_name, run.params['meta']['major'], run.params['meta']['minor']),
                     'confidence': confidence, 'cfg': run.params, 'original_fitness': fitnesses[0]})
    return data


def prepare_data_gen_fitness(versions, confidence, fitness_function=None):
    data = []
    for version_list, n_gen, offset, colors in versions:
        for exp_name, major, minors in version_list:
            for minor in minors:
                data_i = post_analysis.prepare_data_gen_fitness_run(exp_name, major, minor, n_gen,
                                                                    fitness_function=fitness_function)
                if data_i is not None:
                    data_metrics = metrics.metrics_acute(data_i, include_data=True,
                                                         confidence_median=confidence)
                    data.append({'color': colors, 'offset': offset, 'metrics': data_metrics,
                                 'exp': (exp_name, major, minor), 'confidence': confidence,
                                 'fitness_function': fitness_function,
                                 'cfg': loadlib.load_cfg(exp_name, major, minor)})

    return data

def prepare_data_max_fitness(versions, confidence, fitness_function=None, rotation=None):
    data = []
    for version_list, f_perf, offset, colors in versions:
        for exp_name, major, minors in version_list:
            for minor in minors:
                seeds, data_i = post_analysis.prepare_data_max_fitness_run(exp_name, major, minor, f_perf,
                                                                        fitness_function=fitness_function,
                                                                        rotation=rotation)
                data_metrics = None
                if data_i is not None:
                    data_metrics = metrics.metrics_acute(data_i, include_data=True,
                                                        confidence_median=confidence)

                data.append({'color': colors, 'offset': offset, 'metrics': data_metrics,
                             'exp': (exp_name, major, minor), 'confidence': confidence,
                             'fitness_function': fitness_function, 'seeds': seeds,
                             'cfg': loadlib.load_cfg(exp_name, major, minor)})

    return data

def prepare_data_rotations(versions, confidence, fitness_function=None):
    data = []
    for version_list, f_perf, offset, colors in versions:
        for exp_name, major, minors in version_list:
            for minor in minors:
                cfg = loadlib.load_cfg(exp_name, major, minor)
                filepath = loadlib.angle_filepath_from_cfg(cfg)
                angles = utils.load_data(filepath)
                data_i = list([-a[0] for a in angles.values()])
                data.append({'color': colors, 'offset': offset, 'metrics': None,
                             'exp': (exp_name, major, minor), 'confidence': confidence,
                             'cfg': cfg})

                if data_i is not None:
                    data_metrics = metrics.metrics_acute(data_i, include_data=True,
                                                         confidence_median=confidence)
                    data[-1]['metrics'] = data_metrics
    return data




def genotype_distance_seed(exp_name, major, minor, seed):
    frozen_run = loadlib.load_results(exp_name, major, minor, seed, verbose=False)
    best = evo.analysis.select_best(frozen_run.population, n=1, window=(-1, None))[0]
    genealogy = evo.analysis.genealogy(frozen_run.population, best)
    m = frozen_run.population.repository.member
    d0_2000    = evo.analysis.genotype_distance(m(genealogy[-1]),   m(genealogy[2000]))
    d2000_4000 = evo.analysis.genotype_distance(m(genealogy[2000]), best)
    return d0_2000, d2000_4000

def prepare_data_genotype_dist(versions, confidence):
    data = []
    for version_list, f_perf, offset, colors in versions:
        for exp_name, major, minors in version_list:
            for minor in minors:
                d2000s, d4000s = [], []
                for seed in loadlib.find_done(exp_name, major, minor):
                    d0_2000, d2000_4000 = genotype_distance_seed(exp_name, major, minor, seed)
                    d2000s.append(d0_2000)
                    d4000s.append(d2000_4000)
                for d, g in [(d2000s, 2000), (d4000s, 4000)]:
                    data_metrics = metrics.metrics_acute(d2000s, include_data=True,
                                                         confidence_median=confidence, scale_factor=1.0)

                    data.append({'color': colors, 'offset': offset, 'metrics': data_metrics,
                                 'exp': (exp_name, major, minor), 'confidence': confidence,
                                 'xlabel': 'v{}.{}_{}'.format(major, minor, g),
                                 #'fitness_function': fitness_function,
                                 'cfg': loadlib.load_cfg(exp_name, major, minor)})
                    print(data[-1]['xlabel'], np.mean(d))
    return data


def prepare_data_avg_gen_distance(versions, confidence, window, leap):
    data = []
    for version_list, f_perf, offset, colors in versions:
        for exp_name, major, minors in version_list:
            for minor in minors:
                data_filepath = loadlib.genealogy_distance_best_filepath(exp_name, major, minor, leap=leap)
                distances = loadlib.load_data(data_filepath)

                distances = np.array(distances).T  # now first index is timesteps
                if window is not None:
                    distances = distances[window[0]:window[1]]
                distances = np.mean(distances, axis=0)
                data_metrics = metrics.metrics_acute(distances, include_data=True,
                                                     confidence_median=confidence, scale_factor=1000.0)

                data.append({'color': colors, 'offset': offset, 'metrics': data_metrics,
                             'exp': (exp_name, major, minor), 'confidence': confidence,
                             #'fitness_function': fitness_function,
                             'cfg': loadlib.load_cfg(exp_name, major, minor)})
    return data


def prepare_correlation_distrib(versions, confidence):
    single_path = loadlib.genealogy_distance_correlation_filepath

    count = 0
    p_corrs = []
    for exp_name, major, minors, window, marker, colors in versions:
        # xs, ys, vs = [], [], []
        for minor in minors:
            filepath = Path(paths.POST_DIR, single_path(exp_name, major, minor, window))
            if not filepath.is_file():
                utils.print_warning('correlation_graph: {} not found, skipping'.format(filepath))
                return
            fit_distances = loadlib.load_data(filepath)

            elite_fitnesses, elite_distances, seeds = zip(*fit_distances)
            elite_fitnesses = np.array(elite_fitnesses)/78.92820323027551
            p_corr, p_val_p_corr = pearsonr(elite_fitnesses, elite_distances)

            if p_val_p_corr < 0.01:
                p_corrs.append(p_corr)
            count += 1

    print('{}/{} correlations'.format(len(p_corrs), count))
    data_metrics = metrics.metrics_acute(p_corrs, include_data=True,
                                         confidence_median=confidence, scale_factor=1.0)
    return [{'color': color.development, 'offset': 0, 'metrics': data_metrics,
             'exp': (exp_name,), 'confidence': confidence,
             'fitness_function': None}]


def boxplot_graph(data_metrics, x_fun, title='', filename=None, exp_name=None,
                  y_range=None, x_range=None, x_ticker=None, x_labels=None,
                  boxwidth=100, width=700, height=200, baseline=None, trendline=True,
                  color_trendlines=False, open_output=False, verbose=False, save=True,
                  font_size=8, open_folder=False, redraw=True, autoyticker=None,
                  outliers_size=1, scatter_only=False, rotation_scatter=False, g=None,
                  notches=True, outlier_shape='diamond', **kwargs):
    if len([data_i for data_i in data_metrics if data_i['metrics'] is not None]) == 0:
        utils.print_warning('no data for graph {}, skipping generation'.format(title))
        return

    ## figure out filename and title
    if filename is None:
        filename = title.replace(' ', '_')
    fig_filepath = Path(paths.FIGURES_DIR, '{}/boxplot/{}.pdf'.format(exp_name, filename))
    if not redraw and fig_filepath.exists():
        utils.print_warning('{} exists, use --redraw to force regeneration'.format(fig_filepath))
        return

    confidences = [data_i['confidence'] for data_i in data_metrics]
    assert len(set(confidences)) == 1
    if confidences[0] is not None:
        title = '{} (confidence {}%)'.format(title, 100*confidences[0])

    ## compute x locations
    for data_i in data_metrics:
        try:
            data_i['x'] = x_fun(None)
        except Exception:
            cfg = loadlib.load_cfg(*data_i['exp'])
            data_i['x'] = x_fun(cfg)

    ## create graph and setup x-axis
    if g is None:
        g = graph.Graph(title=title, x_range=x_range, y_range=y_range, tools=None, width=width, height=height)
        if x_ticker is None:
            x_ticker = sorted(set([data_i['x'] for data_i in data_metrics]))
            if x_labels is None:
                g.fig.xaxis.major_label_text_font_size = '0pt'
            else:
                g.fig.xaxis.major_label_text_font_size = '{}pt'.format(font_size)
        else:
            g.fig.xaxis.major_label_standoff = 10

        x_ticker = [int(x_i) if x_i == int(x_i) else x_i for x_i in x_ticker]  # bug with ints being different than floats
        g.fig.xaxis.ticker = x_ticker
        if x_range is None:
            x_range = (min(x_ticker) - 0.5, max(x_ticker) + 0.5)
            g.fig.x_range.start = x_range[0]
            g.fig.x_range.end   = x_range[1]
        if x_labels is not None:
            assert len(set(x_ticker)) == len(x_labels), 'number of tickers ({}) does not match the number of labels ({})'.format(
                len(set(x_ticker)), len(x_labels))
            g.fig.xaxis.major_label_overrides = {tick: label for tick, label in zip(x_ticker, x_labels)}
        g.fig.yaxis.major_label_text_font_size = '{}pt'.format(font_size)

    ## check that all data is displayed
    if y_range is not None:
        min_y = min(data_i['metrics']['min'] for data_i in data_metrics if data_i['metrics'] is not None)
        if min_y < y_range[0]:
            utils.print_warning('{}: minimum y value ({}) below minimum displayed value ({})'.format(filename, min_y, y_range[0]))
        max_y = max(data_i['metrics']['max'] for data_i in data_metrics if data_i['metrics'] is not None)
        if max_y > y_range[1]:
            utils.print_warning('{}: maximum y value ({}) above maximum displayed value ({})'.format(filename, max_y, y_range[1]))
    if x_range is not None:
        min_x = min(data_i['x'] for data_i in data_metrics if data_i['metrics'] is not None) - boxwidth/2
        if min_x < x_range[0]:
            utils.print_warning('{}: minimum x value ({}) below minimum displayed value ({})'.format(filename, min_x, x_range[0]))
        max_x = max(data_i['x'] for data_i in data_metrics if data_i['metrics'] is not None) + boxwidth/2
        if max_x > x_range[1]:
            utils.print_warning('{}: maximum x value ({}) above maximum displayed value ({})'.format(filename, max_x, x_range[1]))


    ## plot data
    if not scatter_only:
        if trendline:
            medians = [(data_i['x'] + data_i['offset'], data_i['metrics']['q0.5']) for data_i in data_metrics
                                                                             if data_i['metrics'] is not None]
            m_xs, m_ys = zip(*sorted(medians))
            g.fig.line(m_xs, m_ys, line_color='grey', line_alpha=0.5, line_width=0.5, line_dash=[2,2])

        if color_trendlines:
            colors = set([data_i['color'] for data_i in data_metrics])
            for color in colors:
                medians = [(data_i['x'] + data_i['offset'], data_i['metrics']['q0.5']) for data_i in data_metrics
                           if color == data_i['color'] and data_i['metrics'] is not None]
                m_xs, m_ys = zip(*sorted(medians))
                g.fig.line(m_xs, m_ys, line_color=color[0], line_alpha=0.5, line_width=0.5, line_dash=[2,2])

        if baseline is not None:
            for data_i in data_metrics :
                if data_i['metrics'] is not None and baseline == data_i['exp']:
                    _baselines(g.fig, data_i, x_range=x_range)

        if autoyticker is not None:
            for data_i in data_metrics:
                if data_i['metrics'] is not None and autoyticker == data_i['exp']:
                    m = data_i['metrics']
                    g.fig.yaxis.ticker = [y_range[0], m['min_iqr'], m['q0.25'], m['q0.5'],
                                          m['q0.75'], m['max_iqr'], y_range[1]]

    for data_i in data_metrics:
        if data_i['metrics'] is not None:
            if scatter_only or (len(data_i['metrics']['data']) < 8):
                x = data_i['x'] + data_i['offset']
                line_color, dark_color, light_color = data_i['color']
                if rotation_scatter:
                    seeds = data_i['seeds']
                    exp_name, major, minor = data_i['exp']
                    cfg = loadlib.load_cfg(exp_name, major, minor)
                    filepath = loadlib.angle_filepath_from_cfg(cfg)
                    data_angles = utils.load_data(filepath)
                    for seed in seeds:
                        assert len(data_angles[seed]) == 1
                    angles = [abs(data_angles[seed][0])/2/math.pi for seed in seeds]
                    #print(angles)
                    data_rot   = [d for d, a in zip(data_i['metrics']['data'], angles) if a >  3]
                    data_norot = [d for d, a in zip(data_i['metrics']['data'], angles) if a <= 3]
                    g.fig.scatter([random.gauss(x-boxwidth/6, boxwidth/20) for _ in data_rot],
                                  data_rot, marker='circle',
                                  size=4*outliers_size, line_color=dark_color, fill_color=light_color,
                                  line_alpha=0.5, fill_alpha=0.5, line_width=0.5)
                    g.fig.scatter([random.gauss(x+boxwidth/6, boxwidth/20) for _ in data_norot],
                                  data_norot, marker='triangle',
                                  size=4*outliers_size, line_color=dark_color, fill_color=light_color,
                                  line_alpha=0.5, fill_alpha=0.5, line_width=0.5)

                else:
                    g.fig.scatter([random.gauss(x, boxwidth/16) for _ in data_i['metrics']['data']],
                                  data_i['metrics']['data'], marker='circle',
                                  size=4*outliers_size, line_color=dark_color, fill_color=light_color,
                                  line_alpha=0.5, fill_alpha=0.5, line_width=0.5)
            else:
                _box_plot(g.fig, data_i, boxwidth, outliers_size=outliers_size, notches=notches,
                          outlier_shape=outlier_shape)
                if verbose:
                    print(('boxplot x: {x} min_iqr: {min_iqr:.2f}, q25: {q0.25:.2f}, q50: {q0.5:.2f}, '
                           ' q75: {q0.75:.2f}, max_iqr: {max_iqr:.2f}').format(**data_i['metrics']))
                    if len(metrics['outliers']) > 0:
                        print('         outliers: ', data_i['metrics']['outliers'])

    ## save fig
    if save:
        fig_filepath.parent.mkdir(parents=True, exist_ok=True)
        graph.save_fig(g.fig, str(fig_filepath), scale_factor=2.0)

        if sys.platform == 'darwin':
            if open_output:
                subprocess.check_output(['open', fig_filepath])
            if open_folder:
                subprocess.check_output(['open', fig_filepath.parent])

    return g, filename


def _box_plot(fig, data_i, boxwidth, outliers_size=1, notches=True, outlier_shape='diamond'):
    line_color, dark_color, light_color = data_i['color']
    line_width = 0.5
    size = line_width * 7
    w = boxwidth / 2
    m = data_i['metrics']
    x = data_i['x'] + data_i['offset']

    # box
    if notches and 'median_ci_low' in m:
        w_median = 0.5*w
        ci_low, ci_high = m['median_ci_low'], m['median_ci_high']
        p_xs, p_ys = [x-0.5*w], [m['q0.5']]

        # if ci_low < m['q0.25']:
        #     p_xs.extend([x-w, x-w, x+w, x+w])
        #     p_ys.extend([ci_low, m['q0.25'], m['q0.25'], ci_low])
        # else:
        p_xs.extend([x-w, x-w, x+w, x+w])
        p_ys.extend([ci_low, m['q0.25'], m['q0.25'], ci_low])

        p_xs.append(x + 0.5*w)
        p_ys.append(m['q0.5'])

        p_xs.extend([x+w, x+w, x-w, x-w])
        p_ys.extend([ci_high, m['q0.75'], m['q0.75'], ci_high])
        # if ci_high > m['q0.75']:
        #     p_xs.extend([x+w, x+w, x-w, x-w])
        #     p_ys.extend([ci_high, m['q0.75'], m['q0.75'], ci_high])
        # else:
    else:
        w_median = w
        p_xs = [x-w, x-w, x+w, x+w]
        p_ys = [m['q0.25'], m['q0.75'], m['q0.75'], m['q0.25']]

    # stems
    fig.segment(x, m['max_iqr'], x, m['q0.75'], line_color=light_color, line_width=line_width)
    fig.segment(x, m['min_iqr'], x, m['q0.25'], line_color=light_color, line_width=line_width)

    # whiskers
    fig.segment(x - 0.2*boxwidth, m['min_iqr'], x + 0.2*boxwidth, m['min_iqr'],
                line_color=light_color, line_width=line_width)
    fig.segment(x - 0.2*boxwidth, m['max_iqr'], x + 0.2*boxwidth, m['max_iqr'],
                line_color=light_color, line_width=line_width)

    # outliers
    fig.scatter([random.gauss(x, boxwidth/16) for _ in m['outliers']], m['outliers'], marker=outlier_shape,
                size=4*outliers_size, line_color=dark_color, fill_color=light_color,
                line_alpha=0.5, fill_alpha=0.5, line_width=line_width)

    # notched (or not) box
    fig.patch(p_xs, p_ys, line_color=None, fill_color=light_color)

    # median line
    fig.segment(x - w_median, m['q0.5'], x + w_median, m['q0.5'],
                line_color=dark_color, line_width=2*line_width)


def _baselines(fig, data_i, x_range=None):
    line_color, dark_color, light_color = data_i['color']
    line_width = 0.5
    x_min, x_max = -100, 100
    if x_range is not None:
        x_min, x_max = x_range

    m = data_i['metrics']
    if 'median_ci_low' in m:
        square_x = [x_min, x_max, x_max, x_min]
        square_y = [m['median_ci_low'], m['median_ci_low'], m['median_ci_high'], m['median_ci_high']]
        fig.patch(square_x, square_y, line_color=None, fill_color=light_color, fill_alpha=0.25)

    fig.segment(x_min, m['q0.25'], x_max, m['q0.25'], line_color=line_color, line_width=line_width, line_dash="1 1")
    fig.segment(x_min, m['q0.5'],  x_max, m['q0.5'],  line_color=line_color, line_width=line_width, line_dash="1 1")
    fig.segment(x_min, m['q0.75'], x_max, m['q0.75'], line_color=line_color, line_width=line_width, line_dash="1 1")
