import shutil
import pickle
from pathlib import Path

import numpy as np
import scipy.stats
from scipy.stats import pearsonr, spearmanr
from bokeh.models import ColumnDataSource, LabelSet
import pandas as pd

from ... import paths, utils
from .. import loadlib
from . import graph
from . import color


def correlation_graph(params, major, minor, pool=None):

    exp_name, kind, window = params['exp_name'], params['kind'], params.get('window', None)

    if kind == 'genealogy':
        max_x = 4.0
        filepath = loadlib.genealogy_distance_correlation_filepath(exp_name, major, minor, window)
    elif kind == 'genotype':
        max_x = 0.2
        filepath = loadlib.genotype_distance_correlation_filepath(exp_name, major, minor, window)
    else:
        raise ValueError('kind {} not known'.format(kind))

    post_filepath = Path(paths.POST_DIR, filepath)
    fig_filepath  = Path(paths.FIGURES_DIR, filepath).with_suffix('.pdf')
    if not params['redraw'] and fig_filepath.is_file():
        return

    if not post_filepath.is_file():
        utils.print_warning('correlation_graph: {} not found, skipping'.format(post_filepath))
        return
    with open(post_filepath, 'rb') as fd:
        fit_distances = pickle.load(fd)

    elite_fitnesses, elite_distances, seeds = zip(*fit_distances)
    elite_fitnesses = np.array(elite_fitnesses)/1000

    p_corr, _ = pearsonr(elite_fitnesses, elite_distances)
    s_corr, _ = spearmanr(elite_fitnesses, elite_distances)

    window_str = '' if window is None else ' [{}-{}]'.format(window[0], window[1])
    title = "{}_distance/fitness correlation {}:v{}.{}{} (Pearson's {:.3f}, Spearman's {:.3f})".format(
            kind, exp_name, major, minor, window_str, p_corr, s_corr)

    if max(elite_fitnesses) > y_max:
        utils.print_warning('v{}.{}: max y value {} exceeds max displayed y value {}'.format(major,
                            minor, utils.red('{:.2f}'.format(max(elite_fitnesses))), y_max))
    if max(elite_distances) > x_max:
        utils.print_warning('v{}.{}: max x value {} exceeds max displayed x value {}'.format(major,
                            minor, utils.red('{:.2f}'.format(max(elite_distances))), x_max))

    g = graph.Graph(title=title, width=500, height=500, tools=None, y_range=(0, 16.0)) #, x_range=(0, max_x))
    g.fig.scatter(elite_distances, elite_fitnesses,
                  marker='circle', size=6.0, line_color=None, line_width=0.5, alpha=0.5)
    g.fig.x_range.start = 0
    g.fig.y_range.start = 0
    # title = '{}/{}/genealogy v{}.{}.png'.format(paths.FIGURES_DIR, exp_name, major, minor)
    fig_filepath.parent.mkdir(parents=True, exist_ok=True)
    graph.save_fig(g.fig, fig_filepath, scale_factor=2.0, verbose=True)


def correlation_double_graph(params, major_a, minor_a, major_b, minor_b, pool=None):

    exp_name, kind, window = params['exp_name'], params['kind'], params.get('window', None)
    color_names, seed_labels = params['color_names'], params.get('seed_labels', False)
    copy_folder = params.get('copy_folder', None)
    y_max = params.get('y_max', 12.0)

    single_path = {'genealogy': loadlib.genealogy_distance_correlation_filepath,
                   'genotype' : loadlib.genotype_distance_correlation_filepath}[kind]
    double_path = {'genealogy': loadlib.genealogy_distance_correlation_double_filepath,
                   'genotype' : loadlib.genotype_distance_correlation_double_filepath}[kind]

    fig_filepath = double_path(exp_name, major_a, minor_a, major_b, minor_b, window=window).with_suffix('.pdf')
    fig_abs_filepath = Path(paths.FIGURES_DIR, fig_filepath)
    if not params['redraw'] and fig_abs_filepath.is_file():
        return

    # x_max = 4.5 if kind == 'genealogy' else 0.2
    g = graph.Graph(width=500, height=500, tools=None, y_range=(0, y_max)) # , x_range=(0, x_max))
    # g.fig.x_range.start = 0
    # g.fig.y_range.start = 0

    p_corrs, s_corrs = [], []
    for (major, minor), color_name in zip([(major_a, minor_a), (major_b, minor_b)], color_names):
        filepath = Path(paths.POST_DIR, single_path(exp_name, major, minor, window))
        if not filepath.is_file():
            utils.print_warning('correlation_graph: {} not found, skipping'.format(filepath))
            return
        with open(filepath, 'rb') as fd:
            fit_distances = pickle.load(fd)

        elite_fitnesses, elite_distances, seeds = zip(*fit_distances)
        elite_fitnesses = np.array(elite_fitnesses)/78.92820323027551
        if max(elite_fitnesses) > y_max:
            utils.print_warning('v{}.{}: max y value {} exceeds max displayed y value {}'.format(major,
                                minor, utils.red('{:.2f}'.format(max(elite_fitnesses))), y_max))
        # if max(elite_distances) > x_max:
        #     utils.print_warning('v{}.{}: max x value {} exceeds max displayed x value {}'.format(major,
        #                         minor, utils.red('{:.2f}'.format(max(elite_distances))), x_max))

        p_corr, p_val_p_corr = pearsonr( elite_fitnesses, elite_distances)
        s_corr, p_val_s_corr = spearmanr(elite_fitnesses, elite_distances)
        p_corrs.append(p_corr)
        s_corrs.append(s_corr)

        stars = 'n.s.'
        if p_val_p_corr < 0.05:  stars = '*'
        if p_val_p_corr < 0.01:  stars = '**'
        if p_val_p_corr < 0.001: stars = '***'
        if major != 1:
            print('v{}.{}: Pearson  : {:.3f} (p-value {:.3e}) {}'.format(major, minor, p_corr, p_val_p_corr, stars))
            # print('Spearman : {:.3f} (p-value {:.3e})'.format(s_corr, p_val_s_corr))


        # p.xaxis[0].axis_label = '{} distance'.format(kind)
        # p.yaxis[0].axis_label = 'fitness'

        slope, intercept = scipy.stats.siegelslopes(elite_fitnesses, elite_distances)
        # print(slope, intercept)
        slope, intercept, _, _ = scipy.stats.theilslopes(elite_fitnesses, elite_distances)
        # print(slope, intercept)
        # slope, intercept, rvalue, pvalue, stderr = scipy.stats.linregress(elite_distances, elite_fitnesses)
        # print(slope, intercept)

        line_color, dark_color, light_color = getattr(color, color_name)


        marker = 'square' if color_name == 'control' else 'triangle'
        size   = 11 if color_name == 'control' else 13

        g.fig.scatter(elite_distances, elite_fitnesses, fill_color=dark_color,
                      marker=marker, size=size, line_color=None, line_width=0.5, alpha=0.5)

        # g.fig.line(elite_distances, [slope * x + intercept for x in elite_distances],
        #            line_color=line_color, line_width=2.0, alpha=0.5)

        if seed_labels:
            source = ColumnDataSource(data={'fitnesses': elite_fitnesses,
                                            'distances': elite_distances,
                                            'seeds': [str(s) for s in seeds]})
            labels = LabelSet(x='distances', y='fitnesses', text='seeds', level='glyph',
                              x_offset=0, y_offset=0, source=source, render_mode='canvas', text_font_size='4pt')
            g.fig.add_layout(labels)

        # creating a csv version of the data
        # csv_filepath = str(filepath) + '.csv'
        # data = {'fitness': elite_fitnesses, 'distance': elite_distances}
        # df = pd.DataFrame(data=data)
        # df.to_csv(csv_filepath)
        # utils.print_info('saved {}'.format(csv_filepath))

    window_str = '' if window is None else ' [{}-{}]'.format(window[0], window[1])
    title = "{}_distance/fitness correlation {}:v{}.{}/v{}.{}{} (Pearson's {:.3f}/{:.3f}, Spearman's {:.3f}/{:.3f})".format(
            kind, exp_name, major_a, minor_a, major_b, minor_b, window_str, p_corrs[0], p_corrs[1],
            s_corrs[0], s_corrs[1])
    g.fig.title.text = title
    g.fig.title.text_font_size = '6pt'

    fig_abs_filepath.parent.mkdir(parents=True, exist_ok=True)
    graph.save_fig(g.fig, fig_abs_filepath, scale_factor=2.0, verbose=True)

    if copy_folder is not None:
        copy_filepath = Path(paths.FIGURES_DIR, exp_name, 'copies', copy_folder, fig_filepath.name)
        copy_filepath.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(fig_abs_filepath, copy_filepath)
        utils.print_info('copy @{}'.format(copy_filepath))
