import multiprocessing as mp
from pathlib import Path

import matplotlib.pyplot as plt
import matplotlib as mpl

from ... import utils
from ... import paths
from ..  import loadlib
from .   import color
from .   import graph



def evoscatter_fig_seeds(params, major_a, minor_a, major_b, minor_b, pool=None):
    exp_name, seeds = params['exp_name'], params['seeds']

    if seeds is None:
        done_a = loadlib.find_done(exp_name, major_a, minor_a)
        done_b = loadlib.find_done(exp_name, major_b, minor_b)
        seeds = set(done_a).intersection(set(done_b))

    y_maxs = []
    for seed in seeds:
        y_maxs.append(evoscatter_fig_seed(params, major_a, minor_a, major_b, minor_b, seed))

    utils.print_info('maximum value over all seeds is {:.1f}'.format(max(y_maxs)))


def evoscatter_fig_seed(params, major_a, minor_a, major_b, minor_b, seed):
    exp_name, backend = params['exp_name'], params['backend']
    x_range, y_range, color_names = params['x_range'], params['y_range'], params['color_names']
    width, height = params.get('fig_size', (800, 300))
    y_ticker = params.get('y_ticker', None)

    def _get_perfs(params, major, minor, seed):
        exp_name = params['exp_name']
        cfg = loadlib.load_cfg(exp_name, major, minor)
        fitness_function = params.get('fitness_function', None)
        fitness_filepath = loadlib.fitness_filepath_from_cfg(cfg, seed, fitness_function=fitness_function)
        if fitness_filepath.is_file():
            perfs = loadlib.load_perffile(cfg, seed, fitness_function=fitness_function)  # makes y-axis labels shorter
        else:
            utils.print_error('{} not found; skipping'.format(fitness_filepath))
            return None
        return perfs

    def splice(perfs):
        x_perfs, y_perfs = [], []
        for t, gen_perfs in enumerate(perfs):
            for p in gen_perfs:
                x_perfs.append(t)
                y_perfs.append(p)
        return x_perfs, y_perfs

    name = 'seed{}'.format(seed)
    version = 'v{}_v{}'.format(loadlib.version_str(major_a, minor_a),
                               loadlib.version_str(major_b, minor_b))
    title = 'evoscatter {} {}'.format(version, name)

    fig_filepath = Path(paths.FIGURES_DIR, exp_name, 'evoscatter', version, name)
    fig_filepath.parent.mkdir(parents=True, exist_ok=True)
    file_filepath = str(fig_filepath) + '.png'

    y_max = float('-inf')

    if backend == 'matplotlib':
        plt.rcParams.update({'font.size': 6})
        plt.rcParams['axes.linewidth'] = 0.4

        fig, ax = plt.subplots(figsize=(12, 4), dpi=150)
        ax.tick_params(width=0.4)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.grid(False)
        ax.set_title(title)
        plt.subplots_adjust(left=0.05, right=0.95, top=0.95, bottom=0.05)
        if x_range is not None:
            plt.xlim(*x_range)
        if y_range is not None:
            plt.ylim(*y_range)

        assert len(color_names) == 2
        # b is usually less, better put in front.
        for (major, minor), color_name in zip([(major_b, minor_b), (major_a, minor_a)], reversed(color_names)):
            color_ = getattr(color, color_name)
            x_perfs, y_perfs = _get_perfs(params, major, minor, seed)
            y_max = max(y_max, max(y_perfs))
            ax.scatter(x_perfs, y_perfs, c=color_[2], marker="o", s=2.0, alpha=0.125, edgecolors='white', linewidth=0)

        plt.savefig(file_filepath, bbox_inches='tight', pad_inches=0.05)
        plt.close(fig)
        utils.print_info('saved {}'.format(file_filepath))

    elif backend == 'bokeh':
        g = graph.Graph(title=title, y_range=y_range, x_range=x_range, width=width, height=height,
                        tools=None, y_axis_location='right')
        # g.fig.yaxis.ticker = [0, 2500, 5000]
        g.fig.yaxis[0].ticker.desired_num_ticks = 3
        if y_ticker is not None:
            g.fig.yaxis.ticker = y_ticker

        g.fig.xaxis.ticker = [0, 2000, 4000]
        g.fig.axis.major_label_text_font_size = '50pt'
        g.fig.axis.axis_line_width = 3
        g.fig.axis.major_tick_line_width = 3
        g.fig.axis.major_tick_out = 15
        g.fig.axis.major_label_standoff = 20
        g.fig.min_border_right = 100

        y_ticker = [y_range[0], y_range[1]]
        for (major, minor), color_name in zip([(major_b, minor_b), (major_a, minor_a)], reversed(color_names)):
            color_ = getattr(color, color_name)
            perfs = _get_perfs(params, major, minor, seed)
            x_perfs, y_perfs = splice(perfs)

            y_max = max(y_max, max(y_perfs))
            g.fig.scatter(x_perfs, y_perfs, alpha=0.125, size=2.0, color=color_)
            y_ticker.append(max(perfs[-1]))

        if y_range is not None and y_range[1] < y_max:
            g.fig.y_range.end = int(y_max) + 2
            y_ticker[1] = int(y_max) + 2

        y_ticker = sorted([float(y_i) for y_i in y_ticker])
        g.fig.yaxis.ticker = y_ticker
        g.fig.yaxis.major_label_overrides = {y_i: '{:.2f}'.format(y_i) for y_i in y_ticker}


        graph.save_fig(g.fig, file_filepath, scale_factor=1.0, verbose=True)

    else:
        raise ValueError('unknow backend "{}"'.format(backend))

    if y_range is not None and y_range[1] < y_max:
        utils.print_warning('max value {:.1f} exceeded set display boundary {:.1f}, autoresized to {}'.format(
                            y_max, y_range[1], int(y_max) + 2))

    return y_max
