import shutil
from pathlib import Path

from scipy.signal import savgol_filter
import numpy as np

from ... import evo, utils, paths
from .. import loadlib, metrics
from . import graph, color


def genealogy_distance_best_seeds_graphs(params, major, minor, pool=None):
    exp_name = params['exp_name']
    seeds = loadlib.find_done(exp_name, major, minor)

    ref_seeds = params.get('seeds', None)
    if ref_seeds is not None:
        ref_seeds = set(range(ref_seeds[0], ref_seeds[1]+1))
        seeds = sorted(ref_seeds.intersection(seeds))

    if pool is None:
        for seed in seeds:
            genealogy_distance_best_seed_graph(params, major, minor, seed)
    else:
        pool.starmap(genealogy_distance_best_seed_graph, [(params, major, minor, seed)
                                                          for seed in done])

def genealogy_distance_best_seed_graph(params, major, minor, seed, pool=None):
    exp_name = params['exp_name']
    frozen_run = loadlib.load_results(exp_name, major, minor, seed, verbose=False)

    best = evo.analysis.select_best(frozen_run.population, n=1, window=(-50, None))[0]
    genealogy = evo.analysis.genealogy(frozen_run.population, best)
    genealogy = [frozen_run.population.repository.member(uid) for uid in genealogy]

    ds = [evo.analysis.genotype_distance(genealogy[i], genealogy[i+1]) for i in range(len(genealogy) - 2, -1, -1)]

    window, poly = 101, 1
    ds_smooth = savgol_filter(ds, window, poly, mode='nearest')

    filename = 'speeds_v{}.{}_seed{}_comp.pdf'.format(major, minor, seed)
    fig_filepath = Path(paths.FIGURES_DIR, exp_name, 'speeds_genealogy', filename)

    g = graph.Graph(title=filename, width=600, height=300, tools=None, y_range=[0, 0.0025])
    g.fig.line(range(len(ds_smooth)), ds_smooth)
    g.fig.x_range.start = 0
    #g.fig.y_range.start = 0
    fig_filepath.parent.mkdir(parents=True, exist_ok=True)
    graph.save_fig(g.fig, fig_filepath, scale_factor=2.0, verbose=True)


def genealogy_distance_best_graph(params, major, minor, pool=None):
    exp_name, rolling_window = params['exp_name'], params['rolling_window']
    color_name = params['color_name']

    data_filepath = loadlib.genealogy_distance_best_filepath(exp_name, major, minor)
    distances = loadlib.load_data(data_filepath)

    ma_distances = []
    for d in distances:
        ma_distances.append(savgol_filter(d, rolling_window, 1, mode='nearest'))

    ma_distances = np.array(ma_distances).T  # now first index is timesteps
    distance_metrics = metrics.metrics_longitudinal(ma_distances, confidence_mean=0.95)
    # means = np.mean(ma_distances, axis=0)

    fig_filepath = Path(paths.FIGURES_DIR, loadlib.prefix_path(exp_name, major, minor), 'genealogy_distance_best.png')
    title = 'genealogy distance v{}.{} ({} seeds, window = {})'.format(major, minor, len(distances), rolling_window)

    mean, high, low = distance_metrics['mean'], distance_metrics['mean_ci_high'], distance_metrics['mean_ci_low']
    x = range(len(mean))
    c = getattr(color, color_name)

    g = graph.Graph(title=title, width=600, height=300, tools=None, y_range=[0, 0.003])
    g.patch(x, high, low, line_color=None, fill_color=c[2], fill_alpha=0.25)
    g.fig.line(x, mean, line_color=c[0], line_width=2)
    g.fig.x_range.start = 0

    fig_filepath.parent.mkdir(parents=True, exist_ok=True)
    graph.save_fig(g.fig, fig_filepath, scale_factor=2.0, verbose=True)

    copy_folder = params.get('copy_folder', None)
    if copy_folder is not None:
        copy_filepath = Path(paths.FIGURES_DIR, exp_name, 'copies', copy_folder, 'gen_dist_best_v{}.{}.png'.format(major, minor))
        copy_filepath.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(fig_filepath, copy_filepath)
        utils.print_info('copy @{}'.format(copy_filepath))


def genealogy_distance_best_double_graph(params, major_a, minor_a, major_b, minor_b, pool=None):
    exp_name, leap, rolling_window = params['exp_name'], params['leap'], params['rolling_window']
    y_range, color_names = params['y_range'], params['color_names']
    x_offsets = params.get('x_offsets', (0, 0))

    try:
        title = 'genealogy distance v{}.{}-v{}.{} (leap = {}, window = {})'.format(
                 major_a, minor_a, major_b, minor_b, leap, rolling_window)
        g = graph.Graph(title=title, width=600, height=200, tools=None, y_range=y_range)
        g.fig.x_range.start = 0
        offset_str = '' if x_offsets[0] == x_offsets[1] == 0 else '_x{}_{}'.format(*x_offsets)
        filename = 'genealogy_distance_best_v{}.{}-v{}.{}_leap{}_w{}{}.png'.format(
                    major_a, minor_a, major_b, minor_b, leap, rolling_window, offset_str)
        fig_filepath = Path(paths.FIGURES_DIR, exp_name, 'genealogy_distance_best', filename)

        for i, (major, minor) in enumerate([(major_a, minor_a), (major_b, minor_b)]):
            data_filepath = loadlib.genealogy_distance_best_filepath(exp_name, major, minor, leap)
            distances = loadlib.load_data(data_filepath)

            w_half = rolling_window // 2
            ma_distances = []
            for d in distances:
                ma_distances.append(savgol_filter(d, rolling_window, 1, mode='nearest')[w_half:-w_half])

            ma_distances = np.array(ma_distances).T  # now first index is timesteps
            distance_metrics = metrics.metrics_longitudinal(ma_distances, confidence_mean=params['confidence'])
            # means = np.mean(ma_distances, axis=0)

            mean, high, low = distance_metrics['mean'], distance_metrics['mean_ci_high'], distance_metrics['mean_ci_low']
            x_off = x_offsets[i]
            x = range(w_half + x_off, min(4000, x_off + w_half + len(mean)))
            c = getattr(color, color_names[i])

            mean, high, low = mean[:len(x)], high[:len(x)], low[:len(x)]
            y_max = max(max(mean), max(high), max(low))
            if y_max > y_range[1]:
                utils.print_warning('y_max {} superior to displayed values ({})'.format(y_max, y_range[1]))

            g.patch(x, high, low, line_color=None, fill_color=c[2], fill_alpha=0.35)
            g.fig.line(x, mean, line_color=c[0], line_width=2)

        fig_filepath.parent.mkdir(parents=True, exist_ok=True)
        graph.save_fig(g.fig, fig_filepath, scale_factor=2.0, verbose=True)

        copy_folder = params.get('copy_folder', None)
        if copy_folder is not None:
            copy_filepath = Path(paths.FIGURES_DIR, exp_name, 'copies', copy_folder, filename)
            copy_filepath.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(fig_filepath, copy_filepath)
            utils.print_info('copy @{}'.format(copy_filepath))
    except FileNotFoundError:
        utils.print_warning('{}:{}.{}-{}.{} missing data; skipping.'.format(exp_name,
                             major_a, minor_a, major_b, minor_b))


def parent_stability_double_graph(params, major_a, minor_a, major_b, minor_b, pool=None):
    exp_name, rolling_window = params['exp_name'], params['rolling_window']
    y_range, color_names = params['y_range'], params['color_names']
    x_offsets = params.get('x_offsets', (0, 0))

    try:
        title = 'v{}.{}-v{}.{} (window = {})'.format(
                 major_a, minor_a, major_b, minor_b, rolling_window)
        g = graph.Graph(title=title, width=600, height=300, tools=None, y_range=y_range)
        g.fig.x_range.start = 0
        g.fig.title.text_font_size = '40pt'
        offset_str = '' if x_offsets[0] == x_offsets[1] == 0 else '_x{}_{}'.format(*x_offsets)
        filename = 'parent_stability_v{}.{}-v{}.{}_w{}{}.png'.format(
                    major_a, minor_a, major_b, minor_b, rolling_window, offset_str)
        fig_filepath = Path(paths.FIGURES_DIR, exp_name, 'parent_stability', filename)

        for i, (major, minor) in enumerate([(major_a, minor_a), (major_b, minor_b)]):
            data_filepath = loadlib.parent_stability_filepath(exp_name, major, minor)
            distances = loadlib.load_data(data_filepath)

            w_half = rolling_window // 2
            ma_distances = []
            for d in distances:
                ma_distances.append(savgol_filter(d, rolling_window, 1, mode='nearest')[w_half:-w_half])

            ma_distances = np.array(ma_distances).T  # now first index is timesteps
            distance_metrics = metrics.metrics_longitudinal(ma_distances, confidence_mean=params['confidence'])
            # means = np.mean(ma_distances, axis=0)

            mean, high, low = distance_metrics['mean'], distance_metrics['mean_ci_high'], distance_metrics['mean_ci_low']
            x_off = x_offsets[i]
            x = range(w_half + x_off, min(4000, x_off + w_half + len(mean)))
            c = getattr(color, color_names[i])

            mean, high, low = mean[:len(x)], high[:len(x)], low[:len(x)]
            y_max = max(max(mean), max(high), max(low))
            if y_max > y_range[1]:
                utils.print_warning('y_max {} superior to displayed values ({})'.format(y_max, y_range[1]))

            g.patch(x, high, low, line_color=None, fill_color=c[2], fill_alpha=0.35)
            g.fig.line(x, mean, line_color=c[0], line_width=2)

        fig_filepath.parent.mkdir(parents=True, exist_ok=True)
        graph.save_fig(g.fig, fig_filepath, scale_factor=2.0, verbose=True)

        copy_folder = params.get('copy_folder', None)
        if copy_folder is not None:
            copy_filepath = Path(paths.FIGURES_DIR, exp_name, 'copies', copy_folder, filename)
            copy_filepath.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(fig_filepath, copy_filepath)
            utils.print_info('copy @{}'.format(copy_filepath))
    except FileNotFoundError:
        utils.print_warning('{}:{}.{}-{}.{} missing data; skipping.'.format(exp_name,
                             major_a, minor_a, major_b, minor_b))
