import os
import warnings

import numpy as np

import bokeh as bk

import bokeh.plotting as bpl
from bokeh.models import FixedTicker, FuncTickFormatter
from bokeh.models import LinearColorMapper, BasicTicker, ColorBar, Label
from bokeh.models import HoverTool

from . import utils_bokeh
from .utils_bokeh import unroll, save_fig



TOOLS = ['save', 'box_zoom', 'reset']


class Graph:

    def __init__(self, x_range=None, y_range=None, title='', height=250, width=900,
                       tools=TOOLS, border=False, **kwargs):
        # bk.plotting.output_notebook(hide_banner=True)
        self._tools = () if tools is None else tuple(tools)
        self._hover_tool = HoverTool(tooltips=[("name", "$name"),
                                               ("(x,y)", "(@x{%d}, @y{%0.1f})")],
                                     formatters={'x' : 'printf', 'y': 'printf'})

        self._handle = None
        self.lines = []
        self.title = title
        self.N = 0

        self.fig = utils_bokeh.figure(title=self.title, plot_width=width, plot_height=height,
                                      x_range=x_range, y_range=y_range,
                                      toolbar_sticky=False, tools=self.tools, **kwargs)
        if len(self._tools) == 0:
            #p.toolbar.logo = None
            self.fig.toolbar_location = None

        self._glyph_map = {self.fig.circle:  bk.models.markers.Circle,
                           self.fig.line:    bk.models.glyphs.Line}
        if not border:
            self.fig.outline_line_color = None

    @property
    def tools(self):
        return self._tools + (self._hover_tool,)

    def set_x_ticks(self, ticks):
        """Manually set the ticks displayed on the x-axis."""
        self.fig.xaxis[0].ticker = FixedTicker(ticks=ticks)

    def set_y_ticks(self, ticks):
        """Manually set the ticks displayed on the y-axis."""
        self.fig.yaxis[0].ticker = FixedTicker(ticks=ticks)

    def save(self, filepath:str=None, verbose:bool=True):
        """Save the graph

        :param filepath:  'path/to/filepath.pdf' or None. The extension can be '.pdf', '.png', or
                          '.svg'. If None, a pdf filepath will be automatically generated.
        :param verbose:   If True, displays a confirmation dialog after the operation is completed.
        """
        if filepath is None:
            filename = self.title if len(title) > 0 else 'figure'
            filepath = '{}.pdf'.format(filename)
        utils_bokeh.save_fig(self.fig, filepath, verbose=verbose)

    def show(self):
        """Displays the graph"""
        bpl.show(self.fig)

    def patch(self, x, high, low, **kwargs):
        patch_x = list(x) + list(reversed(x))
        patch_y = list(low) + list(reversed(high))
        self.fig.patch(patch_x, patch_y, **kwargs)
