import os
import random
import multiprocessing as mp
import colorsys

import matplotlib as mpl
import matplotlib.cm
import numpy as np
from bokeh.colors import HSL, RGB

from ... import utils
from ... import paths
from ... import evo
from ..  import loadlib
from .   import color
from .   import graph


def color_map(values):
    colors = []
    for r, g, b, _ in 255*mpl.cm.viridis(mpl.colors.Normalize()(values)):
        colors.append("#{:02x}{:02x}{:02x}".format(int(r), int(g), int(b)))
    return colors


def fig_aged_perfs(versions, seed, figname,
                   force=True, t_max=4000, width=1200, height=200, y_range=(0, 10000)):
    suffix = '_distance.png'
    fig_filepath = os.path.join(paths.FIGURES_DIR, figname + suffix)
    os.makedirs(os.path.dirname(fig_filepath), exist_ok=True)

    # check if the figure is already computed
    if ((not force) and os.path.isfile(fig_filepath) and
        os.path.getmtime(fig_filepath) >= loadlib.last_change(filename)):
        #print('{} is up-to-date; skipping.'.format(fig_filepath))
        print(fig_filepath, os.path.isfile(fig_filepath))
        return

    # y_range = y_range if dist_fitness else (y_range[0], y_range[1]/800)
    graph = graph.Graph(title=figname, y_range=y_range, x_range=(-50, t_max+50),
                         width=width, height=height, tools=None)
    graph.fig.xaxis.ticker = [0, 400, 2000, 4000]

    for exp_name, major, minor, scatter_color in versions:
        cfg = loadlib.load_cfg(exp_name, major, minor)
        perfs = loadlib.load_perffile(cfg, seed)

        x_perfs, y_perfs, age_perfs, alphas = [], [], [], []
        for t, gen_perfs in enumerate(perfs):
            if t < t_max:
                for p in gen_perfs:
                    # if new_member:
                    x_perfs.append(t)
                    y_perfs.append(p)
                    # age_perfs.append(t_root)
                    alphas.append(0.125) #  if new_member else 0.125)
        #colors = color_map(age_perfs)
        graph.fig.scatter(x_perfs, y_perfs, alpha=alphas, size=2.0, color=scatter_color[2])

    return graph.save_fig(graph.fig, fig_filepath, scale_factor=2.0, verbose=True)


def evoscatter_fig(versions, dist_fitness=False, seeds=None, width=1200, height=400,
                   y_range=(0, 6000), open_output=False):
    exp_name, major, minor, _ = versions[0]
    cfg = loadlib.load_cfg(exp_name, major, minor)

    if seeds is None:
        seeds = loadlib.find_done(exp_name, major, minor, cfg=cfg)

    for seed in seeds:
        #if seed in range(cfg['meta']['slurm_array'][0], cfg['meta']['slurm_array'][1] + 1):
        try: # FIXME: load config instead of hard-coding path
            vers = '_'.join('v{:03d}.{:03d}'.format(major, minor) for exp_name, major, minor, _ in versions)
            figname = 'evoscatter_comp/{exp_name}/{vers}/seed{seed}'.format(exp_name=exp_name, vers=vers, seed=seed)
            fig_filepath = fig_aged_perfs(versions, seed, figname,
                                          width=width, height=height, y_range=y_range)
            if open_output:
                subprocess.check_output(['open', fig_filepath])
        except FileNotFoundError as exc:
            print('{}. Skipping.'.format(exc))


def rand_colors(n, seed=2):
    if seed is not None:
        random.seed(seed)
    random_hsv = [(random.uniform(0.0, 360.0),
                   random.uniform(0.2, 1.0),
                   random.uniform(0.45, 0.55)) for i in range(n)]

    return [HSL(*hsv, a=0.25) for hsv in random_hsv]

RANDOM_COLOR = rand_colors(10000)


def fig_genotype_distance(versions, seed, figname,
                          force=True, t_max=4000, width=1200, height=200, y_range=(0, 1.0)):
    suffix = '_mutd.png'
    fig_filepath = os.path.join(paths.FIGURES_DIR, figname + suffix)
    os.makedirs(os.path.dirname(fig_filepath), exist_ok=True)

    # check if the figure is already computed
    if ((not force) and os.path.isfile(fig_filepath) and
        os.path.getmtime(fig_filepath) >= loadlib.last_change(filename)):
        #print('{} is up-to-date; skipping.'.format(fig_filepath))
        print(fig_filepath, os.path.isfile(fig_filepath))
        return

    # y_range = y_range if dist_fitness else (y_range[0], y_range[1]/800)
    graph = graph.Graph(title=figname, y_range=y_range, x_range=(-50, t_max+50),
                         width=width, height=height, tools=None)
    graph.fig.xaxis.ticker = [0, 400, 2000, 4000]

    for exp_name, major, minor, scatter_color in versions:
        run = loadlib.load_results(exp_name, major, minor, seed)
        uids, dists = evo.analysis.genotype_distance_pop(run.population)

        family_count = {}
        for gen_dist in dists:
            for family, d in gen_dist:
                family_count[family] = family_count.get(family, 0) + 1

        family_colors = {}
        #random_colors = rand_colors(len(family_count), seed=0)
        offset = 0
        count_family = sorted([(count, family) for family, count in family_count.items()], reverse=True)
        for count, family in count_family:
            if count > 10:
                family_colors[family] = RANDOM_COLOR[offset]
                offset += 1
            else:
                family_colors[family] = RGB(10, 10, 10, a=0.0)


        xs, ys, colors, alphas = [], [], [], []
        for t, gen_dist in enumerate(dists):
            for family, d in gen_dist:
                xs.append(t)
                ys.append(d)
                colors.append(family_colors[family])
                # alphas.append(0.5) #  if new_member else 0.125)
        graph.fig.scatter(xs, ys, size=1.0, color=colors)

    return graph.save_fig(graph.fig, fig_filepath, scale_factor=2.0, verbose=True)

def mutdscatter_fig(versions, seeds=None, width=1200, height=200,
                    y_range=(0, 0.2), open_output=False):
    exp_name, major, minor, _ = versions[0]
    cfg = loadlib.load_cfg(exp_name, major, minor)

    if seeds is None:
        seeds = range(cfg['meta']['slurm_array'][0], cfg['meta']['slurm_array'][1] + 1)

    for seed in seeds:
        #if seed in range(cfg['meta']['slurm_array'][0], cfg['meta']['slurm_array'][1] + 1):
        try: # FIXME: load config instead of hard-coding path
            vers = '_'.join('v{:03d}.{:03d}'.format(major, minor) for exp_name, major, minor, _ in versions)
            figname = 'mutdscatter_comp/{exp_name}/{vers}/seed{seed}'.format(exp_name=exp_name, vers=vers, seed=seed)
            fig_filepath = fig_genotype_distance(versions, seed, figname,
                                                 width=width, height=height, y_range=y_range)
            if open_output:
                subprocess.check_output(['open', fig_filepath])
        except FileNotFoundError as exc:
            print('{}. Skipping.'.format(exc))


# TODO: cleanup
if __name__ == '__main__':
    evoscatter_fig(((1, 4, color.blue2), (8, 1, color.orange2)), seeds=None, #[1, 2, 27, 87, 64, 52],
                   width=1400, height=660, dist_fitness=True)

    # args = []
    # for major, minor, scatter_color in :
    # # for major, minor, _ in loadlib.find_cfgs('devolearn'):
    #     evoscatter_fig(major, minor, None, scatter_color)
    #     args.append((major, minor, None))
    #
    # with mp.Pool() as pool:
    #    pool.starmap(evoscatter_fig, args)
