import os
import shutil
import warnings
import tempfile
import subprocess
import collections.abc
from pathlib import Path

import bokeh as bk

from .. import utils


# try:
#     get_ipython() # checking if (it seems) we are in a notebook
#     bk.plotting.output_notebook(hide_banner=True)
#     JUPYTER = True
# except NameError:
JUPYTER = False


# Creating figures with nice default

def figure(*args, tools=None, **kwargs):
    """Create a figure with appropriate default."""
    fig = bk.plotting.figure(*args, **kwargs)
    adjust_fig(fig)
    transparent_background(fig)
    return fig

def adjust_fig(fig):
    """Set the layout's defaults for a figure."""
    tight_layout(fig)
    disable_minor_ticks(fig)
    disable_grid(fig)
    fig.toolbar.logo = None

def tight_layout(fig):
    """Make a figure borders tighter."""
    fig.min_border_top    = 35
    fig.min_border_bottom = 35
    fig.min_border_right  = 35
    fig.min_border_left   = 35

def disable_minor_ticks(fig):
    """Remove minor ticks from a figure's axis."""
    fig.axis.major_label_text_font_size = '8pt'
    fig.axis.minor_tick_line_color = None
    fig.axis.major_tick_in = 0

def disable_grid(fig):
    """Remove a grid lines from a figure."""
    fig.xgrid.grid_line_color = None
    fig.ygrid.grid_line_color = None

def transparent_background(fig):
    """Make the background's figure transparent."""
    fig.background_fill_color = None
    fig.border_fill_color     = None


# Saving figure to disk

def save_fig(fig, filepath, verbose=True, scale_factor=1.0):
    """Save figure as png, svg, or pdf.

    :param fig:       The figure to save to disk.
    :param filepath:  'path/to/filepath.pdf'. The extension can be '.pdf', '.png', or '.svg'.
    :param verbose:   If True, displays a confirmation dialog after the operation is completed.
    """
    # filename = filename.replace(' ', '_')
    filepath = os.path.realpath(filepath)
    _, ext = os.path.splitext(filepath)
    ext = ext.lower()
    dirname, filename = os.path.split(filepath)
    if not os.path.exists(dirname):
        Path(dirname).mkdir(parents=True)

    with warnings.catch_warnings():
        #warnings.simplefilter("ignore")
        if ext == '.png':
            original_size = fig.plot_height, fig.plot_width
            fig.plot_height, fig.plot_width = int(scale_factor * fig.plot_height), int(scale_factor * fig.plot_width)
            bk.io.export_png(fig, filename=filepath, timeout=5)
            fig.plot_height, fig.plot_width = original_size
        elif ext == '.svg':
            fig.output_backend = 'svg'
            bk.io.export_svg(fig, filename=filepath, timeout=5)
        elif ext == '.pdf':
            fig.output_backend = 'svg'
            tmp_filepath = os.path.join(tempfile.gettempdir(), filename+'.svg')
            bk.io.export_svg(fig, filename=tmp_filepath, timeout=5)
            if ext == '.pdf':
                svg2pdf(tmp_filepath, destination=filepath)
        else:
            raise ValueError(('The extension must be `.svg`, `.png`, or `.pdf`'
                              ' (was `{}`)').format(ext))
    if verbose:
        print(utils.info('generated {}'.format(filepath)))

    return filepath


def svg2pdf(filename, destination=None):
    """Convert a svg into a pdf.

    Use inkscape if available, else relies on `svglib`
    :param filename:  the path to the svg file, i.e. `path/to/file.svg`
    """
    abs_filename = os.path.abspath(os.path.expanduser(filename))
    pdf_filename = (os.path.abspath(os.path.expanduser(destination))
                    if destination is not None else
                    os.path.splitext(abs_filename)[0] + '.pdf')

    if shutil.which('inkscape') is not None:
        version = str(subprocess.run(['inkscape', '--version'], capture_output=True).stdout)
        if version[11] == '0': # e.g. version == "Inkscape 0.92.4 ..."
            cmd = 'inkscape "{}" --export-pdf="{}"'.format(abs_filename, pdf_filename)
        else:
            cmd = 'inkscape "{}" --export-filename="{}"'.format(abs_filename, pdf_filename)
        with open(os.devnull, 'w') as devnull:
            assert subprocess.call(cmd, shell=True, stdout=devnull) == 0 #, stderr=devnull) == 0
    else:
        try:
            import svglib.svglib
            import reportlab.graphics.renderPDF
            rlg_file = svglib.svglib.svg2rlg(abs_filename)
            reportlab.graphics.renderPDF.drawToFile(rlg_file, pdf_filename)
        except ImportError as e:
            print('Inkscape was not detected. Please install it, or svglib '
                  'using `pip install svglib reportlab`')


# Misc

def unroll(x, size):
    """Unroll `x` if `x` is not already an iterable."""
    if isinstance(x, dict):
        return x
    elif isinstance(x, str):
        return size * (x,)
    elif isinstance(x, collections.abc.Iterable):
        assert len(x) == size
        return x
    else:
        return size * (x,)
