import os
import pickle
from pathlib import Path

import numpy as np
import matplotlib as mpl
mpl.use('Agg')
# import matplotlib
# matplotlib.use("Qt5Agg")
from scipy.signal import savgol_filter
import matplotlib.pyplot as plt
import matplotlib.cm
import matplotlib.gridspec as gridspec
from matplotlib import collections  as mc
from bokeh.models import markers
from matplotlib import ticker

from ... import utils, paths
from .. import loadlib
from . import graph, color



def visu2d_seeds_graph(params, major_a, minor_a, major_b, minor_b, pool=None):
    exp_name = params['exp_name']
    done_a = loadlib.find_done(exp_name, major_a, minor_a)
    done_b = loadlib.find_done(exp_name, major_b, minor_b)
    done = set(done_a).intersection(set(done_b))
    seeds = params['seeds']
    if seeds is not None:
        done = done.intersection(range(seeds[0], seeds[1]+1))

    if pool is None:
        for seed in done:
            visu2d_seed_graph(params, major_a, minor_a, major_b, minor_b, seed)
    else:
        pool.starmap(visu2d_seed_graph, [(params, major_a, minor_a, major_b, minor_b, seed)
                                         for seed in done])


def visu2d_seed_graph(params, major_a, minor_a, major_b, minor_b, seed, pool=None):
    exp_name = params['exp_name']
    analysis_params, graph_params = params['analysis_params'], params['graph_params']
    kind, backend = analysis_params['analysis'], graph_params['backend']
    copy_folder = params.get('copy_folder', None)

    filepaths = loadlib.visu2d_filepaths(analysis_params, exp_name, major_a, minor_a, major_b, minor_b, seed)

    for filepath in filepaths:
        if not filepath.is_file():
            print(utils.warning('{} not found. skipping'.format(filepath)))
            continue

        rel_filepath = filepath.relative_to(paths.POST_DIR)

        metadata = {}
        with open(filepath, 'rb') as fd:
            results = pickle.load(fd)
            #print(results.keys())

        def color_map(values, cmap=mpl.cm.viridis, v_min=None, v_max=None, alpha=1.0):
            colors = []
            for r, g, b, _ in 255*cmap(mpl.colors.Normalize(vmin=v_min, vmax=v_max)(values)):
                colors.append("#{:02x}{:02x}{:02x}{:02x}".format(int(r), int(g), int(b), int(255*alpha)))
            return colors

        xs         = results['map2d'][:,0]
        ys         = results['map2d'][:,1]
        run_id     = results['run_index']
        elite_flag = results['elite_flag']
        metadata   = results['metadata']

        gen_min, gen_max = min(results['generations']), 1.25*max(results['generations'])

        def _separate(id_ref, elite, alpha=1.0):
            xs_sub  = [x for x, r, e in zip(xs,                     run_id, elite_flag) if r == id_ref and e == elite]
            ys_sub  = [y for y, r, e in zip(ys,                     run_id, elite_flag) if r == id_ref and e == elite]
            gen_sub = [g for g, r, e in zip(results['generations'], run_id, elite_flag) if r == id_ref and e == elite]
            colors_sub = color_map(gen_sub, cmap=mpl.cm.viridis, v_min=gen_min, v_max=gen_max, alpha=alpha)
            return xs_sub, ys_sub, colors_sub


        def _cutoff(id_ref, elite, gen_count):
            gen_sub = [g for g, r, e in zip(results['generations'], run_id, elite_flag) if r == id_ref and e == elite]
            for i, g in enumerate(gen_sub):
                if g >= gen_count:
                    return i
            return len(gen_sub)

        title = '{} v{}.{} vs v{}.{}'.format(kind, major_a, minor_a, major_b, minor_b)
        if kind.lower() == 'pca' and len(metadata) > 0:
            percent_explained = 100*float(sum(metadata['explained_variance_ratio']))
            print('explained variation: {:.1f}% ({})'.format(percent_explained,
                                                             metadata['explained_variance_ratio']))
            title += ' (explained variation {:.1f}%)'.format(100*sum(metadata['explained_variance_ratio']))

        xl_0s, yl_0s, xl_1s, yl_1s, line_gen = [], [], [], [], []
        for coo, parent, gen in zip(results['map2d'], results['parent_index'], results['generations']):
            if parent is not None:
                xl_0s.append(coo[0])
                yl_0s.append(coo[1])
                xl_1s.append(results['map2d'][parent][0])
                yl_1s.append(results['map2d'][parent][1])
                line_gen.append(gen)

        xs_a_non_elite, ys_a_non_elite, colors_a_non_elite = _separate(0, False, alpha=0.5)
        xs_b_non_elite, ys_b_non_elite, colors_b_non_elite = _separate(1, False, alpha=0.5)
        xs_a_elite, ys_a_elite, colors_a_elite = _separate(0, True)
        xs_b_elite, ys_b_elite, colors_b_elite = _separate(1, True)

        if backend == 'bokeh':
            g = graph.Graph(title=title, width=1000, height=1000, tools=None)
            g.fig.segment(xl_0s, yl_0s, xl_1s, yl_1s, line_alpha=0.5, line_color=line_colors, line_width=1.0)

            g.fig.scatter(xs_a_non_elite, ys_a_non_elite, marker='square',   size=6.0,
                              fill_color='white', fill_alpha=1.0, line_alpha=0.5, line_color=colors_a_non_elite, line_width=0.5)
            g.fig.scatter(xs_b_non_elite, ys_b_non_elite, marker='triangle', size=9.0,
                              fill_color='white', fill_alpha=1.0, line_alpha=0.5, line_color=colors_b_non_elite, line_width=0.5)

            g.fig.scatter(xs_a_elite, ys_a_elite, marker='square', size=6.0,
                              fill_alpha=1.0, fill_color=colors_a_elite, line_color=None)
            # graph.fig.scatter(xs_a_elite, ys_a_elite, marker='square',   size=6.0, alpha=0.5, color=colors_a_elite)
            g.fig.scatter(xs_b_elite, ys_b_elite, marker='triangle', size=9.0,
                              fill_alpha=1.0, fill_color=colors_b_elite, line_color=None)
            # graph.fig.scatter(xs_b_elite, ys_b_elite, marker='triangle', size=7.0, alpha=0.5, color=colors_b_elite)

            g.fig.scatter(xs[:1], ys[:1], marker='circle', size=10.0, alpha=1.0, fill_color='red', line_color=None)
            g.save_fig(graph.fig, str(fig_filepath) + '.png', scale_factor=2.0, verbose=True)
            #graphs.save_fig(graph.fig, fig_filepath + '.pdf', scale_factor=2.0, verbose=True)  # does not work well

        elif backend == 'matplotlib':
            for smoothing in [True, False]:
                alpha_factor = 0.1 if smoothing else 1.0

                plt.rcParams.update({'font.size': 2})
                plt.rcParams['axes.linewidth'] = 0.2

                fig, ax = plt.subplots()
                ax.set_aspect('equal')
                ax.tick_params(width=0.2)

                line_colors = color_map(line_gen, cmap=mpl.cm.viridis, v_min=gen_min, v_max=gen_max, alpha=alpha_factor*0.5)
                lines = [((x0, y0), (x1, y1)) for x0, y0, x1, y1 in zip(xl_0s, yl_0s, xl_1s, yl_1s)]
                lc = mc.LineCollection(lines, colors=line_colors, linewidths=0.1, zorder=-1)
                ax.add_collection(lc)

                sc = ax.scatter(xs_a_non_elite, ys_a_non_elite, c='white', marker="s", s=2.0, alpha=alpha_factor*1.0, edgecolors=colors_a_non_elite, linewidths=0.1)
                ax.scatter(xs_b_non_elite, ys_b_non_elite, c='white', marker='^', s=2.5, alpha=alpha_factor*1.0, edgecolors=colors_b_non_elite, linewidths=0.1)

                ax.scatter(xs_a_elite, ys_a_elite, c=colors_a_elite,  marker="s", s=2.0, alpha=alpha_factor*1.0, edgecolors='none')
                ax.scatter(xs_b_elite, ys_b_elite, c=colors_b_elite,  marker='^', s=2.5, alpha=alpha_factor*1.0, edgecolors='none')

                if smoothing:
                    window, poly = 101, 3
                    xa_smooth = savgol_filter([xs[0]] + xs_a_elite, window, poly, mode='nearest') # window size 51, polynomial order 3
                    ya_smooth = savgol_filter([ys[0]] + ys_a_elite, window, poly, mode='nearest')
                    xb_smooth = savgol_filter([xs[0]] + xs_b_elite, window, poly, mode='nearest') # window size 51, polynomial order 3
                    yb_smooth = savgol_filter([ys[0]] + ys_b_elite, window, poly, mode='nearest')

                    xa_cutoff, xb_cutoff = _cutoff(0, True, 2000), _cutoff(1, True, 2000)
                    # print(len(xa_smooth), xa_cutoff, len(xb_smooth), xb_cutoff)
                    ax.plot(xa_smooth[:xa_cutoff], ya_smooth[:xa_cutoff], color=color.control[1],     linewidth=1, alpha=1.0)
                    ax.plot(xa_smooth[xa_cutoff:], ya_smooth[xa_cutoff:], color=color.control[2],     linewidth=1, alpha=1.0)
                    ax.plot(xb_smooth[:xb_cutoff], yb_smooth[:xb_cutoff], color=color.development[1], linewidth=1, alpha=1.0)
                    ax.plot(xb_smooth[xb_cutoff:], yb_smooth[xb_cutoff:], color=color.development[2], linewidth=1, alpha=1.0)

                ax.scatter(xs[:1], ys[:1], c='red',  marker="s", s=4.0, alpha=1.0, edgecolors='none')


                ax.spines['right'].set_visible(False)
                ax.spines['top'].set_visible(False)
                ax.set_title(title)
                plt.xlim(-1.4, 1.4)
                plt.ylim(-1.4, 1.4)
                ax.grid(False)
                plt.subplots_adjust(left=0.05, right=0.95, top=0.95, bottom=0.05)

                fig_filename = rel_filepath.with_suffix('')
                if smoothing:
                    fig_filename = str(fig_filename) + '_smoothed'
                fig_filepath = Path(paths.FIGURES_DIR, fig_filename)
                fig_filepath.parent.mkdir(parents=True, exist_ok=True)

                pdf_filepath = str(fig_filepath) + '.pdf'
                plt.savefig(pdf_filepath, bbox_inches='tight',pad_inches=0.05)
                plt.close(fig)
                utils.print_info('saved {}'.format(pdf_filepath))

                if copy_folder is not None:
                    copy_filepath = Path(paths.FIGURES_DIR, copy_folder, str(fig_filename).replace('/', '_'))
                    shutil.copy(pdf_filepath, copy_filepath)
                    utils.print_info('copy @{}'.format(copy_filepath))

        else:
            raise ValueError('unknow backend "{}"'.format(backend))



def visu2d_seeds_graph_separate(params, major_a, minor_a, major_b, minor_b, pool=None):
    exp_name = params['exp_name']
    done_a = loadlib.find_done(exp_name, major_a, minor_a)
    done_b = loadlib.find_done(exp_name, major_b, minor_b)
    done = set(done_a).intersection(set(done_b))
    seeds = params['seeds']
    if seeds is not None:
        done = done.intersection(range(seeds[0], seeds[1]+1))

    if pool is None:
        for seed in done:
            visu2d_seed_graph_separate(params, major_a, minor_a, major_b, minor_b, seed)
    else:
        pool.starmap(visu2d_seed_graph_separate, [(params, major_a, minor_a, major_b, minor_b, seed)
                                                  for seed in done])

def visu2d_seed_graph_separate(params, major_a, minor_a, major_b, minor_b, seed, pool=None):
    exp_name = params['exp_name']
    analysis_params, graph_params = params['analysis_params'], params['graph_params']
    kind = analysis_params['analysis']
    copy_folder = params.get('copy_folder', None)

    filepaths_a = loadlib.visu2d_filepaths_single(analysis_params, exp_name, major_a, minor_a, seed)
    filepaths_b = loadlib.visu2d_filepaths_single(analysis_params, exp_name, major_b, minor_b, seed)
    filepaths_ab = loadlib.visu2d_filepaths(analysis_params, exp_name, major_a, minor_a, major_b, minor_b, seed)

    for filepath_a, filepath_b, filepath_ab in zip(filepaths_a, filepaths_b, filepaths_ab):
        if not (filepath_a.is_file() and filepath_b.is_file()):
            print(utils.warning('{} or {} not found. skipping'.format(filepath_a, filepath_b)))
            continue

        rel_filepath = filepath_ab.relative_to(paths.POST_DIR)

        def color_map(values, cmap=mpl.cm.viridis, v_min=None, v_max=None, alpha=1.0):
            colors = []
            for r, g, b, _ in 255*cmap(mpl.colors.Normalize(vmin=v_min, vmax=v_max)(values)):
                colors.append("#{:02x}{:02x}{:02x}{:02x}".format(int(r), int(g), int(b), int(255*alpha)))
            return colors

        def prepare_data(results):
            gen_min, gen_max = min(results['generations']), 1.25*max(results['generations'])

            xs, ys     = results['map2d'][:,0], results['map2d'][:,1]
            elite_flag = results['elite_flag']
            metadata   = results['metadata']

            x_e,  y_e,  gen_e,  fit_e  = [], [], [], []
            x_ne, y_ne, gen_ne, fit_ne = [], [], [], []
            for x, y, g, f, e in zip(xs, ys, results['generations'], results['fitnesses'], results['elite_flag']):
                if e:
                    x_e.append(x)
                    y_e.append(y)
                    gen_e.append(g)
                    fit_e.append(f)
                else:
                    x_ne.append(x)
                    y_ne.append(y)
                    gen_ne.append(g)
                    fit_ne.append(f)

            xl_0s, yl_0s, xl_1s, yl_1s, line_gen = [], [], [], [], []
            for coo, parent, gen in zip(results['map2d'], results['parent_index'], results['generations']):
                if parent is not None:
                    xl_0s.append(coo[0])
                    yl_0s.append(coo[1])
                    xl_1s.append(results['map2d'][parent][0])
                    yl_1s.append(results['map2d'][parent][1])
                    line_gen.append(gen)

            assert results['elite_flag'][-1] # last is elite
            xld_0s, yld_0s, xld_1s, yld_1s, line_gen_d = [], [], [], [], []
            index, parent_index = -1, results['parent_index'][-1]
            while parent_index is not None:
                xld_0s.append(xs[index])
                yld_0s.append(ys[index])
                xld_1s.append(xs[parent_index])
                yld_1s.append(ys[parent_index])
                line_gen_d.append(results['generations'][index])
                index, parent_index = parent_index, results['parent_index'][parent_index]

            color_e  = color_map(gen_e,  cmap=mpl.cm.viridis, v_min=gen_min, v_max=gen_max, alpha=1.00)
            color_ne = color_map(gen_ne, cmap=mpl.cm.viridis, v_min=gen_min, v_max=gen_max, alpha=0.25)
            return (x_e, y_e, color_e), (x_ne, y_ne, color_ne), (xl_0s, yl_0s, xl_1s, yl_1s, line_gen), (xld_0s, yld_0s, xld_1s, yld_1s, line_gen_d)

        title = ''
        for smoothing in [False]:
            alpha_factor = 0.1 if smoothing else 1.0

            plt.rcParams.update({'font.size': 2})
            plt.rcParams['axes.linewidth'] = 0.2

            fig, (ax_a, ax_b) = plt.subplots(1, 2)
            # fig = plt.figure()
            # gs = gridspec.GridSpec(1, 2, width_ratios=[1,1])
            #
            # ax1 = plt.subplot(gs[0])
            # ax2 = plt.subplot(gs[1])

            min_x, max_x, min_y, max_y = float('inf'), float('-inf'), float('inf'), float('-inf')

            for i, (ax, filepath) in enumerate([(ax_a, filepath_a), (ax_b, filepath_b)]):
                results = loadlib.load_data(filepath)
                point_e, point_ne, lines, lines_d = prepare_data(results)
                x_e, y_e, color_e    = point_e
                x_ne, y_ne, color_ne = point_ne
                xl_0s, yl_0s, xl_1s, yl_1s, line_gen = lines
                xld_0s, yld_0s, xld_1s, yld_1s, line_gen_d = lines_d

                index_2000 = 0
                for i, g in enumerate(results['generations']):
                    if g >= 2000:
                        index_2000 = i
                        break

                local_max_x = max(np.max(x_e), np.max(x_ne))
                local_min_x = min(np.min(x_e), np.min(x_ne))

                max_x = np.max([max_x, np.max(x_e), np.max(x_ne)])
                min_x = np.min([min_x, np.min(x_e), np.min(x_ne)])
                max_y = np.max([max_y, np.max(y_e), np.max(y_ne)])
                min_y = np.min([min_y, np.min(y_e), np.min(y_ne)])

                ax.set_aspect('equal')
                # ax.set_xlim(local_min_x - 0.1, local_max_x + 0.1)
                ax.tick_params(width=0.2, length=1, pad=1)

                if i == 0:
                    # line_colors = color_map(line_gen, cmap=mpl.cm.viridis, v_min=gen_min, v_max=gen_max, alpha=alpha_factor*0.5)
                    lines2 = [((x0 + 0.15, y0 + 0.4), (x1 + 0.15, y1 + 0.4)) for x0, y0, x1, y1 in zip(xld_0s, yld_0s, xld_1s, yld_1s)]
                    lc2 = mc.LineCollection(lines2, colors='grey', linewidths=0.1, zorder=-1)
                    ax.add_collection(lc2)
                    ax.scatter([x_e[0] + 0.15],  [y_e[0] + 0.4],  c='white', marker="v", s=5.0, alpha=alpha_factor*1.0, edgecolors='grey', linewidths=0.1)
                    ax.scatter([x_e[-1] + 0.15], [y_e[-1] + 0.4], c='white', marker="*", s=7.0, alpha=alpha_factor*1.0, edgecolors='grey', linewidths=0.1)

                gen_min, gen_max = min(results['generations']), 1.25*max(results['generations'])
                line_colors = color_map(line_gen, cmap=mpl.cm.viridis, v_min=gen_min, v_max=gen_max, alpha=alpha_factor*0.5)
                lines = [((x0, y0), (x1, y1)) for x0, y0, x1, y1 in zip(xl_0s, yl_0s, xl_1s, yl_1s)]
                lc = mc.LineCollection(lines, colors=line_colors, linewidths=0.1, zorder=-1)
                ax.add_collection(lc)

                ax.scatter(x_ne, y_ne, c='white', marker='^', s=1, alpha=alpha_factor*1.0, edgecolors=color_ne, linewidths=0.1)
                sc = ax.scatter(x_e,  y_e,  c=color_e, marker='^', s=1, alpha=alpha_factor*1.0, edgecolors='none')
                cbar = plt.colorbar(sc, ax=ax, shrink=0.25)
                tick_locator = ticker.MaxNLocator(nbins=4)
                cbar.locator = tick_locator
                cbar.update_ticks()

                metadata   = results['metadata']
                if kind.lower() == 'pca':
                    title = '(explained variation {:.1f}%)'.format(100*sum(metadata['explained_variance_ratio']))
                #
                # title = '{} v{}.{} vs v{}.{}'.format(kind, major_a, minor_a, major_b, minor_b)
                # if kind.lower() == 'pca' and len(metadata) > 0:
                #     percent_explained = 100*float(sum(metadata['explained_variance_ratio']))
                #     print('explained variation: {:.1f}% ({})'.format(percent_explained,
                #                                                      metadata['explained_variance_ratio']))
                #     title += ' (explained variation {:.1f}%)'.format(100*sum(metadata['explained_variance_ratio']))
                #

                # if smoothing:
                #     window, poly = 101, 3
                #     xa_smooth = savgol_filter([xs[0]] + xs_a_elite, window, poly, mode='nearest') # window size 51, polynomial order 3
                #     ya_smooth = savgol_filter([ys[0]] + ys_a_elite, window, poly, mode='nearest')
                #     xb_smooth = savgol_filter([xs[0]] + xs_b_elite, window, poly, mode='nearest') # window size 51, polynomial order 3
                #     yb_smooth = savgol_filter([ys[0]] + ys_b_elite, window, poly, mode='nearest')
                #
                #     xa_cutoff, xb_cutoff = _cutoff(0, True, 2000), _cutoff(1, True, 2000)
                #     # print(len(xa_smooth), xa_cutoff, len(xb_smooth), xb_cutoff)
                #     ax.plot(xa_smooth[:xa_cutoff], ya_smooth[:xa_cutoff], color=color.control[1],     linewidth=1, alpha=1.0)
                #     ax.plot(xa_smooth[xa_cutoff:], ya_smooth[xa_cutoff:], color=color.control[2],     linewidth=1, alpha=1.0)
                #     ax.plot(xb_smooth[:xb_cutoff], yb_smooth[:xb_cutoff], color=color.development[1], linewidth=1, alpha=1.0)
                #     ax.plot(xb_smooth[xb_cutoff:], yb_smooth[xb_cutoff:], color=color.development[2], linewidth=1, alpha=1.0)

                x_2000, y_2000 = results['map2d'][index_2000]
                ax.scatter([x_2000], [y_2000], c='red', marker="o", s=4.0, alpha=0.75, edgecolors='none')
                ax.scatter(x_e[:1],  y_e[:1],  c='red', marker="v", s=5.0, alpha=0.75, edgecolors='none')
                ax.scatter(x_e[-1:], y_e[-1:], c='red', marker="*", s=9.0, alpha=0.75, edgecolors='none')

                ax.spines['right'].set_visible(False)
                ax.spines['top'].set_visible(False)
                ax.set_title(title)
                ax.grid(False)

            for ax in [ax_a, ax_b]:
                ax.set_xlim(min_x - 0.1, max_x + 0.1)
                ax.set_ylim(min_y - 0.1, max_y + 0.1)

            plt.subplots_adjust(left=0.05, right=0.95, top=0.95, bottom=0.05)

            fig_filename = rel_filepath.with_suffix('')
            if smoothing:
                fig_filename = str(fig_filename) + '_smoothed'
            fig_filepath = Path(paths.FIGURES_DIR, fig_filename)
            fig_filepath.parent.mkdir(parents=True, exist_ok=True)

            pdf_filepath = str(fig_filepath) + '.pdf'
            plt.savefig(pdf_filepath, bbox_inches='tight',pad_inches=0.05)
            plt.close(fig)
            utils.print_info('saved {}'.format(pdf_filepath))

            if copy_folder is not None:
                copy_filepath = Path(paths.FIGURES_DIR, copy_folder, str(fig_filename).replace('/', '_'))
                shutil.copy(pdf_filepath, copy_filepath)
                utils.print_info('copy @{}'.format(copy_filepath))




def visu2d_graph(params, major_a, minor_a, major_b, minor_b, pool=None):
    exp_name = params['exp_name']
    analysis_params, graph_params = params['analysis_params'], params['graph_params']
    kind, backend = analysis_params['analysis'], graph_params['backend']

    filepaths = loadlib.visu2d_filepaths(analysis_params, exp_name, major_a, minor_a, major_b, minor_b)

    for filepath in filepaths:
        try:
            if not filepath.is_file():
                print(utils.warning('{} not found. skipping'.format(filepath)))

            rel_filepath = filepath.relative_to(paths.POST_DIR)

            metadata = {}
            with open(filepath, 'rb') as fd:
                data = pickle.load(fd)
                tsne_results, exps, fitnesses = data[:3]
                if len(data) == 4:
                    metadata = data[3]

            def color_map(values, cmap=mpl.cm.viridis):
                colors = []
                for r, g, b, _ in 255*cmap(mpl.colors.Normalize()(values)):
                    colors.append("#{:02x}{:02x}{:02x}".format(int(r), int(g), int(b)))
                return colors

            xs = tsne_results[:,0]
            ys = tsne_results[:,1]

            colors = color_map(fitnesses, cmap=mpl.cm.viridis)
            xs_a     = [x for x, e in zip(xs, exps)     if e == 0]
            ys_a     = [y for y, e in zip(ys, exps)     if e == 0]
            colors_a = [c for c, e in zip(colors, exps) if e == 0]
            xs_b     = [x for x, e in zip(xs, exps)     if e == 1]
            ys_b     = [y for y, e in zip(ys, exps)     if e == 1]
            colors_b = [c for c, e in zip(colors, exps) if e == 1]

            fig_filename = rel_filepath.with_suffix('')
            fig_filepath = Path(paths.FIGURES_DIR, fig_filename)
            fig_filepath.parent.mkdir(parents=True, exist_ok=True)

            title = '{} v{}.{} vs v{}.{}'.format(kind, major_a, minor_a, major_b, minor_b)
            if kind.lower() == 'pca' and len(metadata) > 0:
                title += ' (explained variation {:.1f}%)'.format(100*sum(metadata['explained_variance_ratio']))

            if backend == 'bokeh':
                graph = graphs.Graph(title=title, width=500, height=500, tools=None)
                graph.fig.scatter(xs_a, ys_a, marker='square',   size=20.0, alpha=0.75, fill_color=color.control,     line_color=None, line_width=0.5)
                graph.fig.scatter(xs_b, ys_b, marker='triangle', size=25.0, alpha=0.75, fill_color=color.development, line_color=None, line_width=1.0)
                # graph.fig.scatter(xs_a, ys_a, marker='square',   size=20.0, alpha=0.75, fill_color=colors_a, line_color=None, line_width=0.5)
                # graph.fig.scatter(xs_b, ys_b, marker='triangle', size=25.0, alpha=0.75, fill_color=colors_b, line_color=None, line_width=1.0)
                if exps[0] == -1:
                    graph.fig.scatter(xs[:1], ys[:1], marker='square',   size=20.0, alpha=0.75, fill_color='red', line_color=None,  line_width=0.5)
                    graph.fig.scatter(xs[:1], ys[:1], marker='triangle', size=25.0, alpha=0.75, fill_color='red', line_color=None,  line_width=1.0)
                    # graph.fig.scatter(xs[:1], ys[:1], marker='circle',   size=50.0, alpha=0.25, fill_color=None,  line_color='red', line_width=1.0)

                graphs.save_fig(graph.fig, fig_filepath + '.png', scale_factor=2.0, verbose=True)
                graphs.save_fig(graph.fig, fig_filepath + '.pdf', scale_factor=2.0, verbose=True)

            elif backend == 'matplotlib':
                plt.rcParams.update({'font.size': 8})
                plt.rcParams['axes.linewidth'] = 1.0

                fig, ax = plt.subplots()
                ax.tick_params(width=1.0, length=4, pad=2)
                ax.spines['right'].set_visible(False)
                ax.spines['top'].set_visible(False)
                ax.set_title(title)
                ax.grid(False)
                plt.subplots_adjust(left=0.05, right=0.95, top=0.95, bottom=0.05)
                ax.set_aspect('equal')
                #plt.xlim(-1.5, 1.5)
                #plt.ylim(-1.5, 1.5 )

                size = 45.0
                # graph.fig.scatter(xs_a, ys_a, marker='square',   size=10.0, alpha=0.75, fill_color=colors_a, line_color=None, line_width=0.5)
                # graph.fig.scatter(xs_b, ys_b, marker='triangle', size=15.0, alpha=0.75, fill_color=colors_b, line_color=None, line_width=1.0)
                ax.scatter(xs_a, ys_a, c=color.control[1],     marker="s", s=size, alpha=0.75, edgecolors=None, linewidths=0.5)
                ax.scatter(xs_b, ys_b, c=color.development[1], marker="^", s=1.2*size, alpha=0.75, edgecolors=None, linewidths=0.5)
                # ax.scatter(xs_a, ys_a, c=colors_a, marker="s", s=size, alpha=0.75, edgecolors=None, linewidths=0.5)
                # ax.scatter(xs_b, ys_b, c=colors_b, marker="^", s=1.2*size, alpha=0.75, edgecolors=None, linewidths=0.5)

                # first population genotype
                if exps[0] == -1:
                    # ax.scatter(xs[:1], ys[:1], c='white',  marker="o", s=30.0, alpha=0.25, edgecolors='red', linewidths=0.2)
                    ax.scatter(xs[:1], ys[:1], c='red', marker="s", s=size, alpha=0.75, edgecolors='none', linewidths=0.0)
                    ax.scatter(xs[:1], ys[:1], c='red', marker="^", s=size, alpha=0.75, edgecolors='none', linewidths=0.0)

                plt.savefig(str(fig_filepath) + '.pdf', bbox_inches='tight', pad_inches=0.05)
                plt.close(fig)
                utils.print_info('saved {}'.format(str(fig_filepath) + '.pdf'))
            else:
                raise ValueError('unknow backend "{}"'.format(backend))
        except FileNotFoundError as exc:
            print(exc)


if __name__ == '__main__':
    exp_name = 'devoevo_draft_delay'
    for major_a, minor_a, major_b, minor_b in [(15, 8, 15, 9),
                                               # (15, 8, 15, 10),
                                               # (15, 10, 15, 9)
                                              ]:
        #visu_graph('PCA', exp_name, major_a, minor_a, major_b, minor_b)
        visu_graph_seed_all('PCA', exp_name, major_a, minor_a, major_b, minor_b, seeds=(1, 5))
        # for n_neighbors in [10, 15, 20, 30, 50, 100]:
        #     visu_graph('UMAP', exp_name, major_a, minor_a, major_b, minor_b, n_neighbors=n_neighbors, min_d=0.3)
        # for perplexity in [10, 15, 20, 30, 50, 100]:
        #     visu_graph('t-SNE', exp_name, major_a, minor_a, major_b, minor_b, perplexity=perplexity, n_iter=10000)
