import os
import re
import glob
import pickle
import platform
import subprocess
from pathlib import Path

import numpy as np
import dateutil.parser

from .. import utils, evo, paths
from . import array_str



DEFAULT_EXP = 'devoevo_draft_delay'

seed_regexp = re.compile('seed(?P<seed>[0-9]+)')
seed_perfs_regexp = re.compile('seed(?P<seed>[0-9]+)_.*')

def last_change(filename):
    """Return the date of the last change"""
    if platform.system() == 'Darwin':  # macos: catches rsync uploads
        st = subprocess.check_output(["mdls", "-name", "kMDItemDateAdded", "-raw", filename])
        return dateutil.parser.parse(st).timestamp()
    else:
        return os.path.getmtime(filename)

def find_cfg(exp_name, major, minor, verbose=True):
    regexp = 'v{major:03d}*/v{major}_{minor}_*'.format(major=major, minor=minor)
    regexp = os.path.join(paths.EXP_DIR, exp_name, regexp)
    matches = glob.glob(regexp)
    regexp = 'v{major:03d}*/v{major}_{minor:02d}_*'.format(major=major, minor=minor)
    regexp = os.path.join(paths.EXP_DIR, exp_name, regexp)
    matches.extend(glob.glob(regexp))
    matches = list(set(matches))

    if len(matches) == 0:
        if verbose:
            utils.print_warning('find_cfg: no config found for {}.{}'.format(major, minor))
        return None
    if len(matches) > 1:
        if verbose:
            utils.print_warning('find_cfg: multiple matches: {}'.format(matches))
        return None
    return matches[0]

def load_cfg(exp_name, major, minor, notfound_okay=True):
    filepath = find_cfg(exp_name, major, minor)
    return None if filepath is None else utils.load_config(filepath)


def version_str(major, minor):
    return '{:03d}.{:03d}'.format(major, minor)

def prefix_path(exp_name, major, minor, seed=None):
    if seed is None:
        return '{}/v{}'.format(exp_name, version_str(major, minor))
    else:
        return '{}/v{}/seed{}'.format(exp_name, version_str(major, minor), seed)

def prefix_path_from_cfg(cfg, seed=None):
    exp_name = cfg['meta'].get('exp_name', None)
    if exp_name is not None:
        return prefix_path(exp_name, cfg['meta']['major'], cfg['meta']['minor'], seed)
    else:
        # deprecated, only there for backward compatibility
        version = version_str(cfg['meta']['major'], cfg['meta']['minor'])
        filepath = cfg['meta']['filename_prefix'].format(exp_name=exp_name, version=version, seed=seed)
        if seed is None:
            return str(Path(filepath).parent)
        else:
            return filepath

def majors(exp_name):
    """Return all the majors cfgs founds."""
    found = find_cfgs(exp_name=exp_name)
    return sorted(set([major for major, _, _ in found]))

def minors(exp_name, major):
    """Return all the minors cfgs founds for a given major experiment numbering."""
    found = find_cfgs(exp_name=exp_name, major=major)
    return [minor for major, minor, _ in found]

def find_cfgs(exp_name=DEFAULT_EXP, major=None, directory=None):
    if directory is None:
        directory = os.path.abspath(os.path.join(paths.EXP_DIR, exp_name))
    found = []
    glob_exp = '{}/v*/v*_*_*'.format(directory)
    for filepath in glob.glob(glob_exp):
        reg_exp = directory + '/v([0-9]+)((_.*)?)/v(?P<major>[0-9]+)_(?P<minor>[0-9]+)_(.*).yaml'
        m = re.fullmatch(reg_exp, filepath)
        if m is not None:
            major_found, minor = int(m.group('major')), int(m.group('minor'))
            if major is not None:
                if major == major_found:
                    found.append((major_found, minor, filepath))
            else:
                found.append((major_found, minor, filepath))
    return sorted(found)

def find_done(exp_name, major, minor, cfg=None):
    """Return a list of seed that are done, even outside of scheduled jobs"""
    if cfg is None:
        cfg = load_cfg(exp_name, major, minor)
        assert major == cfg['meta']['major'] and minor == cfg['meta']['minor']

    return find_done_cfg(cfg)

# def check_integrity(path):
#     try:
#         if pool is None:
#             pool = mp.Pool()
#             pool.map()
#
#     finally:
#         if pool is not None:
#             pool.close()
#             pool.join()
#             pool.terminate()


def find_started_cfg(cfg, with_results=False, check_integrity=False, done_only=False,
                     directories=paths.RESULTS_DIRS):
    """Return a list of seed that are started or done, even outside of scheduled jobs

    :param with_results:     if True, check that the manifest is also present.
    :param check_integrity:  if True, if a manifest file is found, check that the file hashes
                             match the content of the files. Raise
    """
    seeds = []
    seeds_to_check = []
    for results_dir in directories:
        path = Path(results_dir, prefix_path_from_cfg(cfg, 0)).parent
        #print(path.is_dir(), path)
        if path.is_dir():
            for subpath in os.listdir(path):
                full_subpath = Path(path, subpath)
                if full_subpath.is_dir() and ((not done_only) or Path(full_subpath, 'done').exists()):
                    m = seed_regexp.fullmatch(subpath)
                    if m is not None:
                        seed = int(m.group('seed'))
                        manifest_filepath = Path(full_subpath, 'manifest.yaml')
                        verify_seed = False
                        if manifest_filepath.is_file():
                            if check_integrity:
                                if verify_run_from_cfg(cfg, seed):
                                    seeds.append(seed)
                            else:
                                seeds.append(seed)
                        else:
                            if not with_results:
                                seeds.append(seed)


    return sorted(set(seeds))

def find_done_cfg(cfg, with_results=False, check_integrity=False, directories=paths.RESULTS_DIRS):
    """Return a list of seed that are done, even outside of scheduled jobs

    :param with_results:     if True, check that the manifest is also present.
    :param check_integrity:  if True, if a manifest file is found, check that the file hashes
                             match the content of the files.
    """
    return find_started_cfg(cfg, with_results=with_results, check_integrity=check_integrity,
                            done_only=True, directories=directories)


    ## verify

def verify_run(exp_name, major, minor, seed, verbose=True):
    cfg = load_cfg(exp_name, major, minor)
    return verify_run_from_cfg(cfg, seed, verbose=verbose)

def verify_run_from_cfg(cfg, seed, verbose=True):
    filepath_seed = manifest_path_from_cfg(cfg, seed)
    try:
        return evo.Run.verify(Path(filepath_seed, 'manifest.yaml'))
    except ValueError as exc:
        if verbose:
            print(exc)
        return False



def done_path(cfg, seed):
    for results_dir in paths.RESULTS_DIRS:
        path = Path(results_dir, prefix_path_from_cfg(cfg, seed))
        if Path(path, 'done').exists():
            return path

def manifest_path(exp_name, major, minor, seed, directories=paths.RESULTS_DIRS):
    cfg = load_cfg(exp_name, major, minor)
    return manifest_path_from_cfg(cfg, seed, directories=directories)

def manifest_path_from_cfg(cfg, seed, directories=paths.RESULTS_DIRS):
    for results_dir in directories:
        path = Path(results_dir, prefix_path_from_cfg(cfg, seed))
        if Path(path, 'manifest.yaml').is_file():
            return path

def manifest_done_path(exp_name, major, minor, seed, directories=paths.RESULTS_DIRS):
    cfg = load_cfg(exp_name, major, minor)
    return manifest_done_path_from_cfg(cfg, seed, directories=directories)

def manifest_done_path_from_cfg(cfg, seed, directories=paths.RESULTS_DIRS):
    prefix_path =  prefix_path_from_cfg(cfg, seed)
    for results_dir in directories:
        path = Path(results_dir, prefix_path)
        if Path(path, 'manifest.yaml').is_file() and Path(path, 'done').is_file():
            return path
    raise FileNotFoundError('no manifest found in any result dir at {}'.format(prefix_path))


def is_done(exp_name, major, minor, seed):
    for results_dir in paths.RESULTS_DIRS:
        if Path(results_dir, prefix_path(exp_name, major, minor, seed), 'done').is_file():
            return True
    return False

def is_done_cfg(cfg, seed):
    for results_dir in paths.RESULTS_DIRS:
        if Path(results_dir, prefix_path_from_cfg(cfg, seed), 'done').is_file():
            return True
    return False


    ## fitness ##

def find_done_perfs(exp_name, major, minor, cfg=None):
    """Return a list of seeds for which the perf file is present, even outside of scheduled jobs"""
    if cfg is None:
        cfg = load_cfg(exp_name, major, minor)
        assert major == cfg['meta']['major'] and minor == cfg['meta']['minor']
    return find_done_perfs_cfg(cfg)

def find_done_perfs_cfg(cfg, fitness_function=None):
    """Return a list of seeds for which the perf file is present, even outside of scheduled jobs"""
    path = fitness_filepath_from_cfg(cfg, 0, fitness_function).parent

    seeds = []
    if os.path.isdir(path):
        for filename in os.listdir(path):
            if filename.startswith('seed') and filename.endswith('.npy'):
                seed = int(filename[4:].split('_')[0])
                seeds.append(seed)
    return sorted(seeds)

def perf_suffix(cfg, fitness_function=None):
    if fitness_function is None:
        fitness_function = cfg['member']['environment'].get('fitness_function', 'distance')
    return {'distance':     '_distance.fitness.npy',
            'body_lengths': '_bodylength.fitness.npy'}[fitness_function]

def load_perffile(cfg, seed, fitness_function=None, verbose=True):
    filepath = fitness_filepath_from_cfg(cfg, seed, fitness_function)
    if verbose:
        print('loading {}'.format(filepath))
    try:
        data = np.load(filepath, allow_pickle=True)
    except EOFError as e:
        print(utils.error('file {} appear to be corrupted'.format(filepath)))
        raise e
    if len(data) <= 2:
        return np.array(data, dtype=np.float32)
    return np.array([d for d in data[2:]], dtype=np.float32)



    ## Filepaths ##

    # fitness

def fitness_filepath(exp_name, major, minor, seed, fitness_function=None, subdir='fitness'):
    return fitness_filepath_from_cfg(load_cfg(exp_name, major, minor), seed, fitness_function,
                                     subdir=subdir)

def fitness_filepath_from_cfg(cfg, seed, fitness_function=None, subdir='fitness'):
    name = 'seed{}{}'.format(seed, perf_suffix(cfg, fitness_function))
    return Path(paths.POST_DIR, prefix_path_from_cfg(cfg), subdir, name)

    # elite fitness

def elite_fitness_filepath(exp_name, major, minor, confidence, fitness_function=None):
    return elite_fitness_filepath_from_cfg(load_cfg(exp_name, major, minor), confidence, fitness_function)

def elite_fitness_filepath_from_cfg(cfg, confidence, fitness_function=None):
    name = 'elite_distrib_c{}'.format(confidence) + perf_suffix(cfg, fitness_function)
    return Path(paths.POST_DIR, prefix_path_from_cfg(cfg), name)

    # end fitness

def end_fitness_filepath(exp_name, major, minor, fitness_function=None):
    return end_fitness_filepath_from_cfg(load_cfg(exp_name, major, minor), fitness_function)

def end_fitness_filepath_from_cfg(cfg, fitness_function=None):
    name = 'end_fitness' + perf_suffix(cfg, fitness_function)
    name = Path(Path(name).stem).stem + '.pickle'
    return Path(paths.POST_DIR, prefix_path_from_cfg(cfg), name)

    # angle

def angle_filepath_from_cfg(cfg, window=None, all=False):
    window = (-50, None) if window is None else window
    # window_str = '' if window in [None, (-50, None)] else '[{}-{}]'.format(*window)
    window_str = '[{}-{}]'.format(*window)
    suffix = '_all' if all else ''

    return Path(paths.POST_DIR, prefix_path_from_cfg(cfg), 'angle{}{}.pickle'.format(window_str, suffix))


    # visu2d

def visu2d_filepaths_single(analysis_params, exp_name, major, minor, seed=None):
    kind = analysis_params['analysis']
    window = analysis_params.get('window', None)
    window_str = '' if window is None else '_w{}_{}'.format(window[0], window[1])

    filepaths = []
    seed_str = '' if seed is None else '_seed{}'.format(seed)
    if kind.lower() == 'pca':
        filename = 'pca{}_elites{}.pickle'.format(seed_str, window_str)
        filepaths.append(Path(paths.POST_DIR, prefix_path(exp_name, major, minor), 'pca', filename))
    elif kind.lower() in ['t-sne', 'tsne']:
        perplexities = analysis_params['perplexities']
        n_iter       = analysis_params['n_iter']
        for perplexity in perplexities:
            filename = 'tsne{}_elites_p{}_it{}{}.pickle'.format(seed_str, perplexity, n_iter, window_str)
            filepaths.append(Path(paths.POST_DIR, prefix_path(exp_name, major, minor), 'tsne', filename))

    elif kind.lower() == 'umap':
        n_neighbors = analysis_params['n_neighbors']
        min_d       = analysis_params['min_d']
        for nn in n_neighbors:
            filename = 'umap{}_elites_nn{}_d{}{}.pickle'.format(seed_str, nn, min_d, window_str)
            filepaths.append(Path(paths.POST_DIR, prefix_path(exp_name, major, minor), 'umap', filename))
    else:
        raise ValueError('Kind {} not known'.format(kind.lower()))

    return filepaths


def visu2d_filepaths(analysis_params, exp_name, major_a, minor_a, major_b, minor_b, seed=None):
    kind = analysis_params['analysis']
    window = analysis_params.get('window', None)
    window_str = '' if window is None else '_w{}_{}'.format(window[0], window[1])

    filepaths = []
    seed_str = '' if seed is None else '_seed{}'.format(seed)
    if kind.lower() == 'pca':
        if analysis_params.get('differential', None) is None:
            diff_str = ''
        else:
            diff_str = '_diff{}'.format(analysis_params['differential'])
        filename = 'pca_v{}.{}_v{}.{}{}_elites{}{}.pickle'.format(major_a, minor_a, major_b, minor_b,
                                                                  seed_str, diff_str, window_str)
        filepaths.append(Path(paths.POST_DIR, exp_name, 'visu2d', 'pca', filename))
    elif kind.lower() in ['t-sne', 'tsne']:
        perplexities = analysis_params['perplexities']
        n_iter       = analysis_params['n_iter']
        for perplexity in perplexities:
            filename = 'tsne_v{}.{}_v{}.{}{}_elites_p{}_it{}{}.pickle'.format(major_a, minor_a, major_b, minor_b,
                                                                              seed_str, perplexity, n_iter, window_str)
            filepaths.append(Path(paths.POST_DIR, exp_name, 'visu2d', 'tsne', filename))

    elif kind.lower() == 'umap':
        n_neighbors = analysis_params['n_neighbors']
        min_d       = analysis_params['min_d']
        for nn in n_neighbors:
            filename = 'umap_v{}.{}_v{}.{}{}_elites_nn{}_d{}{}.pickle'.format(major_a, minor_a, major_b,
                                                                              minor_b, seed_str, nn, min_d, window_str)
            filepaths.append(Path(paths.POST_DIR, exp_name, 'visu2d', 'umap', filename))
    else:
        raise ValueError('Kind {} not known'.format(kind.lower()))

    return filepaths

    # genotype distance

def genotype_distance_filepath(exp_name, major, minor, *args):
    return Path(paths.POST_DIR, prefix_path(exp_name, major, minor), 'genotype_distance.pickle')

def genealogy_distance_best_filepath(exp_name, major, minor, leap):
    return Path(paths.POST_DIR, prefix_path(exp_name, major, minor), 'genealogy_dist_best_leap{}.npy'.format(leap))

def parent_stability_filepath(exp_name, major, minor, *args):
    return Path(paths.POST_DIR, prefix_path(exp_name, major, minor), 'parent_stability.npy')


    # correlations

def _correlation_filepath(kind, exp_name, major, minor, window, n_elites):
    window_str = '' if window is None else '_[{}-{}]'.format(window[0], window[1])
    filename = '{}_dist_fitness{}_n{}.pickle'.format(kind, window_str, n_elites)
    return Path(prefix_path(exp_name, major, minor), 'correlations', filename)

def genotype_distance_correlation_filepath(exp_name, major, minor, window=None, n_elites=10):
    return _correlation_filepath('genotype', exp_name, major, minor, window, n_elites)

def genealogy_distance_correlation_filepath(exp_name, major, minor, window=None, n_elites=10):
    return _correlation_filepath('genealogy', exp_name, major, minor, window, n_elites)

def _correlation_double_filepath(kind, exp_name, major_a, minor_a, major_b, minor_b, window):
    window_str = '' if window is None else '_[{}-{}]'.format(window[0], window[1])
    filename = '{}_dist_fitness_v{}_v{}{}.pickle'.format(kind, version_str(major_a, minor_a),
                                                          version_str(major_b, minor_b), window_str)
    return Path(exp_name, 'correlations', filename)

def genotype_distance_correlation_double_filepath(exp_name, major_a, minor_a, major_b, minor_b, window=None):
    return _correlation_double_filepath('genotype', exp_name, major_a, minor_a, major_b, minor_b, window)

def genealogy_distance_correlation_double_filepath(exp_name, major_a, minor_a, major_b, minor_b, window=None):
    return _correlation_double_filepath('genealogy', exp_name, major_a, minor_a, major_b, minor_b, window)



    ## Misc

def scheduled_seeds(cfg):
    return array_str.str_to_indexes(cfg['meta']['slurm_array'])

def end_of_dev(perfs):
    pass

def age_suffix():
    return '_ages.npy'

def genotype_distance_suffix():
    return '_mutd.npy'

def load_mutdfile(cfg, seed, quiet=False):
    suffix = genotype_distance_suffix()

    perf_filename = prefix_path_from_cfg(cfg, seed) + suffix
    filepath = os.path.join(paths.POST_DIR, perf_filename)
    if not quiet:
        print('loading {}'.format(filepath))
    data = np.load(filepath)
    return np.array(data, dtype=np.float32)

def post_filepath(exp_name, major, minor, seed):
    cfg = load_cfg(exp_name, major, minor)
    return post_filepath_from_cfg(cfg, seed)

def post_filepath_from_cfg(cfg, seed):
    return os.path.join(paths.POST_DIR, prefix_path_from_cfg(cfg, seed))

def results_paths(exp_name, major, minor, seed):
    cfg = load_cfg(exp_name, major, minor)
    return results_paths_from_cfg(cfg, seed)

def results_paths_from_cfg(cfg, seed):
    return [Path(results_dir, prefix_path_from_cfg(cfg, seed)) for results_dir in paths.RESULTS_DIRS]

def load_results(exp_name, major, minor, seed, **kwargs):
    return evo.Run.from_results(manifest_done_path(exp_name, major, minor, seed), frozen=True, **kwargs)

def load_results_from_cfg(cfg, seed, **kwargs):
    return evo.Run.from_results(manifest_done_path_from_cfg(cfg, seed), frozen=True, **kwargs)

def move_results_to_work(cfg, seed, is_done_ok=False):
    """Move (probably half-way) results to the WORK directory, if needed"""
    seed_path = None
    for results_dir in paths.RESULTS_DIRS:
        _seed_path = Path(results_dir, prefix_path_from_cfg(cfg, seed))
        if Path(_seed_path, 'manifest.yaml').exists():
            if (not is_done_ok) and Path(_seed_path, 'done').exists():
                raise ValueError(utils.error('{} is already done'.format(_seed_path)))
            seed_path = _seed_path
            break

    if seed_path is not None and results_dir != paths.RESULTS_WORK_DIR:
        work_path = Path(paths.RESULTS_WORK_DIR, prefix_path_from_cfg(cfg, seed))
        shutil.rmtree(work_path)  # let's keep it clean
        shutil.copytree(seed_path, work_path)
        print(utils.info('copied results from {} to {}'.format(seed_path, work_path)))

def load_data(filepath):
    suffix = Path(filepath).suffix
    if suffix == '.pickle':
        with open(filepath, 'rb') as fd:
            return pickle.load(fd)
    elif suffix == '.npy':
        return np.load(filepath)
    else:
        raise ValueError('unknow extension for {}'.format(filepath))

def save_data(data, filepath, verbose=True):
    suffix = Path(filepath).suffix
    if suffix == '.pickle':
        with open(filepath, 'wb') as fd:
            pickle.dump(data, fd)
    elif suffix == '.npy':
        np.save(filepath, data)
    else:
        raise ValueError('unknow extension for {}'.format(filepath))
    if verbose:
        utils.print_info('saved {}'.format(filepath))
