import numpy as np

from . import stats


def metrics_longitudinal(data, include_data=False,
                         confidence_median=None, confidence_mean=None):
    """Expect a 2D array, compute metrics for every timestep

    :param include_data:  if True, the original data is included.
    """
    data = np.array(data, dtype=np.float32)
    m = {'mean': np.mean(data, axis=1), 'std': np.std(data, axis=1),
         'min' : np.min(data, axis=1), 'max' : np.max(data, axis=1),
         'q0.25': np.quantile(data, 0.25, axis=1),
         'q0.5' : np.median(data, axis=1),
         'q0.75': np.quantile(data, 0.75, axis=1)}

    iqr = m['q0.75'] - m['q0.25']
    m['max_iqr'] = np.minimum(np.max(data, axis=1), m['q0.75'] + 1.5 * iqr)
    m['min_iqr'] = np.maximum(np.min(data, axis=1), m['q0.25'] - 1.5 * iqr)

    m['outliers'] = []
    for data_t, max_iqr_t, min_iqr_t in zip(data, m['max_iqr'], m['min_iqr']):
        outliers_t = data_t[(data_t > max_iqr_t) | (data_t < min_iqr_t)]
        m['outliers'].append(outliers_t)

    if len(data) > 5:
        if confidence_median is not None:
            ci_lows, ci_highs = [], []
            for data_t in data:
                ci_low, ci_high = stats.confidence_interval_median(data_t, confidence=confidence_median)
                ci_lows.append(ci_low)
                ci_highs.append(ci_high)
            m.update({'median_ci_low': np.array(ci_lows), 'median_ci_high': np.array(ci_highs)})

        if confidence_mean is not None:
            ci_lows, ci_highs = [], []
            for data_t in data:
                ci_low, ci_high = stats.confidence_interval_mean(data_t, confidence=confidence_mean)
                ci_lows.append(ci_low)
                ci_highs.append(ci_high)
            m.update({'mean_ci_low': np.array(ci_lows), 'mean_ci_high': np.array(ci_highs)})

    if include_data:
        m['data'] = data

    return m


def metrics_acute(data, include_data=False,
                  confidence_median=None, confidence_mean=None, scale_factor=1.0):
    """Expect an array, compute metrics for the distribution

    :param include_data:  if True, the original data is included.
    """
    data_original = data
    data = np.array(scale_factor*np.array(data), dtype=np.float32)
    m = {'mean': np.mean(data)/scale_factor, #'std': np.std(data)/scale_factor,
         'min' : np.min(data)/scale_factor, 'max' : np.max(data)/scale_factor,
         'q0.25': np.quantile(data, 0.25)/scale_factor,
         'q0.5' : np.median(data)/scale_factor,
         'q0.75': np.quantile(data, 0.75)/scale_factor}

    iqr = (m['q0.75'] - m['q0.25'])
    m['max_iqr'] = np.minimum(np.max(data)/scale_factor, m['q0.75'] + 1.5 * iqr)
    m['min_iqr'] = np.maximum(np.min(data)/scale_factor, m['q0.25'] - 1.5 * iqr)
    m['outliers'] = [d_i for d_i in data_original if d_i > m['max_iqr'] or d_i < m['min_iqr']]

    if len(data) > 5:
        if confidence_median is not None:
            ci_low, ci_high = stats.confidence_interval_median(data, confidence=confidence_median)
            m.update({'median_ci_low': ci_low/scale_factor, 'median_ci_high': ci_high/scale_factor})

        if confidence_mean is not None:
            ci_low, ci_high = stats.confidence_interval_mean(data, confidence=confidence_mean)
            m.update({'mean_ci_low': ci_low/scale_factor, 'mean_ci_high': ci_high/scale_factor})

    if include_data:
        m['data'] = data_original

    return m
