import sys
import argparse
import itertools
import subprocess
import multiprocessing as mp
from pathlib import Path

from .. import utils, paths
from . import loadlib

from .postprocessing import compute_post, compute_visu2d, compute_visu2d_single, post_angle
from . import graphs


function_map = {'genotype_distance_metric' : (compute_post.compute_genotype_distance_metrics, 1),
                'fitness'                  : (compute_post.compute_fitness, 1),
                'elite_fitness'            : (compute_post.compute_elite_fitness, 1),
                'end_fitness'              : (compute_post.compute_end_fitness, 1),
                'visu2d_seeds'             : (compute_visu2d.compute_visu_seeds, 2),
                'visu2d_single_seeds'      : (compute_visu2d_single.compute_visu_single_seeds, 1),
                'visu2d_seeds_graph'       : (graphs.visu2d_seeds_graph, 2),
                'visu2d_seeds_graph_separate' : (graphs.visu2d_seeds_graph_separate, 2),
                'visu2d'                   : (compute_visu2d.compute_visu, 2),
                'visu2d_graph'             : (graphs.visu2d_graph, 2),
                'genotype_correlation'     : (compute_post.genotype_correlation, 1),
                'genealogy_correlation'    : (compute_post.genealogy_correlation, 1),
                'correlation_graph'        : (graphs.correlation_graph, 1),
                'correlation_double_graph' : (graphs.correlation_double_graph, 2),
                'elite_fitness_graph'      : (graphs.elite_fitness_graph, 2),
                'parent_stability'         : (compute_post.parent_stability_metrics, 1),
                'genealogy_distance_best_seeds_graphs': (graphs.genealogy_distance_best_seeds_graphs, 1),
                'genealogy_distance_best'             : (compute_post.compute_genealogy_distance_best, 1),
                'genealogy_distance_best_graph'       : (graphs.genealogy_distance_best_graph, 1),
                'genealogy_distance_best_double_graph': (graphs.genealogy_distance_best_double_graph, 2),
                'parent_stability_double_graph'       : (graphs.parent_stability_double_graph, 2),
                'angle'                               : (post_angle.compute_angle, 1),
                'angle_all'                           : (post_angle.compute_angle_all, 1),
              }

def unfold_majors(exp_name, majors_a, majors_b, excluded_majors_a=(), excluded_majors_b=()):
    if excluded_majors_a is None:
        excluded_majors_a = ()
    if excluded_majors_b is None:
        excluded_majors_b = ()

    if majors_a is not None and majors_b is not None:
        if len(majors_a) == 1:
            majors_a = len(majors_b) * majors_a
        if len(majors_b) == 1:
            majors_b = len(majors_a) * majors_b
        assert len(majors_a) == len(majors_b)
        return zip(majors_a, majors_b)

    if majors_a is None or majors_b is None:
        all_majors = loadlib.majors(exp_name)
        if majors_a is None:
            assert majors_b is not None
            selected_majors_a = sorted(set(all_majors) - set(excluded_majors_a))
            selected_majors_b = sorted(set(majors_b) - set(excluded_majors_b))
            return itertools.product(selected_majors_a, selected_majors_b)
        if majors_b is None:
            assert majors_a is not None
            selected_majors_b = sorted(set(all_majors) - set(excluded_majors_b))
            selected_majors_a = sorted(set(majors_a) - set(excluded_majors_a))
            return itertools.product(selected_majors_a, selected_majors_b)

    raise ValueError # defensive


def unfold_minors(exp_name, major_a, major_b, minors_a, minors_b):
    if minors_a is not None and minors_b is not None:
        if len(minors_a) == 1:
            minors_a = len(minors_b) * minors_a
        if len(minors_b) == 1:
            minors_b = len(minors_a) * minors_b
        assert len(minors_a) == len(minors_b)
        return zip(minors_a, minors_b)

    if minors_a is None:
        minors_a = loadlib.minors(exp_name, major_a)
    if minors_b is None:
        minors_b = loadlib.minors(exp_name, major_b)
    return itertools.product(minors_a, minors_b)


def post_compute(task_name, params, multiprocess=True):
    try:
        suffix = ' singlethreaded.'
        pool = None
        if multiprocess:
            n_process = params.get('n_process', None)
            if n_process is not None and n_process != 1:
                n_process = min(mp.cpu_count(), n_process)
                suffix = ' using {} processes.'.format(utils.bold(n_process))
                pool = mp.Pool(n_process)

        print('## starting task {}{}...'.format(task_name, suffix))

        f, n_exp = function_map[params['task']]
        # mp.Pool can't be passed between processes
        pool_arg = pool if not params.get('pool_majors', False) else None

        def add_version_to_queue_1(params, major, minor, pool):
            for majors, minors in params.get('excluded_versions', ()):
                for major_excluded in majors:
                    if major_excluded == major and (minors is None or minor in minors):
                        return
            args_list.append((params, major, minor, pool))

        def add_version_to_queue_2(params, major_a, minor_a, major_b, minor_b, pool):
            for excluded_majors_a, excluded_minors_a, excluded_majors_b, excluded_minors_b in params.get('excluded_versions', ()):
                for major_excluded in excluded_majors_a:
                    if major_excluded == major_a and (excluded_minors_a is None or minor_a in excluded_minors_a):
                        return
                for major_excluded in excluded_majors_b:
                    if major_excluded == major_b and (excluded_minors_b is None or minor_b in excluded_minors_b):
                        return
            args_list.append((params, major_a, minor_a, major_b, minor_b, pool))

        exp_name = params['exp_name']
        args_list = []
        if n_exp == 1:
            if params['versions'] is None:
                for major, minor, _ in loadlib.find_cfgs(exp_name=exp_name):
                    add_version_to_queue_1(params, major, minor, pool_arg)
            else:
                for majors, minors in params['versions']:
                    for major in majors:
                        if minors is None:
                            for major, minor, _ in loadlib.find_cfgs(exp_name, major=major):
                                add_version_to_queue_1(params, major, minor, pool)
                        else:
                            for minor in minors:
                                if loadlib.find_cfg(exp_name, major, minor, verbose=False) is None:
                                    utils.print_warning('config not found for {}:v{}.{}; skipping'.format(
                                                        exp_name, major, minor))
                                else:
                                    add_version_to_queue_1(params, major, minor, pool_arg)

        if n_exp == 2:  # comparison between two exps, only explicit major/minor are supported
            for majors_a, minors_a, majors_b, minors_b in params['versions']:
                for major_a, major_b in unfold_majors(exp_name, majors_a, majors_b):
                    for minor_a, minor_b in unfold_minors(exp_name, major_a, major_b, minors_a, minors_b):
                        add_version_to_queue_2(params, major_a, minor_a, major_b, minor_b, pool_arg)

        if pool is not None and params.get('pool_majors', False):
            pool.starmap(f, args_list)
        else:
            for args in args_list:
                f(*args)

        # for graphs
        copy_folder = params.get('copy_folder', None)
        if sys.platform == 'darwin' and copy_folder is not None:
            subprocess.check_output(['open', Path(paths.FIGURES_DIR, exp_name, 'copies', copy_folder)])


    finally:
        if pool is not None:
            pool.close()
            pool.join()
            pool.terminate()


def run_task_file(filepath, single_threaded=False, directory='', include=True):
    params = utils.load_config(Path(directory, filepath))

    for task_name, task_params in params['tasks'].items():
        post_compute(task_name, task_params, multiprocess=not single_threaded)

    if include:
        for include_path in params.get('include', ()):
            run_task_file(include_path, single_threaded=single_threaded,
                          directory=Path(directory, filepath).parent, include=True)


def post_cmd():
    global pool
    parser = argparse.ArgumentParser()
    parser.add_argument('params', help="parameters of the post processing")
    parser.add_argument('-s', '--single', help="do not parallelize processing",
                        default=False, action='store_true')
    parser.add_argument('-n', '--no_include', help="do no run tasks in secondary include files",
                        default=False, action='store_true')
    args = parser.parse_args()

    run_task_file(args.params, single_threaded=args.single, directory='', include=not args.no_include)
