import gc
import sys
import pickle
import traceback
from pathlib import Path

import numpy as np

from ... import Run, evo, utils, paths
from ..  import loadlib, metrics
from .   import utils_post



    ##  Per-seed Fitness computation  ##

def compute_fitness_seed_nosave(cfg, seed, fitness_function=None):
    run = loadlib.load_results_from_cfg(cfg, seed, verbose=False)
    perfs = evo.analysis.performances(run.population, fitness_function=fitness_function)

    def format_array(a):
        if np.finfo(np.float16).min < np.min(a) and np.max(a) < np.finfo(np.float16).max:
            return np.array(a, dtype=np.float16)
        return np.array(a, dtype=np.float32)

    if  all(len(p_i) == len(perfs[0]) for p_i in perfs):
        np_perfs = format_array(perfs)
    else:
        np_perfs = np.array([format_array(perf_row) for perf_row in perfs], dtype=object)
    return np_perfs/78.92820323027551

def compute_fitness_seed(cfg, seed, fitness_function=None):
    try:
        if not loadlib.is_done_cfg(cfg, seed):
            print('error: {} not done; skipping.'.format(filepath_done) , file=sys.stderr)
            return

        perf_filepath = loadlib.fitness_filepath_from_cfg(cfg, seed, fitness_function)
        perf_filepath.parent.mkdir(parents=True, exist_ok=True)

        if not utils_post.uptodate(cfg, [seed], perf_filepath):
            np_perfs = compute_fitness_seed_nosave(cfg, seed, fitness_function=fitness_function)
            np.save(perf_filepath, np_perfs)
            utils_post.set_mtime(cfg, [seed], perf_filepath)
            utils.print_info('saved  {}'.format(perf_filepath))
    except Exception as exc:
        traceback.print_exc()
        utils.print_error('error encountered when computing fitness file for v{}.{}, seed {}'.format(
                          cfg['meta']['major'], cfg['meta']['minor'], seed))

def compute_fitness(params, major, minor, pool=None):
    cfg = loadlib.load_cfg(params['exp_name'], major, minor)
    fitness_function = params['fitness_function']

    args_list = []
    seeds_done = loadlib.find_done_cfg(cfg)
    seeds = params.get('seeds', None)
    if seeds is not None:
        seeds_done = sorted(set(seeds_done).intersection(seeds))
    for seed in seeds_done:
        args_list.append((cfg, seed, fitness_function))

    if pool is not None:
        pool.starmap(compute_fitness_seed, args_list)
    else:
        for args in args_list:
            compute_fitness_seed(*args)


        ## End fitness computation ##

def get_end_fitness(cfg, seed, fitness_function=None, nosave=False):
    if nosave:
        perfs = compute_fitness_seed_nosave(cfg, seed, fitness_function=fitness_function)
    else:
        perfs = loadlib.load_perffile(cfg, seed, fitness_function=fitness_function, verbose=False)
    return np.mean(np.sort(np.stack(perfs[-50:]).flatten())[-10:])

def compute_end_fitness(params, major, minor, pool=None):
    """Compute the distribution of the best performers across timesteps and seeds"""
    nosave = True

    cfg = loadlib.load_cfg(params['exp_name'], major, minor)
    if cfg['meta'].get('disabled', False):
        return
    fitness_function = params['fitness_function']

    args_list = []
    if nosave:
        seeds = loadlib.find_done_cfg(cfg)
    else:
        seeds = loadlib.find_done_perfs_cfg(cfg)
    if len(seeds) < 5:
        utils.print_warning('v{}.{}: not enough datapoints ({}) to compute distribution, skipping'.format(
                            major, minor, len(seeds)))
        return

    filepath = loadlib.end_fitness_filepath_from_cfg(cfg, fitness_function=fitness_function)
    if not utils_post.uptodate(cfg, seeds, filepath):
        for seed in seeds:
            args_list.append((cfg, seed, fitness_function, nosave))

        if pool is not None:
            end_fitnesses = pool.starmap(get_end_fitness, args_list)
        else:
            end_fitnesses = []
            for args in args_list:
                end_fitnesses.append(get_end_fitness(*args))

        end_fitnesses = np.array(end_fitnesses).T
        filepath.parent.mkdir(parents=True, exist_ok=True)
        with open(filepath, 'wb') as fd:
            pickle.dump((seeds, end_fitnesses), fd)
        # set_mtime(cfg, [seed], filepath)
        utils_post.set_mtime_fitness(cfg, seeds, filepath)
        print(utils.info('saved  {}'.format(filepath)))



        ##  Elite Fitness computation  ##


def get_elites_fitness(cfg, seed, fitness_function=None, nosave=False):
    if nosave:
        perfs = compute_fitness_seed_nosave(cfg, seed, fitness_function=fitness_function)
    else:
        perfs = loadlib.load_perffile(cfg, seed, fitness_function=fitness_function, verbose=False)
    return np.array([np.max(perfs_t) for perfs_t in perfs])

def compute_elite_fitness(params, major, minor, pool=None):
    """Compute the distribution of the best performers across timesteps and seeds"""
    nosave = True

    cfg = loadlib.load_cfg(params['exp_name'], major, minor)
    if cfg['meta'].get('disabled', False):
        return
    fitness_function = params['fitness_function']

    args_list = []
    if nosave:
        seeds = loadlib.find_done_cfg(cfg)
    else:
        seeds = loadlib.find_done_perfs_cfg(cfg)
    if len(seeds) < 5:
        utils.print_warning('v{}.{}: not enough datapoints ({}) to compute distribution, skipping'.format(
                            major, minor, len(seeds)))
        return

    filepath = loadlib.elite_fitness_filepath_from_cfg(cfg, params['confidence'], fitness_function=fitness_function)

    if not (utils_post.uptodate if nosave else utils_post.uptodate_fitness)(cfg, seeds, filepath):
        for seed in seeds:
            args_list.append((cfg, seed, fitness_function, nosave))

        if pool is not None:
            elite_fitnesses = pool.starmap(get_elites_fitness, args_list)
        else:
            elite_fitnesses = []
            for args in args_list:
                elite_fitnesses.append(get_elites_fitness(*args))
        elite_fitnesses = np.array(elite_fitnesses).T

        metrics_data = metrics.metrics_longitudinal(elite_fitnesses, include_data=False,
                                                    confidence_median=params['confidence'],
                                                    confidence_mean=params['confidence'])

        filepath.parent.mkdir(parents=True, exist_ok=True)
        with open(filepath, 'wb') as fd:
            pickle.dump(metrics_data, fd)
        # set_mtime(cfg, [seed], filepath)
        utils_post.set_mtime_fitness(cfg, seeds, filepath)
        print(utils.info('saved  {}'.format(filepath)))


        ## Genealogy/Genealogy distance/Fitness Correlation ##


def get_genealogy_distance(args):
    exp_name, major, minor, seed, window, fitness_window, n_elites = args
    frozen_run = loadlib.load_results(exp_name, major, minor, seed, verbose=False)

    best = evo.analysis.select_best(frozen_run.population, n=n_elites, window=fitness_window)
    members, distances = evo.analysis.genealogy_distance(frozen_run.population, window=window)

    uid_map = {member.uid: distance for member, distance in zip(members, distances)}
    mean_fitness  = np.mean([b.fitness for b in best])
    mean_distance = np.mean([uid_map[b.uid] for b in best])
    return mean_fitness, mean_distance, seed

def get_genotype_distance(args):
    exp_name, major, minor, seed, window, fitness_window, n_elites = args
    frozen_run = loadlib.load_results(exp_name, major, minor, seed, verbose=False)

    best = evo.analysis.select_best(frozen_run.population, n=n_elites, window=fitness_window)
    generations, distances = evo.analysis.genotype_distance_pop(frozen_run.population, window=window)

    uid_map = {}
    for gen, d_gen in zip(generations, distances):
        for uid, d_rooted in zip(gen, d_gen):
            uid_map[uid] = d_rooted[1]
    mean_fitness  = np.mean([b.fitness for b in best])
    mean_distance = np.mean([uid_map[b.uid] for b in best])
    return mean_fitness, mean_distance, seed


def genealogy_correlation(params, major, minor, pool):
    return _correlation(params, major, minor, pool, loadlib.genealogy_distance_correlation_filepath, get_genealogy_distance)

def genotype_correlation(params, major, minor, pool):
    return _correlation(params, major, minor, pool, loadlib.genotype_distance_correlation_filepath, get_genotype_distance)


def _correlation(params, major, minor, pool, filepath_fun, get_fun):
    exp_name = params['exp_name']
    window, n_elites = params.get('window', None), params['n_elites']
    fitness_window = params.get('fitness_window', (-50, None))
    filepath = filepath_fun(exp_name, major, minor, window, n_elites)
    post_filepath = Path(paths.POST_DIR, filepath)

    cfg = loadlib.load_cfg(exp_name, major, minor)
    seeds = loadlib.find_done(exp_name, major, minor)
    if len(seeds) > 0:
        if utils_post.uptodate(cfg, seeds, post_filepath):
            # utils.print_info('{} exists; skipping'.format(post_filepath))
            return

        args_list = [(exp_name, major, minor, seed, window, fitness_window, n_elites) for seed in seeds]
        if pool is not None:
            fit_distances = pool.map(get_fun, args_list)
        else:
            fit_distances = [get_fun(args) for args in args_list]

        post_filepath.parent.mkdir(parents=True, exist_ok=True)
        with open(post_filepath, 'wb') as fd:
            pickle.dump(fit_distances, fd)
            utils_post.set_mtime(cfg, seeds, post_filepath)
            utils.print_info('saved {}'.format(filepath))

    gc.collect()



        ##  Genotype distance  ##


def _genotype_distance_seed(exp_name, major, minor, seed, leap=None):
    """Return the mean genotype distance for each generation"""
    run = loadlib.load_results(exp_name, major, minor, seed)
    uids, dists = evo.analysis.genotype_distance_pop(run.population)
    return np.array([np.mean(dist_i) for dist_i in dists])

def _genealogy_distance_best_seed(exp_name, major, minor, seed, leap):
    frozen_run = loadlib.load_results(exp_name, major, minor, seed, verbose=False)

    best = evo.analysis.select_best(frozen_run.population, n=1, window=(-1, None))[0]
    genealogy = evo.analysis.genealogy(frozen_run.population, best)
    genealogy = [frozen_run.population.repository.member(uid) for uid in genealogy]

    return [evo.analysis.genotype_distance(genealogy[i], genealogy[i+leap]) for i in range(len(genealogy) - 1 - leap, -1, -1)]

def _parent_stability_seed(exp_name, major, minor, seed, leap):
    frozen_run = loadlib.load_results(exp_name, major, minor, seed, verbose=False)

    pop = frozen_run.population
    stability = [0.0]  # first generation has neutral stability.
    for generation in pop.generations[1:]:
        fitness, fitness_diff = [], []
        for uid in generation:
            m      = pop.repository.member(uid)
            parent = pop.repository.member(m.puid)
            if parent.genotype == m.genotype:
                fitness.append(abs(parent.fitness))
                fitness.append(abs(m.fitness))
                fitness_diff.append(abs(parent.fitness - m.fitness))
        s = np.mean(fitness_diff) / np.mean(fitness)
        stability.append(s)

    return stability


def compute_genealogy_distance_best(params, major, minor, pool=None):
    return _stack_metrics(loadlib.genealogy_distance_best_filepath, _genealogy_distance_best_seed,
                          params, major, minor, pool=pool)

def compute_genotype_distance_metrics(params, major, minor, pool=None):
    return _stack_metrics(loadlib.genotype_distance_filepath, _genotype_distance_seed,
                          params, major, minor, pool=pool)

def parent_stability_metrics(params, major, minor, pool=None):
    return _stack_metrics(loadlib.parent_stability_filepath, _parent_stability_seed,
                          params, major, minor, pool=pool)

def _stack_metrics(filepath_fun, get_fun, params, major, minor, pool=None):
    try:
        exp_name, leap = params['exp_name'], params.get('leap', 1)
        cfg = loadlib.load_cfg(exp_name, major, minor)
        if cfg['meta'].get('disabled', False):
            return

        # check that not already computed
        filepath = filepath_fun(exp_name, major, minor, leap)
        filepath.parent.mkdir(parents=True, exist_ok=True)

        seeds = loadlib.find_done(exp_name, major, minor, cfg=cfg)

        if not utils_post.uptodate(cfg, seeds, filepath):
            # compute individual seeds distances
            args_list = []
            for seed in seeds:
                args_list.append((exp_name, major, minor, seed, leap))

            if pool is not None:
                distances = pool.starmap(get_fun, args_list)
            else:
                distances = [get_fun(*args) for args in args_list]

            if len(distances) == 0:
                utils.print_warning('no data for {}:v{}.{}'.format(exp_name, major, minor))
                return

            if np.max(distances) < np.finfo(np.float16).max:
                distances = np.array(distances, dtype=np.float16)
            else:
                distances = np.array(distances)
            loadlib.save_data(distances, filepath, verbose=True)

    except Exception as exc:
        traceback.print_exc()
        utils.print_error('error encountered when trying to compute {}'.format(filepath))
