"""
Dimension reduction computations: PCA, t-SNE, UMAP

"""
import os
import time
import pickle
import traceback
import multiprocessing as mp

os.environ['OMP_NUM_THREADS'] = '5'

import numpy as np
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
#import umap

from ... import paths, Member, evo, utils
from ..  import loadlib
from .   import utils_post



def save(results, filepath):
    filepath.parent.mkdir(parents=True, exist_ok=True)
    with open(filepath, 'wb') as fd:
        pickle.dump(results, fd)
        utils.print_info('saved {}'.format(filepath))


def extract_data(args):
    exp_name, major, minor, seed, first_pop, run_id = args

    run = loadlib.load_results(exp_name, major, minor, seed)
    assert run.population.generations[0] == first_pop
    best = evo.analysis.select_best(run.population, n=1, window=[3000, 4000])[0]
    return best.genotype, best.fitness, run_id


def prepare_data(exp_name, major_a, minor_a, major_b, minor_b, load_n_process=None):
    done_a = loadlib.find_done(exp_name, major_a, minor_a)
    done_b = loadlib.find_done(exp_name, major_b, minor_b)
    done = set(done_a).intersection(set(done_b))

    data, exps, fitnesses = [], [], []
    if len(done) == 0:
        return data, exps, fitnesses

    first_pop = None

    run = loadlib.load_results(exp_name, major_a, minor_a, list(done)[0])
    if first_pop is None:
        first_pop = run.population.generations[0]
        if len(first_pop) == 1:
            uid = first_pop[0]
            member = run.population.repository.member(uid)
            data.append(member.genotype)
            fitnesses.append(member.fitness)
            exps.append(-1)

    args_list = []
    for seed in done:
        args_list.append((exp_name, major_a, minor_a, seed, first_pop, 0))
        args_list.append((exp_name, major_b, minor_b, seed, first_pop, 1))

    if load_n_process is not None:
        try:
            pool = mp.Pool(load_n_process)
            for genotype, fitness, run_id in pool.map(extract_data, args_list):
                data.append(genotype)
                fitnesses.append(fitness)
                exps.append(run_id)
        finally:
                pool.close()
                pool.join()
                pool.terminate()
                pool = None
    else:
        for args in args_list:
            genotype, fitness, run_id = extract_data(args)
            data.append(genotype)
            fitnesses.append(fitness)
            exps.append(run_id)

    return np.array(data), exps, np.array(fitnesses)


def compute_visu_(exp_name, major_a, minor_a, major_b, minor_b, analysis_params):
    if analysis_params.get('differential', None) is not None:
        raise NotImplementedError

    np.random.seed(1)

    kind = analysis_params['analysis']
    cfgs = [loadlib.load_cfg(exp_name, major_a, minor_a), loadlib.load_cfg(exp_name, major_b, minor_b)]

    # create filepaths and check if the computation is done already
    filepaths = loadlib.visu2d_filepaths(analysis_params, exp_name, major_a, minor_a, major_b, minor_b)
    if kind.lower() == 'pca':
        if utils_post.uptodate_intersection_done(cfgs, filepaths[0]):
            return
    elif kind.lower() in ['t-sne', 'tsne']:
        perplexities = analysis_params['perplexities']
        filtered = [(perplexity, filepath) for perplexity, filepath in zip(perplexities, filepaths)
                    if not utils_post.uptodate_intersection_done(cfgs, filepath)]
        if len(filtered) == 0: return
        perplexities, filepaths = zip(*filtered)
    elif kind.lower() == 'umap':
        n_neighbors = analysis_params['n_neighbors']
        filtered = [(nn, filepath) for nn, filepath in zip(n_neighbors, filepaths)
                    if not utils_post.uptodate_intersection_done(cfgs, filepath)]
        if len(filtered) == 0: return
        n_neighbors, filepaths = zip(*filtered)

    data, exps, fitnesses = prepare_data(exp_name, major_a, minor_a, major_b, minor_b,
                                         load_n_process=analysis_params.get('load_n_process', None))

    if len(data) == 0:
        utils.print_warning('no finished run found for v{}.{}-v{}.{} combination; skipping'.format(
                            major_a, minor_a, major_b, minor_b))
        return

    time_start = time.time()

    if kind.lower() == 'pca':
        pca = PCA(n_components=2)
        results = pca.fit_transform(data)
        filepath = filepaths[0]
        print('Explained variation per principal component: {}'.format(pca.explained_variance_ratio_))
        print('{} done in {:.1f} seconds'.format(kind, time.time()-time_start))
        metadata = {'explained_variance_ratio': pca.explained_variance_ratio_}
        save((results, exps, fitnesses, metadata), filepath)
        utils_post.set_mtime_intersection_done(cfgs, filepath)
    elif kind.lower() in ['t-sne', 'tsne']:
        n_iter = analysis_params['n_iter']
        for perplexity, filepath in zip(perplexities, filepaths):
            prefix = 'tsne'
            tsne = TSNE(n_components=2, verbose=0, perplexity=perplexity, n_iter=n_iter, n_jobs=1)
            results = tsne.fit_transform(data)
            print('{} done in {:.1f} seconds'.format(kind, time.time()-time_start))
            metadata = {}
            save((results, exps, fitnesses, metadata), filepath)
            utils_post.set_mtime_intersection_done(cfgs, filepath)
    elif kind.lower() == 'umap':
        min_d = analysis_params['min_d']
        for nn, filepath in zip(n_neighbors, filepaths):
            reducer = umap.UMAP(n_components=2, n_neighbors=nn, min_dist=min_d)
            results = reducer.fit_transform(data)
            print('{} done in {:.1f} seconds'.format(kind, time.time()-time_start))
            metadata = {}
            save((results, exps, fitnesses, metadata), filepath)
            utils_post.set_mtime_intersection_done(cfgs, filepath)
    else:
        raise ValueError('"{}" analysis not recognized.'.format(kind))


def compute_visu(params, major_a, minor_a, major_b, minor_b, pool=None):
    return compute_visu_(params['exp_name'], major_a, minor_a, major_b, minor_b, params['analysis_params'])


def prepare_data_seed(exp_name, major_a, minor_a, major_b, minor_b, seed, analysis_params):
    genealogy = analysis_params.get('genealogy', False)
    run_a = loadlib.load_results(exp_name, major_a, minor_a, seed)
    run_b = loadlib.load_results(exp_name, major_b, minor_b, seed)
    genotypes, generations, elite_flag, parent_index, fitnesses, run_index = [], [], [], [], [], []

    for run_i, run in enumerate([run_a, run_b]):
        # compute the uids of the elites
        _dataset, elite_uids = set(), []
        n_elites = analysis_params.get('n_elites', 1)
        for t_gen, gen in enumerate(run.population.generations):
            fitted_states = [(state[Member.fitness_index], state)
                           for state in [run.population.repository.member_state(uid) for uid in gen]]

            for fitness, state in sorted(fitted_states, reverse=True)[:n_elites]:
                genotype = tuple(state[Member.genotype_index])
                if genotype not in _dataset:  # not adding the same genotype twice
                    _dataset.add(genotype)
                    elite_uids.append(state[Member.uid_index])

        elite_uids = set(elite_uids)
        if genealogy:
            m_states, parents = evo.analysis.population_genealogy_subset(run.population, elite_uids)
            offset = len(genotypes)

            for state, parent in zip(m_states, parents):
                uid = state[Member.uid_index]
                genotypes.append(state[Member.genotype_index])
                generations.append(state[Member.generation_index])
                fitnesses.append(state[Member.fitness_index])
                parent_index.append(None if parent is None else parent + offset)
                elite_flag.append(uid in elite_uids)
                run_index.append(run_i)
        else:
            for uid in elite_uids:
                state = run.population.repository.member_state(uid)
                genotypes.append(state[Member.genotype_index])
                generations.append(state[Member.generation_index])
                fitnesses.append(state[Member.fitness_index])
                parent_index.append(None)
                elite_flag.append(True)
                run_index.append(run_i)

    return np.array(genotypes), generations, fitnesses, elite_flag, parent_index, run_index


def compute_visu_seed(exp_name, major_a, minor_a, major_b, minor_b, seed, analysis_params):
    try:
        np.random.seed(1)

        # create filepaths and check if the computation is done already
        kind = analysis_params['analysis']
        cfgs = [loadlib.load_cfg(exp_name, major_a, minor_a), loadlib.load_cfg(exp_name, major_b, minor_b)]
        filepaths = loadlib.visu2d_filepaths(analysis_params, exp_name, major_a, minor_a, major_b, minor_b, seed)
        if kind.lower() == 'pca':
            if utils_post.uptodate_intersection_done(cfgs, filepaths[0]):
                return
        elif kind.lower() in ['t-sne', 'tsne']:
            perplexities = analysis_params['perplexities']
            filtered = [(perplexity, filepath) for perplexity, filepath in zip(perplexities, filepaths)
                        if not utils_post.uptodate_intersection_done(cfgs, filepath)]
            if len(filtered) == 0: return
            perplexities, filepaths = zip(*filtered)
        elif kind.lower() == 'umap':
            n_neighbors = analysis_params['n_neighbors']
            filtered = [(nn, filepath) for nn, filepath in zip(n_neighbors, filepaths)
                        if not utils_post.uptodate_intersection_done(cfgs, filepath)]
            if len(filtered) == 0: return
            n_neighbors, filepaths = zip(*filtered)

        # defensive
        if len(filepaths) == 0:
            return

        data = prepare_data_seed(exp_name, major_a, minor_a, major_b, minor_b, seed,
                                  analysis_params)
        genotypes, generations, fitnesses, elite_flag, parent_index, run_index = data
        results = {'generations': generations, 'fitnesses': fitnesses, 'elite_flag': elite_flag,
                   'parent_index': parent_index, 'run_index': run_index, 'metadata': {},
                   'analysis_params': analysis_params}

        # PCA
        if kind.lower() == 'pca':
            pca      = PCA(n_components=2)
            if analysis_params.get('differential', None) is None:
                results['map2d'] = pca.fit_transform(genotypes)
            else:
                genotypes_fit = [g for g, index in zip(genotypes, run_index)
                                 if index == analysis_params['differential']]
                pca.fit(genotypes_fit)
                results['map2d'] = pca.transform(genotypes)

            filepath = filepaths[0]
            print('Explained variation per principal component: {}'.format(pca.explained_variance_ratio_))
            results['metadata'] = {'explained_variance_ratio': pca.explained_variance_ratio_}
            save(results, filepath)
            utils_post.set_mtime_intersection_done(cfgs, filepath)

        # t-SNE  (params are `perplexities` and `n_iter`)
        elif kind.lower() in ['t-sne', 'tsne']:
            n_iter       = analysis_params['n_iter']
            for perplexity, filepath in zip(perplexities, filepaths):
                time_start = time.time()
                tsne = TSNE(n_components=2, verbose=0, perplexity=perplexity, n_iter=n_iter, n_jobs=1)
                results['map2d'] = tsne.fit_transform(genotypes)
                print('{} done in {:.1f} seconds'.format(kind, time.time()-time_start))
                save(results, filepath)
                utils_post.set_mtime_intersection_done(cfgs, filepath)

        # UMAP  (params are `n_neighbors` and `min_d`)
        elif kind.lower() == 'umap':
            min_d       = analysis_params['min_d']
            for nn, filepath in zip(n_neighbors, filepaths):
                time_start = time.time()
                reducer = umap.UMAP(n_components=2, n_neighbors=nn, min_dist=min_d)
                results['map2d'] = reducer.fit_transform(genotypes)
                print('{} done in {:.1f} seconds'.format(kind, time.time()-time_start))
                save(filepath, results)
                utils_post.set_mtime_intersection_done(cfgs, filepath)

        else:
            raise ValueError('"{}" analysis not recognized.'.format(kind))
    except Exception:
        traceback.print_exc()
        utils.print_error('error encountered in computing visu2d {} v{}.{}-v{}.{}'.format(
                          kind,  major_a, minor_a, major_b, minor_b))


def compute_visu_seeds(params, major_a, minor_a, major_b, minor_b, pool=None):
    exp_name = params['exp_name']
    cfg_a = loadlib.load_cfg(exp_name, major_a, minor_a)
    seeds_a = loadlib.find_done_cfg(cfg_a)
    cfg_b = loadlib.load_cfg(exp_name, major_a, minor_a)
    seeds_b = loadlib.find_done_cfg(cfg_b)
    seeds = set(seeds_a).intersection(set(seeds_b))

    if params.get('seeds', None) is not None:
        seeds = seeds.intersection(set(range(params['seeds'][0], params['seeds'][1])))
    seeds = list(sorted(seeds))
    if len(seeds) == 0:
        print(utils.warning('No seed with enough data found for analysis'))

    args_list = []
    for seed in seeds:
        args_list.append((exp_name, major_a, minor_a, major_b, minor_b, seed, params['analysis_params']))

    if pool is not None:
        pool.starmap(compute_visu_seed, args_list)
    else:
        for args in args_list:
            compute_visu_seed(*args)
