"""
Dimension reduction computations: PCA, t-SNE, UMAP

"""
import os
import time
import pickle
import traceback
import multiprocessing as mp

os.environ['OMP_NUM_THREADS'] = '5'

import numpy as np
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
#import umap

from ... import paths, Member, evo, utils
from ..  import loadlib
from .   import utils_post



def save(results, filepath):
    filepath.parent.mkdir(parents=True, exist_ok=True)
    with open(filepath, 'wb') as fd:
        pickle.dump(results, fd)
        utils.print_info('saved {}'.format(filepath))



def prepare_data_single_seed(exp_name, major, minor, seed, analysis_params):
    genealogy = analysis_params.get('genealogy', False)
    window = analysis_params.get('window', None)
    run = loadlib.load_results(exp_name, major, minor, seed)
    genotypes, generations, elite_flag, parent_index, fitnesses = [], [], [], [], []

    # compute the uids of the elites
    _dataset, elite_uids = set(), []
    n_elites = analysis_params.get('n_elites', 1)
    gens = run.population.generations
    if window is not None:
        gens = gens[window[0]:window[1]]

    for t_gen, gen in enumerate(gens):
        fitted_states = [(state[Member.fitness_index], state)
                         for state in [run.population.repository.member_state(uid) for uid in gen]]

        for fitness, state in sorted(fitted_states, reverse=True)[:n_elites]:
            genotype = tuple(state[Member.genotype_index])
            if genotype not in _dataset:  # not adding the same genotype twice
                _dataset.add(genotype)
                elite_uids.append(state[Member.uid_index])

    elite_uids = set(elite_uids)
    if genealogy:
        m_states, parents = evo.analysis.population_genealogy_subset(run.population, elite_uids, window=window)
        offset = len(genotypes)

        for state, parent in zip(m_states, parents):
            uid = state[Member.uid_index]
            genotypes.append(state[Member.genotype_index])
            generations.append(state[Member.generation_index])
            fitnesses.append(state[Member.fitness_index])
            parent_index.append(None if parent is None else parent + offset)
            elite_flag.append(uid in elite_uids)

    else:
        for uid in elite_uids:
            state = run.population.repository.member_state(uid)
            genotypes.append(state[Member.genotype_index])
            generations.append(state[Member.generation_index])
            fitnesses.append(state[Member.fitness_index])
            parent_index.append(None)
            elite_flag.append(True)

    return np.array(genotypes), generations, fitnesses, elite_flag, parent_index



def compute_visu_single_seed(exp_name, major, minor, seed, analysis_params):
    try:
        np.random.seed(1)

        # create filepaths and check if the computation is done already
        kind = analysis_params['analysis']
        cfgs = [loadlib.load_cfg(exp_name, major, minor)]
        filepaths = loadlib.visu2d_filepaths_single(analysis_params, exp_name, major, minor, seed)
        if kind.lower() == 'pca':
            if utils_post.uptodate_intersection_done(cfgs, filepaths[0]):
                return
        elif kind.lower() in ['t-sne', 'tsne']:
            perplexities = analysis_params['perplexities']
            filtered = [(perplexity, filepath) for perplexity, filepath in zip(perplexities, filepaths)
                        if not utils_post.uptodate_intersection_done(cfgs, filepath)]
            if len(filtered) == 0: return
            perplexities, filepaths = zip(*filtered)
        elif kind.lower() == 'umap':
            n_neighbors = analysis_params['n_neighbors']
            filtered = [(nn, filepath) for nn, filepath in zip(n_neighbors, filepaths)
                        if not utils_post.uptodate_intersection_done(cfgs, filepath)]
            if len(filtered) == 0: return
            n_neighbors, filepaths = zip(*filtered)

        # defensive
        if len(filepaths) == 0:
            return

        data = prepare_data_single_seed(exp_name, major, minor, seed, analysis_params)
        genotypes, generations, fitnesses, elite_flag, parent_index = data
        results = {'generations': generations, 'fitnesses': fitnesses, 'elite_flag': elite_flag,
                   'parent_index': parent_index, 'metadata': {}, 'analysis_params': analysis_params}

        # PCA
        if kind.lower() == 'pca':
            pca      = PCA(n_components=2)
            results['map2d'] = pca.fit_transform(genotypes)

            filepath = filepaths[0]
            print('Explained variation per principal component: {}'.format(pca.explained_variance_ratio_))
            results['metadata'] = {'explained_variance_ratio': pca.explained_variance_ratio_}
            save(results, filepath)
            utils_post.set_mtime_intersection_done(cfgs, filepath)

        # t-SNE  (params are `perplexities` and `n_iter`)
        elif kind.lower() in ['t-sne', 'tsne']:
            n_iter       = analysis_params['n_iter']
            for perplexity, filepath in zip(perplexities, filepaths):
                time_start = time.time()
                tsne = TSNE(n_components=2, verbose=0, perplexity=perplexity, n_iter=n_iter, n_jobs=1)
                results['map2d'] = tsne.fit_transform(genotypes)
                print('{} done in {:.1f} seconds'.format(kind, time.time()-time_start))
                save(results, filepath)
                utils_post.set_mtime_intersection_done(cfgs, filepath)

        # UMAP  (params are `n_neighbors` and `min_d`)
        elif kind.lower() == 'umap':
            min_d       = analysis_params['min_d']
            for nn, filepath in zip(n_neighbors, filepaths):
                time_start = time.time()
                reducer = umap.UMAP(n_components=2, n_neighbors=nn, min_dist=min_d)
                results['map2d'] = reducer.fit_transform(genotypes)
                print('{} done in {:.1f} seconds'.format(kind, time.time()-time_start))
                save(filepath, results)
                utils_post.set_mtime_intersection_done(cfgs, filepath)

        else:
            raise ValueError('"{}" analysis not recognized.'.format(kind))
    except Exception:
        traceback.print_exc()
        utils.print_error('error encountered in computing visu2d {} v{}.{}'.format(kind, major, minor))



def compute_visu_single_seeds(params, major, minor, pool=None):
    exp_name = params['exp_name']
    cfg   = loadlib.load_cfg(exp_name, major, minor)
    seeds = loadlib.find_done_cfg(cfg)

    if params.get('seeds', None) is not None:
        seeds = set(seeds).intersection(params['seeds'])
    seeds = list(sorted(seeds))
    if len(seeds) == 0:
        print(utils.warning('No seed with enough data found for analysis'))

    args_list = []
    for seed in seeds:
        args_list.append((exp_name, major, minor, seed, params['analysis_params']))

    if pool is not None:
        pool.starmap(compute_visu_single_seed, args_list)
    else:
        for args in args_list:
            compute_visu_single_seed(*args)
