import sys
import math
from pathlib import Path

import scipy
import numpy as np
import pandas as pd

from ... import utils, paths
from .. import loadlib, funlib



        ## Max fitness Over All Seeds ##

def prepare_data_gen_fitness_run(exp_name, major, minor, n_gen, fitness_function=None):
    maximums, _warnings = [], []

    cfg = loadlib.load_cfg(exp_name, major, minor)
    if cfg is None:
        utils.print_warning('no config found for {}:{}.{}; skipping'.format(exp_name, major, minor))
        return None

    all_perfs = []
    # print(set(range(1, 101)) - set(loadlib.find_done_perfs_cfg(cfg)))
    for seed in loadlib.find_done_perfs_cfg(cfg):
        try:
            perfs = loadlib.load_perffile(cfg, seed, fitness_function=fitness_function, verbose=False)[n_gen]
            perfs = perfs
            all_perfs.extend(perfs)
        except FileNotFoundError:
            _warnings.append(utils.warning('v{}.{}.{} not found; skipping.'.format(major, minor, seed)))

    if len(all_perfs) == 0:
        for warning in _warnings:
            print('  ' + warning, file=sys.stderr)
        utils.print_warning('no data for exp v{}.{}; skipping.'.format(major, minor))
        return None

    # export csv
    # csv_filepath = Path(paths.POST_DIR, loadlib.prefix_path(exp_name, major, minor), 'elites_fitness.csv')
    # df = pd.DataFrame(data={'fitness': maximums})
    # df.to_csv(csv_filepath)
    # utils.print_info('saved {}'.format(csv_filepath))

    return np.array(all_perfs)



def prepare_data_max_fitness_run(exp_name, major, minor, f_perf, fitness_function=None, rotation=None):
    maximums, _warnings = [], []

    cfg = loadlib.load_cfg(exp_name, major, minor)
    if cfg is None:
        utils.print_warning('no config found for {}:{}.{}; skipping'.format(exp_name, major, minor))
        return (), None
    # print(set(range(1, 101)) - set(loadlib.find_done_perfs_cfg(cfg)))
    seeds = []

    results = None
    end_filepath = loadlib.end_fitness_filepath_from_cfg(cfg, fitness_function=fitness_function)
    if end_filepath.exists():
        seeds, fitnesses = utils.load_data(end_filepath)
        results = {seed: fitness for seed, fitness in zip(seeds, fitnesses)}

    potential_seeds = seeds[:] if len(seeds) > 0 else loadlib.find_done_cfg(cfg)

    if rotation in ['none', 'only']:
        filepath = loadlib.angle_filepath_from_cfg(cfg)
        data_angles = utils.load_data(filepath)
        angles = [abs(data_angles[seed][0])/2/math.pi for seed in potential_seeds]
        if rotation == 'none':
            potential_seeds = [seed for seed, a in zip(potential_seeds, angles) if a <= 2]
        else:
            potential_seeds = [seed for seed, a in zip(potential_seeds, angles) if a  > 2]

    seeds = []
    for seed in potential_seeds:
        if results is not None:
            maximums.append(results[seed])
            seeds.append(seed)
        else:
            try:
                perfs = loadlib.load_perffile(cfg, seed, fitness_function=fitness_function, verbose=False)
                perfs = perfs
                try:
                    maximums.append(f_perf(perfs))
                    seeds.append(seed)
                except FloatingPointError:
                    _warnings.append(utils.warning('v{}.{}.{} was found but did not return any data; skipping.'.format(major, minor, seed)))
            except FileNotFoundError:
                _warnings.append(utils.warning('v{}.{}.{} not found; skipping.'.format(major, minor, seed)))

    if len(maximums) == 0:
        for warning in _warnings:
            print('  ' + warning, file=sys.stderr)
        utils.print_warning('no data for exp v{}.{}; skipping.'.format(major, minor))
        return (), None

    # export csv
    # csv_filepath = Path(paths.POST_DIR, loadlib.prefix_path(exp_name, major, minor), 'elites_fitness.csv')
    # df = pd.DataFrame(data={'fitness': maximums})
    # df.to_csv(csv_filepath)
    # utils.print_info('saved {}'.format(csv_filepath))

    return seeds, np.array(maximums)



        ## Significance ##


def significance(exp_name, major_a, minor_a, major_b, minor_b, perf_fun=funlib.fitness_3950_4000):
    seeds_a, maximums_a = prepare_data_max_fitness_run(exp_name, major_a, minor_a, perf_fun)
    seeds_b, maximums_b = prepare_data_max_fitness_run(exp_name, major_b, minor_b, perf_fun) # funlib.fitness_0_4000)
    # if len(seeds_a) == 101:
    #     assert len(seeds_a)==len(maximums_a)
    #     seeds_a = seeds_a[1:]
    #     maximums_a = maximums_a[1:]
    # if len(seeds_b) == 101:
    #     seeds_b = seeds_b[1:]
    #     maximums_b = maximums_b[1:]
    if len(maximums_a) != len(maximums_b):
        seeds = set(seeds_a).intersection(seeds_b)
        maximums_a = np.array([max_a for seed, max_a in zip(seeds_a, maximums_a) if seed in seeds])
        maximums_b = np.array([max_b for seed, max_b in zip(seeds_b, maximums_b) if seed in seeds])
    if maximums_a is None or maximums_b is None or len(maximums_a) == 0:
        if maximums_a is None:
            utils.print_warning('v{}.{}: no data, skipping significance computation'.format(major_a, minor_a))
        if maximums_b is None:
            utils.print_warning('v{}.{}: no data, skipping significance computation'.format(major_b, minor_b))
        else:
            utils.print_warning('v{}.{}-v{}.{}: no common seeds, skipping significance computation'.format(major_a, minor_a, major_b, minor_b))
        return

        # utils.print_warning('v{}.{}: {} datapoints vs v{}.{}: {} datapoints'.format(
        #                     major_a, minor_a, len(maximums_a),
        #                     major_b, minor_b, len(maximums_b)))
        # return
    # testing normality with Shapiro-Wilk
    W, p_norm = scipy.stats.shapiro(maximums_a - maximums_b)
    # Paired t-test
    # t_value, p_value = scipy.stats.ttest_rel(maximums_a, maximums_b)
    # Signed-Rank Wilcoxon
    w_value, p_value = scipy.stats.wilcoxon(maximums_a, maximums_b)
    stars = 'n.s.'
    if p_value < 0.05:  stars = '*'
    if p_value < 0.01:  stars = '**'
    if p_value < 0.001: stars = '***'

    print('{}:v{}.{:02d}-v{}.{:02d}:   w-value: {:10.2f}, p-value: {:12g} (W: {:8g}, p: {:12g}):  {}'.format(exp_name,
          major_a, minor_a, major_b, minor_b, w_value, p_value, W, p_norm, stars))
