import gc
import sys
import pickle
import traceback
from pathlib import Path

import numpy as np

from ... import Run, evo, utils, paths
from ..  import loadlib
from .   import utils_post


def compute_angle_seed(cfg, seed, window):
    """
    :param all: if True, compute all the angles of the last generation
    """
    try:
        if not loadlib.is_done_cfg(cfg, seed):
            print('error: {} not done; skipping.'.format(filepath_done) , file=sys.stderr)
            return

        run = loadlib.load_results_from_cfg(cfg, seed, verbose=False)
        window_start, window_end = window
        if window_start is not None and window_start < 0:
            window_start = max(0, len(run.population.generations) + window_start)
        if window_end is not None and window_end < 0:
            window_end = max(0, len(run.population.generations) + window_end)
        
        best = evo.analysis.select_best(run.population, n=1, window=[window_start, window_end])[0]
        best.params['sensors'] = {'touch': None, 'angle': None,
                                  'angular_velocity': False, 'center_angle': True}
        run.population.evaluate_member(best)
        return seed, [best.sensor_values[0]]

    except Exception as exc:
        traceback.print_exc()
        utils.print_error('error encountered when computing angle file for v{}.{}, seed {}'.format(
                          cfg['meta']['major'], cfg['meta']['minor'], seed))
        return None, None

def compute_angle(params, major, minor, pool=None):
    cfg = loadlib.load_cfg(params['exp_name'], major, minor)
    window = params.get('window', (-50, None))

    filepath = loadlib.angle_filepath_from_cfg(cfg, window=window)
    filepath.parent.mkdir(parents=True, exist_ok=True)
    seeds = loadlib.find_done_cfg(cfg)

    if not utils_post.uptodate(cfg, seeds, filepath):
        args_list = []
        for seed in seeds:
            args_list.append((cfg, seed, window))

        if pool is not None:
            results = pool.starmap(compute_angle_seed, args_list)
        else:
            results = []
            for args in args_list:
                results.append(compute_angle_seed(*args))

        data = {seed: angle for seed, angle in results if seed is not None}
        utils.save_data(data, filepath)
        utils_post.set_mtime(cfg, seeds, filepath)
        # utils.print_info('saved  {}'.format(filepath))


def compute_angle_member(run, uid):
    member = run.population.repository.member(uid)
    member.params['sensors'] = {'touch': None, 'angle': None,
                                'angular_velocity': False, 'center_angle': True}
    run.population.evaluate_member(member)
    return member.sensor_values[0]

def compute_angle_all(params, major, minor, pool=None):
    cfg = loadlib.load_cfg(params['exp_name'], major, minor)

    filepath = loadlib.angle_filepath_from_cfg(cfg, all=True)
    filepath.parent.mkdir(parents=True, exist_ok=True)
    seeds = loadlib.find_done_cfg(cfg)

    if not utils_post.uptodate(cfg, seeds, filepath):
        args_list = []
        results = {}
        for seed in seeds:
            cfg = loadlib.load_cfg(params['exp_name'], major, minor)
            if not loadlib.is_done_cfg(cfg, seed):
                print('error: {} not done; skipping.'.format(filepath_done) , file=sys.stderr)
                return
            run = loadlib.load_results_from_cfg(cfg, seed, verbose=False)
            uids = run.population.generations[-1]

            args_list = [(run, uid) for uid in uids]
            if pool is not None:
                results[seed] = pool.starmap(compute_angle_member, args_list)
            else:
                results[seed] = []
                for args in args_list:
                    results[seed].append(compute_angle_member(*args))

        utils.save_data(results, filepath)
        utils_post.set_mtime(cfg, seeds, filepath)
        # utils.print_info('saved  {}'.format(filepath))
