import os
import sys
import copy
import math
import lzma
import random
import pickle
import platform
import importlib
import multiprocessing as mp
from pathlib import Path

import yaml
import numpy as np
import resource
import msgpack
import psutil
import colr



# Config files

def n_available_cores():
    """Return the number of available cores.

    Autodectecs SLURM environment.
    """
    if 'SLURM_CPUS_ON_NODE' in os.environ:
        return int(os.environ['SLURM_CPUS_ON_NODE'])
    else:
        return mp.cpu_count()

def import_str(s):
    """Import an object from a string"""
    if isinstance(s, str):
        mod_str, cls_str = s.rsplit('.', 1)
        mod = importlib.import_module(mod_str)
        return getattr(mod, cls_str)
    else:
        return s

def here(path, thisfile=__file__):
    """Return the path relative to the current file"""
    return os.path.join(os.path.dirname(thisfile), path)

def load_yaml(filepath):
    """Load a yaml file"""
    with open(filepath, 'rb') as fd:
        return yaml.safe_load(fd)

def load_config(filepath):
    """Load configuration file

    Configuration files can also contain specific changes relative to another
    configuration file. Therefore, many configuration files might end up need
    to be parsed.
    """
    with open(filepath, 'rb') as fd:
        params = yaml.safe_load(fd)

    if params.get('base', None) is not None:
        base_filepath = params['base']
        if not os.path.isabs(base_filepath):
            base_filepath = os.path.join(os.path.dirname(filepath),
                                         base_filepath)
        base_params = load_config(base_filepath)
        # merging params
        return merge_params(base_params, params)
    return params


def merge_params(base_params, change_params):
    """Merge two set of parameters.

    We merge the parameters of the two nested dictionaries. This means that if a configuration
    value is a dictionary, its key will be updated independently, which may be undesirable. This is
    unavoidable (the other solutions would add much unecessary complexity); the workaround is to
    wrap dictionary values into one-length lists. Note that empty dictionary are always considered
    as values.

    :param base_params:    the base parameters
    :param change_params:  parameter that override the values of the base parameters.
    """
    merged_params = copy.deepcopy(base_params)
    change_params.pop('base', None)

    return _merge_subparams(merged_params, change_params)


def _merge_subparams(base_params, change_params):
    """
    """
    for key, value in change_params.items():
        if isinstance(value, dict) and len(value) > 0:
            base_params.setdefault(key, {})
            _merge_subparams(base_params[key], value)
        else:
            base_params[key] = value
    return base_params


def even_divide_list(values, k):
    """Divide a list of values into `k` sections, as evenly as possible

    > _even_divide([1, 2, 3, 4, 5], 3) == [[1, 2], [3, 4], [5]]
    """
    n = len(values)
    base = (n // k)
    rest = n - base * k

    results, offset = [], 0
    for _ in range(rest):
        results.append(values[offset:offset+base+1])
        offset += base + 1
    if base > 0:
        for _ in range(k - rest):
            results.append(values[offset:offset+base])
            offset += base
    return results

def even_divide(n, k):
    """Divide 'n' into `k` integer chunks, as evenly as possible.

    > even_divide(5, 3) == [2, 2, 1]
    """
    base, rest = (n // k), n % k
    return [base + 1 if i < rest else base for i in range(k)]

def flatten_once(a):
    """Remove the first level of a nested list

    `[[0, 3], [2], [1, 5, 4]]` becomes `[0, 3, 2, 1, 5, 4]`.
    """
    return [a_ij for a_i in a for a_ij in a_i]

def memory_used():
    """return the memory usage in MB"""
    process = psutil.Process(os.getpid())
    mem = process.memory_info()[0] / float(2 ** 20)
    for child in process.children(recursive=True):
        mem += child.memory_info()[0] / float(2 ** 20)
    return mem

# def total_memory():
#     current_process = psutil.Process(os.getpid())
#     mem = current_process.memory_percent()
#     for child in current_process.children(recursive=True):
#         mem += child.memory_percent()
#     return mem

def max_memory():
    """Return max memory in MB"""
    max_rss  = resource.getrusage(resource.RUSAGE_SELF    ).ru_maxrss / 1024
    max_rss += resource.getrusage(resource.RUSAGE_CHILDREN).ru_maxrss / 1024
    if platform.system() == 'Darwin':  # macos: catches rsync uploads
        return max_rss / 1024
    return max_rss

def export_yaml(params, path=None):
    """"""
    with open(path, 'w') as f:
        yaml.safe_dump(params, f, indent=2, allow_unicode=True)

def log_transform(value, v_min, v_max):
    """Given a value in [0, 1], compute its logscale value along the [v_min, v_max] interval."""
    assert v_max > 0 and v_min > 0
    log_min, log_max = math.log(v_min), math.log(v_max)
    return math.exp(log_min + (log_max - log_min) * value)

def save_msgpack_xz(data, filename):
    if not filename.endswith('.msgpack.xz'):
        filename += '.msgpack.xz'
    with lzma.open(filename, mode='wb') as fd:
        def default(o):
            """This keeps the distinction between tuples and lists."""
            if isinstance(o, tuple):
                return {
                    '__type__': 'tuple',
                    'value': list(o),
                    }
            raise TypeError(repr(o))

        msgpack_data = msgpack.packb(data, use_bin_type=True, strict_types=True, default=default)
        fd.write(msgpack_data)
    return filename

def load_msgpack_xz(filename):
#    print(filename)
    if not os.path.isfile(filename):
        # trying with appending the extention
        filename += '.msgpack.xz'

    with lzma.open(filename, mode='rb') as fd:
        msgpack_data = fd.read()

    # import hashlib
    # print(hashlib.md5(msgpack_data).digest())
    def convert(o):
        """Restore tuples (instead of lists)"""
        if o.get('__type__') == 'tuple':
            return tuple(o['value'])
        return o

    return msgpack.unpackb(msgpack_data, raw=False, strict_map_key=False, object_hook=convert)

def load_data(filepath):
    suffix = Path(filepath).suffix
    if suffix == '.pickle':
        with open(filepath, 'rb') as fd:
            return pickle.load(fd)
    elif suffix == '.npy':
        return np.load(filepath)
    else:
        raise ValueError('unknow extension for {}'.format(filepath))

def save_data(data, filepath, verbose=True):
    suffix = Path(filepath).suffix
    if suffix == '.pickle':
        with open(filepath, 'wb') as fd:
            pickle.dump(data, fd)
    elif suffix == '.npy':
        np.save(filepath, data)
    else:
        raise ValueError('unknow extension for {}'.format(filepath))
    if verbose:
        print_info('saved {}'.format(filepath))


# UI

def red(message):
    return colr.color(message, fore='red')

def error(message):
    return '{}: {}'.format(colr.Colr().red('error'), message)

def print_error(message):
    print(error(message), file=sys.stderr)

def yellow(message):
    return colr.color(message, fore='yellow')

def warning(message):
    return '{}: {}'.format(colr.Colr().yellow('warning'), message)

def print_warning(message):
    print(warning(message), file=sys.stderr)

def green(message):
    return colr.color(message, fore='green')

def info(message):
    return '{}: {}'.format(colr.Colr().green('info'), message)

def blue(message):
    return colr.color(message, fore='blue')

def print_info(message):
    print(info(message))

def bold(message):
    return colr.color(message, style='bold')
