import os
import re
import sys
import argparse
from collections import defaultdict

import evodevo
from evodevo import postlib
from evodevo import utils



def check_job_status(exp_name, major, minor, cfg_filename, with_results=False, slurm_state=None):
    version = postlib.version_str(major, minor)
    cfg = evodevo.load_config(cfg_filename)
    assert cfg['meta']['major'] == major
    assert cfg['meta']['minor'] == minor

    planned_seeds = set(postlib.str_to_indexes(cfg['meta']['slurm_array']))
    done_seeds = postlib.find_done_cfg(cfg, with_results=with_results)

    running_seeds, pending_seeds = postlib.slurmlib.scheduled_seeds(slurm_state, exp_name, major, minor)
    unstarted_seeds = planned_seeds.difference(set(done_seeds).union(running_seeds, pending_seeds))

    return sorted(done_seeds), sorted(running_seeds), sorted(pending_seeds), sorted(unstarted_seeds)


def check_filename_collisions(found_cfgs):
    exps, collision = defaultdict(list), False
    for major, minor, cfg_filename in found_cfgs:
        exps[(major, minor)].append(cfg_filename)
    for (major, minor), filenames in exps.items():
        if len(filenames) >= 2:
            collision = True
            print('error: v{}.{} collision, found in {}'.format(major, minor, ' and '.join(filenames)))
    return not collision


def check_numbering_inconsistencies(found_cfgs):
    inconsistent = False
    for major, minor, cfg_filename in found_cfgs:
        cfg = evodevo.load_config(cfg_filename)
        if cfg['meta']['major'] != major or cfg['meta']['minor'] != minor:
            print('error: {} contains inconsistent numbering: v{}.{}'.format(cfg_filename,
                  cfg['meta']['major'], cfg['meta']['minor']))
            inconsistent = True
    return not inconsistent


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('majors', help="restrict the search to those majors", nargs='?', default=None)
    parser.add_argument('--exp_name', help="name of the experiment to process", default=evodevo.CURRENT_EXP_NAME)
    parser.add_argument('--finished', '-f', help="display finished seeds", default=False, action='store_true')
    parser.add_argument('--results', '-r', help="check that the results are presents", default=False, action='store_true')
    parser.add_argument('--nosqueue', help="don't detect active jobs", default=False, action='store_true')
    args = parser.parse_args()
    args.finished = args.finished or args.results

    directories = evodevo.CURRENT_EXP_NAMES
    if args.exp_name is not None:
        directories = [args.exp_name]

    if args.majors is not None:
        args.majors = postlib.str_to_indexes(args.majors)

    for directory in directories:
        print('finding configs in directory \'{}\'...'.format(directory))

        found_cfgs = postlib.find_cfgs(directory)

        # filename collisions
        no_collision = check_numbering_inconsistencies(found_cfgs)
        if not no_collision:
            sys.exit(1)

        # in-config numbering collisions
        consistent = check_filename_collisions(found_cfgs)
        if not consistent:
            sys.exit(1)

        if args.finished:

            slurm_state = None
            if not args.nosqueue:
                slurm_status = postlib.slurmlib.job_status()
                if slurm_status is not None:
                    n_jobs, slurm_state = slurm_status
                    print('found {} jobs scheduled or running in slurm'.format(n_jobs))

            finished_count, unfinished_count = 0, 0
            major_groups = {}
            for major, minor, cfg_filename in found_cfgs:
                if args.majors is None or major in args.majors:
                    major_groups.setdefault(major, [])
                    major_groups[major].append((minor, cfg_filename))

            printed_majors = set()
            for major, minors in sorted(major_groups.items()):
                #print('v{:02d}'.format(major))
                progress = []
                for minor, cfg_filename in minors:
                    cfg = evodevo.load_config(cfg_filename)
                    if cfg['meta'].get('disabled', False):
                        major_str = 'v{:03d}'.format(major) if major not in printed_majors else '    '
                        print('{}.{}: {}'.format(major_str, minor, utils.yellow('disabled')))
                        printed_majors.add(major)
                    else:
                        finished, running, pending, unstarted = check_job_status(directory, major, minor, cfg_filename,
                                                                                 with_results=args.results,
                                                                                 slurm_state=slurm_state)
                        progress.append((minor, finished, running, pending, unstarted))
                        finished_count += len(finished)
                        unfinished_count += len(running) + len(pending) + len(unstarted)

                # print fully finished exp first
                finished_map = {}
                for minor, finished, running, pending, unstarted in progress:
                    if len(running) == len(pending) == len(unstarted) == 0:
                        s = utils.green(postlib.indexes_to_str(finished))
                        finished_map.setdefault(s, [])
                        finished_map[s].append(minor)
                if len(finished_map) > 0:
                    for s, minors in sorted(finished_map.items()):
                        major_str = 'v{:03d}'.format(major) if major not in printed_majors else '    '
                        print('{}.{}: {} {}'.format(major_str, postlib.indexes_to_str(minors), s,
                                                    utils.green(len(minors)*'+')))
                        printed_majors.add(major)

                # print mixed status in the middle
                inprogress_str = []
                for minor, finished, running, pending, unstarted in progress:
                    if ((not (len(finished) == len(running) == len(pending) == 0)) and
                        (not (len(running) == len(pending) == len(unstarted) == 0))):
                        major_str = 'v{:03d}'.format(major) if major not in printed_majors else '    '
                        seeds_str = []
                        if len(finished)  > 0:
                            seeds_str.append(utils.green(postlib.indexes_to_str(finished)))
                        if len(running)   > 0:
                            seeds_str.append(utils.blue(postlib.indexes_to_str(running)))
                        if len(pending)   > 0:
                            seeds_str.append(utils.yellow(postlib.indexes_to_str(pending)))
                        if len(unstarted) > 0:
                            seeds_str.append(utils.red(postlib.indexes_to_str(unstarted)))

                        s = '{}.{}: {}'.format(major_str, minor, '/'.join(seeds_str))
                        printed_majors.add(major)
                        inprogress_str.append(s)
                if len(inprogress_str) > 0:
                    print('\n'.join(inprogress_str))

                # print nothing started last
                notstarted_map = {}
                for minor, finished, running, pending, unstarted in progress:
                    if len(finished) == len(running) == len(pending) == 0:
                        s = utils.red(postlib.indexes_to_str(unstarted))
                        notstarted_map.setdefault(s, [])
                        notstarted_map[s].append(minor)
                if len(notstarted_map) > 0:
                    for s, minors in sorted(notstarted_map.items()):
                        major_str = 'v{:03d}'.format(major) if major not in printed_majors else '    '
                        print('{}.{}: {} {}'.format(major_str, postlib.indexes_to_str(minors), s,
                                                    utils.red(len(minors)*'+')))
                        printed_majors.add(major)

            print('{} exps, {} simulations finished, {} remaining'.format(len(found_cfgs),
                                                                          finished_count, unfinished_count))

        else:
            print('{} exps found'.format(len(found_cfgs)))
