#!/usr/bin/env python

""" Developmental Learning Experiment

A population of controller starts will a small body that grows with each passing generation.
"""
import os
import sys
import time
import shutil
import socket
import argparse
import subprocess
from pathlib import Path

import evodevo
from evodevo import utils, paths, postlib
from evodevo import paths


def resleeve(params, run):
    """Reexecute the genotype of a previous run with a different config."""
    # loading, preparing and saving the config
    original_exp_name, original_major, original_minor = params['meta']['resleeve']
    assert postlib.is_done(original_exp_name, original_major, original_minor, seed)

    # loading and creating runs
    original_run = postlib.load_results(original_exp_name, original_major, original_minor, seed)
    original_params = original_run.params

    original_pop, sleeve_pop = original_run.population, run.population

    start_gen = params['population'].get('start_generation', 0)
    end_gen   = start_gen + params['population']['n_generation']
    if end_gen == 0 and start_gen < 0:
        end_gen = original_run.population.generation_count

    print('resleeving from {}:v{}.{} seed {}, generation {}-{}'.format(original_exp_name, original_major,
                                                                       original_minor, params['meta']['seed'],
                                                                       start_gen, end_gen))
    generations = original_pop.generations[start_gen + sleeve_pop.generation_count:end_gen]
    generations = tqdm.tqdm(generations) if args.progress else generations

    for generation in generations:
        sleeve_pop.manual_generation()
        sleeve_members = []
        for uid in generation:
            original_member = original_pop.repository.member(uid)
            m = sleeve_pop.manual_member(original_member.genotype, puid=original_member.puid)
            sleeve_members.append(m)
        sleeve_pop.post_gen()
        sleeve_pop.evaluate_generation(sleeve_members)
        run.generation_contexts.append(len(run.contexts) - 1)
        run.results.maybe_save()


if __name__ == '__main__':
    start = time.time()
    # parsing external arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('config', help="configuration filepath")
    parser.add_argument('--seed', help="random seed", type=int)
    parser.add_argument('--restart', help="restart from scratch the computation",
                                     action='store_true', default=False)
    parser.add_argument('--nosave', help="do not save the results", action='store_true', default=False)
    parser.add_argument('--progress', help="show progress bar", action='store_true', default=False)
    args = parser.parse_args()

    # loading, preparing and saving the config
    params = evodevo.load_config(args.config)

    # getting the seed
    seed = args.seed
    if seed is None:  # if not provided, we extract the seed from the slurm array id
        try:
            seed = int(os.environ['SLURM_ARRAY_TASK_ID'])
        except KeyError:
            seed = params['meta']['seed']
    assert seed is not None
    params['meta']['seed'] = seed

    # trying to copy existing results to the work directory
    if not args.restart:
        if postlib.is_done_cfg(params, seed):
            utils.print_warning('job seems already done in {}'.format(
                                postlib.done_path(params, seed)))
        else:
            postlib.move_results_to_work(params, seed, is_done_ok=False)

    # preparing save path
    if args.nosave:
        params['meta']['save_period'] = None
    else:
        work_filepath = Path(paths.RESULTS_WORK_DIR, postlib.prefix_path_from_cfg(params, seed))
        os.makedirs(work_filepath.parent, exist_ok=True)
        params['meta']['save_filepath'] = str(work_filepath)

    # running the evolution
    run = evodevo.Run(params, resume=not args.restart)
    if params['meta'].get('resleeve', None) is None:
        run.run_generations(progress_bar=args.progress)
    else:
        resleeve(params, run)

    # saving and locking the run
    if not args.nosave:
        run.save(verbose=True)
        run.lock()

    duration = int(time.time() - start)  # don't care about subsecond
    hours, remainder = divmod(duration, 3600)
    minutes, seconds = divmod(remainder, 60)
    print('duration: {}h {}m {}s'.format(hours, minutes, seconds))

    # event report
    report = sorted(run.population.events.items())
    if len(report) > 0:
        for uid, events in report:
            print('uid:{:d}'.format(uid))
            for name, event in sorted(events.items()):
                print('  {}: {}'.format(name, event))

        utils.export_yaml(report, Path(work_filepath, 'stability.yaml'))

    # flush output (useful for slurm logfiles)
    print('', file=sys.stdout, end='', flush=True)
    print('', file=sys.stderr, end='', flush=True)
    time.sleep(0.5)

    if not args.nosave:
        # on the cluster, copying results back to the main results directory
        if socket.gethostname().startswith('deigo'):
            main_filepath = Path(paths.RESULTS_MAIN_DIR, postlib.prefix_path_from_cfg(params, seed))

            subprocess.run(['ssh', 'deigo', 'mkdir -p {}'.format(main_filepath)], check=True)
            subprocess.run(['scp', '-rp', str(work_filepath), 'deigo:{}'.format(main_filepath.parent)], check=True)
            # note that the two following prints won't appear in the final results...
            utils.print_info('copied results to main directory {}'.format(main_filepath))
            # shutil.rmtree(work_filepath)
            # utils.print_info('deleted results from work directory {}'.format(work_filepath))
