import sys
import random
import argparse
import subprocess
from pathlib import Path

import numpy as np

import evodevo
from evodevo import postlib, utils, paths



MESSAGE = "starting mass 0.5, non-rolling behavior"

def add_angle_sensor(member):
    member.params['sensors'] = {'touch': None, 'angle': None,
                                'angular_velocity': False, 'center_angle': True}

def set_display(member, paused=False, headless=False, zoom=1.0, speed=4.0, size_factor=1.0,
                        message=None):
    member.params['environment']['headless']     = headless
    member.params['environment']['display']      = 1
    member.params['environment']['window_size']  = [3000, 320]
    member.params['environment']['zoom']         = zoom
    member.params['environment']['spf']          = speed / 60.0
    member.params['environment']['start_skip']   = member.params['motor_delay']
    member.params['environment']['start_paused'] = paused
    member.params['environment']['size_factor']  = size_factor
    member.params['environment']['message']      = message


class Selector:

    @classmethod
    def parser(cls, subparsers, name, desc):
        parser = subparsers.add_parser(name, help=desc)
        parser.add_argument('filepath', help="filepath to the config or result package")
        parser.add_argument('--seed', type=int, help="seed to load if a config file is provided")
        parser.add_argument('--paused', default=False, action='store_true',
                            help="start the simulation paused")
        return parser

    @classmethod
    def load_run(cls, filepath, seed=None):
        try:
            cfg = evodevo.load_config(args.filepath)
            if seed is None:
                seed = args.seed
            if seed is None:
                seeds = postlib.find_done_cfg(cfg)
                if len(seeds) == 0:
                    print('error: found no seeds with finished results')
                    sys.exit(1)
                elif len(seeds) == 1:
                    seed = seeds[0]
                else:
                    seed = input('choose a seed (among {}): '.format(sorted(seeds)))
            return postlib.load_results_from_cfg(cfg, seed)
        except (KeyError, IsADirectoryError):
            return evodevo.Run.from_results(args.filepath, frozen=True)


class ElitesSelector(Selector):

    @classmethod
    def parser(cls, subparsers):
        return super().parser(subparsers, 'elites',
                              'show the best creature at regular intervals of the evolution.')

    @classmethod
    def select(cls, args):
        run = cls.load_run(args.filepath)
        members = reversed(evodevo.evo.analysis.select_elite_history(run.population, period=500,
                                                            milestones=(0, 100, 250, 500, 1000, 2000, 3999)))
        return len(members) * [run], members


class BestSelector(Selector):

    @classmethod
    def parser(cls, subparsers):
        return super().parser(subparsers, 'best',
                              'show the best creatures of the result population.')

    @classmethod
    def select(cls, args):
        run = cls.load_run(args.filepath, seed=args.seed)
        best = evodevo.evo.analysis.select_best(run.population, n=1, window=[2000, None])
        return len(best) * [run], best


class VeryBestSelector(Selector):
    """show the best out of all seeds"""

    @classmethod
    def parser(cls, subparsers):
        parser = super().parser(subparsers, 'verybest',
                                'show the best creatures of the result population of all seeds.')
        parser.add_argument('--n', type=int, default=1, help='how many to show.')
        return parser

    @classmethod
    def select(cls, args):
        cfg = utils.load_config(args.filepath)
        seeds = postlib.find_done_cfg(cfg)

        perfseeds = [(np.max(postlib.load_perffile(cfg, seed)), seed) for seed in seeds]
        perfseeds = sorted(perfseeds, reverse=True)

        runs, bests = [], []
        for _, seed in perfseeds[:args.n]:
            print('seed {}'.format(seed))
            run = cls.load_run(args.filepath, seed=seed)
            best = evodevo.evo.analysis.select_best(run.population, n=1)
            runs.append(run)
            bests.append(best[0])
        return runs, bests


class AllSelector(Selector):

    @classmethod
    def parser(cls, subparsers):
        return super().parser(subparsers, 'all',
                              'show all the creatures of the result population.')

    @classmethod
    def select(cls, args):
        run = cls.load_run(args.filepath, seed=args.seed)
        members = []
        for generation in run.population.generations:
            for uid in generation:
                members.append(run.population.repository.member(uid))
        return len(members) * [run], members


class AnySelector(Selector):

    @classmethod
    def parser(cls, subparsers):
        return super().parser(subparsers, 'any',
                              'show random creatures from the result population.')

    @classmethod
    def select(cls, args):
        run = cls.load_run(args.filepath, seed=args.seed)
        def selection():
            while True:
                uid = random.randint(0, run.population.repository.member_count)
                yield run, run.population.repository.member(uid)
        return selection()


class GenerationSelector(Selector):

    @classmethod
    def parser(cls, subparsers):
        parser = super().parser(subparsers, 'generation',
                                'show the creatures from a specific generation.')
        parser.add_argument('gen', type=int, default=0, help="the generation to select.")
        parser.add_argument('--n', type=int, default=None, help="how many to show")
        return parser

    @classmethod
    def select(cls, args):
        run = cls.load_run(args.filepath, seed=args.seed)
        gen = run.population.generations[args.gen]
        members = evodevo.evo.analysis.sorted_creatures(run.population, gen)
        if args.n is not None:
            members = members[:args.n]
        return len(members) * [run], members


class UidSelector(Selector):

    @classmethod
    def parser(cls, subparsers):
        parser = super().parser(subparsers, 'uid',
                                'show the creature with the uid specified.')
        parser.add_argument('uid', type=int, help="the uid of the creature to select.")
        return parser

    @classmethod
    def select(cls, args):
        run = cls.load_run(args.filepath)
        return [run], [run.population.repository.member(args.uid)]


class RandomSelector(Selector):

    @classmethod
    def parser(cls, subparsers):
        parser = super().parser(subparsers, 'random',
                                'generate random genotype and run them — does not need saved results.')
        parser.add_argument('--gen', type=int, default=0, help='the generation number to draw random from')
        return parser

    @classmethod
    def select(cls, args):
        cfg = evodevo.load_config(args.filepath)
        if args.seed is not None:
            cfg['meta']['seed'] = args.seed
        else:
            cfg['meta']['seed'] = random.randint(0, 10e10)
        print('seed: {}'.format(cfg['meta']['seed']))
        cfg['meta']['save_period'] = None

        run = evodevo.Run(cfg)
        pop = run.population
        assert args.gen >= 0
        pop.generations = args.gen * [None]
        pop.generation_count = args.gen
        pop._first_generation()
        members = [pop.repository.member(uid) for uid in pop.generations[-1]]
        return len(members) * [run], members


class GenotypeSelector(Selector):

    @classmethod
    def parser(cls, subparsers):
        parser =  super().parser(subparsers, 'genotype',
                                 'show the creature with the genotype specified.')
        parser.add_argument('genotype', type=str, help="the uid of the creature to select.")
        return parser

    @classmethod
    def select(cls, args):
        assert args.genotype is not None
        genotype = eval(args.genotype)

        cfg = evodevo.load_config(args.filepath)
        if args.seed is not None:
            cfg['meta']['seed'] = args.seed
        else:
            cfg['meta']['seed'] = random.randint(0, 10e10)
        print('seed: {}'.format(cfg['meta']['seed']))
        cfg['meta']['save_period'] = None

        run = evodevo.Run(cfg)
        pop = run.population
        pop.manual_generation()
        member = pop.manual_member(genotype=genotype)

        return [run], [member]


def measure_instability(run, member, noise=None):
    fitness = member.fitness
    fitnesses = evodevo.evo.analysis.noise_fitness(run, member, 100, noise=noise)
    percents = np.array(fitnesses) / np.mean(fitnesses)
    noise = member.params['motor_noise'] if noise is None else noise
    print('fitness unstability: {:.2f}% (noise = {})'.format(100 * np.std(percents), noise))
    return np.std(percents)


def extract_seed(run):
    return int(Path(run.params['meta']['save_filepath']).name[4:])

def make_movie(member, folder):
    width, height = member.params['environment']['window_size']
    framerate = int(1 / member.params['environment']['spf'])
    cmd = ['ffmpeg', '-y', '-r', str(framerate), '-f', 'image2', '-s', '{}x{}'.format(width, height), '-r', '60', '-i',
           '{}/shot%05d.png'.format(folder), '-vcodec', 'libx264', '-crf', '22', '-pix_fmt', 'yuv420p',
           '-r', '25', '{}.mp4'.format(folder)]
    print(' '.join(cmd))
    subprocess.run(cmd)
    utils.print_info('saved {}.mp4'.format(folder))

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('--headless',  help="headless execution", default=False, action='store_true')
    parser.add_argument('--movie',  help="record a movie", default=False, action='store_true')
    parser.add_argument('--zoom',  help="zoom level", default=None, type=float)
    parser.add_argument('--speed',  help="speed, compared to real time", default=2.0, type=float)
    parser.add_argument('--message',  help="message to display", default=None, type=str)
    subparsers = parser.add_subparsers(title="commands", dest="command",
                                       help="what to show among `any`, `best`, `elites`, `random`, `uid`, `genotype`")


    for selector_cls in [AllSelector, RandomSelector, AnySelector, BestSelector, VeryBestSelector,
                         GenotypeSelector, GenerationSelector, UidSelector, ElitesSelector]:
        selector_parser = selector_cls.parser(subparsers)
        selector_parser.set_defaults(func=selector_cls.select)

    args = parser.parse_args()
    runs, members = args.func(args)

    MESSAGE = args.message

    print('  gen        uid   fitness       puid')

    screenshot_folder = None
    for run, member in zip(runs, members):
        size_factor = member.params['environment'].get('size_factor', 1.0)
        zoom        = member.params['environment']['zoom']
        if args.zoom is not None:
            zoom = args.zoom
        print(zoom)
        set_display(member, args.paused, headless=args.headless, zoom=zoom, speed=args.speed,
                    size_factor=size_factor, message=MESSAGE)
        add_angle_sensor(member)
        fitness = member.fitness
        first_eval = fitness is None
        if not first_eval:
            puid_str = '      none' if member.puid is None else '{: 10d}'.format(member.puid)
            print('{: 5d} {: 10d} {: 9.2f} {}'.format(member.generation, member.uid, fitness, puid_str))

        if args.movie:
            member.params['environment']['pause_at_end'] = False
            version = 'v{}.{}'.format(run.params['meta']['major'], run.params['meta']['minor'])
            if args.command == 'random':
                title = '{}_random_seed{}'.format(version, run.params['meta']['seed'])
            else:
                title = '{}_seed{}_g{}_uid{}'.format(version, extract_seed(run), member.generation, member.uid)

            screenshot_folder = Path(paths.MOVIE_DIR, run.params['meta']['exp_name'], title)
            screenshot_folder.mkdir(parents=True, exist_ok=True)
            screenshot_folder = str(screenshot_folder)

        run.population.evaluate_member(member, screenshot_folder=screenshot_folder)

        if args.movie:
            make_movie(member, str(screenshot_folder))

        if first_eval:
            puid_str = '      none' if member.puid is None else '{: 10d}'.format(member.puid)
            print('{: 5d} {: 10d} {: 9.2f} {}'.format(member.generation, member.uid, member.fitness, puid_str))
        elif abs(member.fitness - fitness) > 0.0001:
            print('warning: replay fitness ({:.1f}) diverging from original fitness by {:.5f}'.format(
                  member.fitness, abs(member.fitness - fitness)))

    # measure_instability(best)
    # for noise in [1.0, 0.1, 0.01, 0.0001]:
    #     measure_instability(best, noise)
