import os
import sys
import time
import socket
import argparse
import subprocess
from pathlib import Path

import evodevo
from evodevo import postlib, paths, utils


SCRIPT = """
#!/bin/bash
sbatch << EOT
#!/bin/bash

#SBATCH -p {queue_name}                        \t# Queue name (default short on deigo, compute on mario)
#SBATCH --job-name=v{version}_{exp_name}       \t# Job name
#SBATCH --mail-type=FAIL                       \t# Mail events (NONE, BEGIN, END, FAIL, ALL)
#SBATCH --mail-user=fabien.benureau@oist.jp
#SBATCH --ntasks=1                             \t# Run a single task
#SBATCH --cpus-per-task={n_cores}                    \t# Number of cores per tasks
#SBATCH --mem={memory}                             \t# Job Memory
#SBATCH --time={duration}                      \t# Time limit hrs:min:sec
#SBATCH --output={output_filepath_slurm} \t# Standard output log
#SBATCH --error={error_filepath_slurm}   \t# Standard error log
#SBATCH --array={array_index}                    \t# Array indexes
{extra_script}

pwd; hostname; date; echo {config}

source ~/.venvs/wobbling2022/bin/activate
time {mprof} python run_exp.py {restart}{config}

echo ""
date;

EOT
"""

# TODO: verbose about ignoring configs with minor=None and no base in the name

hostname = socket.gethostname()

parser = argparse.ArgumentParser()
parser.add_argument('configs',    help="configuration filepaths", nargs='+')
parser.add_argument('--quiet',    help="don't display slurm script", default=False, action='store_true')
parser.add_argument('--restart',  help="restart from scratch the computation", action='store_true', default=False)
parser.add_argument('--force',    help="launch even disabled configs", action='store_true', default=False)
parser.add_argument('--pretend',  help="do not actually submit", default=False, action='store_true')
parser.add_argument('--queue',    help="queue name (default short)", default=None, action='store')
parser.add_argument('--tag',      help="only run config with this tag", default=None, action='store')
parser.add_argument('--array',    help="slurm array", default=None)
parser.add_argument('--nosqueue', help="don't try to get active job status", default=False, action='store_true')
parser.add_argument('--mprof',    help="profile memory usage", default=False, action='store_true')
args = parser.parse_args()


def compute_partition(args):
    if args.queue is None:
        if hostname.startswith('deigo'):
            return 'short'
        elif hostname in ['mario64', 'alien']:
            return 'compute'
        else:
            return 'compute'
    else:
        return args.queue

def natural_n_cores(cfg):
    pp = cfg['population']
    n = pp['n_mutant'] + pp['n_random']
    if pp['reevaluate']:
        n += pp['n_elites']
    return n


def build_array_indexes(params, slurm_state):
    """Returing a string of the array of non-finished seeds (or all of them if --restart)."""
    if args.array is not None:
        slurm_array = args.array
    else:
        if hostname in ['mario64', 'alien']:
            slurm_array = params['meta']['slurm_array_fast']
        else:
            slurm_array = params['meta']['slurm_array']

    slurm_indexes = postlib.str_to_indexes(slurm_array)

    exp_name, major, minor = params['meta']['exp_name'], params['meta']['major'], params['meta']['minor']
    running_seeds, pending_seeds = postlib.slurmlib.scheduled_seeds(slurm_state, exp_name, major, minor)
    unstarted_seeds = set(slurm_indexes).difference(set(running_seeds).union(pending_seeds))

    if args.restart:
        selected = sorted(unstarted_seeds)
    else:
        selected = []
        for index in sorted(unstarted_seeds):
            if not postlib.is_done_cfg(params, index):
                selected.append(index)

    if len(selected) == 0:
        return [], None
    return selected, postlib.indexes_to_str(selected)

if __name__ == '__main__':

    slurm_state = {}

    if not args.nosqueue:
        slurm_status = postlib.slurmlib.job_status()
        if slurm_status is not None:
            n_jobs, slurm_state = slurm_status
            print('found {} jobs scheduled or running in slurm'.format(n_jobs))
        else:
            utils.print_error('could not fetch data about running jobs from squeue')
            sys.exit(1)

    for config_name in args.configs:
        if ('_base.yaml' in config_name or not config_name.endswith('.yaml') or
             (not Path(config_name).name.startswith('v'))):
            continue
        params = evodevo.load_config(config_name)

        # do not run disabled configs
        if (not args.force) and params['meta'].get('disabled', False):
            utils.print_warning('{} disabled; skipping.'.format(config_name))
            continue

        # filter by tag if necessary
        if args.tag is not None and args.tag not in params['meta'].get('tags', []):
            continue

        seeds, array_str = build_array_indexes(params, slurm_state)
        if array_str is None:
            utils.print_warning("{}: no unfinished job detected; skipping.".format(config_name))
            continue

        # queue, cores, duration, memory
        queue = compute_partition(args)
        if hostname in ['mario64', 'alien']:
            duration = '24:00:00'
            n_cores = natural_n_cores(params)
        else:
            # the short queue as a 2 hour maximum walltime.
            duration = '2:00:00' if queue == 'short' else params['meta']['slurm_time']
            n_cores = params['meta']['slurm_cores']
        memory = params['meta']['slurm_memory']

        # creating the logpaths before the job, because if it does not exist, the job instantly fails
        for seed in seeds:
            output_filepath = Path(paths.RESULTS_WORK_DIR, postlib.prefix_path_from_cfg(params, seed))
            os.makedirs(output_filepath, exist_ok=True)

        log_prefix = Path(paths.RESULTS_WORK_DIR, postlib.prefix_path_from_cfg(params, ''))
        output_filepath_slurm = str(log_prefix) + '%a/log_%A.out'
        error_filepath_slurm  = str(log_prefix) + '%a/log_%A.err'

        # extra scripts
        extra_script = ''
        # if hostname.startswith('deigo'):
        #     extra_script = '#SBATCH -C cascade\n'

        # profiling
        mprof_str = ''
        if args.mprof:
            raise NotImplementedError # FIXME
            output_filepath_mprof = os.path.join(output_filepath.parent,
                                                 "mprofile_2020032300000.dat") # FIXME
            mprof_str = 'mprof run -o ' + output_filepath_mprof

        # creating the script
        version = postlib.version_str(major=params['meta']['major'], minor=params['meta']['minor'])
        exp_name = params['meta']['exp_name']
        restart_str = '--restart ' if args.restart else ''
        script = SCRIPT.format(version=version, config=config_name, array_index=array_str,
                               memory=memory, duration=duration, n_cores=n_cores,
                               output_filepath_slurm=output_filepath_slurm,
                               error_filepath_slurm=error_filepath_slurm,
                               queue_name=queue, restart=restart_str, mprof=mprof_str,
                               extra_script=extra_script, exp_name=exp_name)

        utils.print_info('submitting {}'.format(config_name))

        if not args.quiet:
            print(script)
            print('')

        if not args.pretend:
            p = subprocess.Popen(script, shell=True)
            p.wait()
            time.sleep(1.0)
