import os
import sys
import glob
import time
import shutil
import socket
import argparse
from pathlib import Path

import evodevo
from evodevo import paths, postlib, utils


def old_enough(seed_path):
    done_path = Path(seed_path, 'done')
    # done file exist and is more than three hours old.
    if done_path.is_file() and time.time() - done_path.stat().st_mtime > 2 * 3600.0:
        return True
    return False

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('majors', help="the major to consider", nargs='?', default=None)
    parser.add_argument('--exp_name', help="name of the experiment to process", default=evodevo.CURRENT_EXP_NAME)
    parser.add_argument('--pretend', '-p', help="only print what the script would do.",
                        default=False, action='store_true')
    parser.add_argument('--force', '-f', default=False, action='store_true',
                        help="required when not on the cluster, to avoid accidental loss of data.")
    parser.add_argument('--active_ok', '-a', default=False, action='store_true',
                        help="allow to run while jobs are running or pending on the cluster")
    args = parser.parse_args()

    # if not on cluster, we need --force
    on_cluster = False
    hostname = socket.gethostname()
    if hostname.startswith('deigo') or hostname.startswith('saion'):
        on_cluster = True

    if not args.force and not on_cluster:
        utils.print_error('not on cluster; use --force to delete files.')
        sys.exit(1)

    if args.majors is not None:
        args.majors = postlib.str_to_indexes(args.majors)

    # check for active jobs
    if not args.active_ok:
        slurm_status = postlib.slurmlib.job_status()
        if slurm_status is not None:
            n_jobs, slurm_state = slurm_status
            if n_jobs > 0:
                utils.print_error('{} jobs are running or pending; use --active_ok to override; exiting.'.format(n_jobs))
                sys.exit(1)
        elif on_cluster:
            utils.print_error('could not fetch data about running jobs from squeue; use --active_ok to override; exiting.')
            sys.exit(1)

    def move_dir(work_path, main_path, reason='unknown reason'):
        if args.pretend:
            utils.print_info('would move {} to {} [{}]'.format(work_path, main_path, reason))
        else:
            main_path.mkdir(parents=True, exist_ok=True)
            shutil.rmtree(main_path.resolve())
            shutil.move(work_path.resolve(), main_path.resolve())
            utils.print_info('moved {} to {} [{}]'.format(work_path, main_path, reason))

    for major, minor, cfg_filepath in postlib.find_cfgs(args.exp_name):
        if not args.majors is None and major not in args.majors:
            continue

        print(postlib.version_str(major, minor))
        cfg = postlib.load_cfg(args.exp_name, major, minor)

        # remove all corrupted runs
        for directory in [paths.RESULTS_WORK_DIR, paths.RESULTS_MAIN_DIR]:
            uncorrupted = set(postlib.find_started_cfg(cfg, with_results=True, check_integrity=True,
                                                       directories=[directory]))
            corrupted   = set(postlib.find_started_cfg(cfg, with_results=True, check_integrity=False,
                                                       directories=[directory])) - uncorrupted
            #print(sorted(uncorrupted), sorted(corrupted))
            for seed in corrupted:
                seed_path = Path(directory, postlib.prefix_path_from_cfg(cfg, seed))
                if args.active_ok and not old_enough(seed_path):
                    continue
                if args.pretend:
                    utils.print_warning('would delete corrupted run in {}'.format(seed_path))
                else:
                    shutil.rmtree(seed_path.resolve())
                    utils.print_warning('deleted corrupted run in {}'.format(seed_path))

        # find all done uncorrupted experiments in work and main
        done_seeds_work = set(postlib.find_done_cfg(cfg, with_results=True, check_integrity=True, # defensive
                                                    directories=[paths.RESULTS_WORK_DIR]))
        done_seeds_main = set(postlib.find_done_cfg(cfg, with_results=True, check_integrity=True,
                                                    directories=[paths.RESULTS_MAIN_DIR]))

        # if a finished run is in work but not main, move it there
        for seed in done_seeds_work - done_seeds_main:
            work_path = Path(paths.RESULTS_WORK_DIR, postlib.prefix_path_from_cfg(cfg, seed))
            main_path = Path(paths.RESULTS_MAIN_DIR, postlib.prefix_path_from_cfg(cfg, seed))
            if args.active_ok and not old_enough(work_path):  # avoid messing with running jobs
                continue
            move_dir(work_path, main_path, reason='absent from main')

        # if finished run are both present in main and work, only move it from work to main
        # if the one in work is more recent
        for seed in done_seeds_work.intersection(done_seeds_main):
            work_path = Path(paths.RESULTS_WORK_DIR, postlib.prefix_path_from_cfg(cfg, seed))
            main_path = Path(paths.RESULTS_MAIN_DIR, postlib.prefix_path_from_cfg(cfg, seed))
            if args.active_ok and not old_enough(work_path):  # avoid messing with running jobs
                continue
            work_timestamp = Path(work_path, 'manifest.yaml').stat().st_mtime
            main_timestamp = Path(work_path, 'manifest.yaml').stat().st_mtime
            if work_timestamp > main_timestamp:
                move_dir(work_path, main_path, reason='more recent')

        # rsync main with archive
