import argparse
from pathlib import Path

from evodevo import postlib, paths, utils
from evodevo.postlib import loadlib
from evodevo.postlib import color
import numpy as np
np.seterr(all='raise')


EXP_NAME = 'wobbling'
CONFIDENCE = 0.99
REDRAW = False


def cfgs(version):
    _cfgs = []
    for version_list, f_perf, offset, colors in versions:
        for exp_name, major, minors in version_list:
            for minor in minors:
                cfg = loadlib.load_cfg(exp_name, major, minor)
                _cfgs.append(cfg)
    return _cfgs


def redraw(filename):
    if not REDRAW:
        filename.replace(' ', '_')
        fig_filepath = Path(paths.FIGURES_DIR, '{}/boxplot/{}.pdf'.format(EXP_NAME, filename))
        if fig_filepath.exists():
            utils.print_warning('{} exists, use --redraw to force regeneration'.format(fig_filepath))
            return False
    return True


width_0, height_0 = 700, 300
Y_MAX = 20


def fig3A_sinusoid_mass_pause():
    title = 'fig3A_sinusoid_mass_pause' + '_conf{}'.format(CONFIDENCE)
    print(title)
    if not redraw(title): return

    ff = postlib.fitness_3950_4000
    offset = 0.25
    versions = [([(EXP_NAME,   1, [1])], ff, 0, color.control)]
    for i in reversed(range(1, 10)):
        versions.extend([
            ([(EXP_NAME, 2, [i])], ff, -offset, color.development),
            ([(EXP_NAME, 3, [i])], ff,       0, color.development2),
            ([(EXP_NAME, 4, [i])], ff,  offset, color.development3),
        ])

    data = postlib.prepare_data_max_fitness(versions, confidence=CONFIDENCE)
    x_labels = ['control']
    for i in range(1, len(versions), 3):
        exp_name, major, minors = versions[i][0][0]
        cfg = loadlib.load_cfg(exp_name, major, minors[0])
        period = cfg['population']['dev_vars']['node_mass']['subgenerator']['subgenerators'][0]['subgenerators'][0]['period']
        x_labels.append(str(period))
        for offset in [1, 2]:
            exp_name, major, minors = versions[i+offset][0][0]
            cfg = loadlib.load_cfg(exp_name, major, minors[0])
            assert period == cfg['population']['dev_vars']['node_mass']['subgenerator']['subgenerators'][0]['subgenerators'][0]['period']



    x_ticker = list(range(1, len(x_labels) + 1))
    postlib.boxplot_graph(data, postlib.one_delay_triple_counter(), title=title, exp_name=EXP_NAME,
                          baseline=(EXP_NAME, 1, 1), boxwidth=0.2,
                          y_range=(0, 19.9), width=500, height=300, trendline=False,
                          color_trendlines=True,
                          x_labels=x_labels, x_ticker=x_ticker, x_range=(0.5, len(x_ticker) + 0.5),
                          redraw=REDRAW)


def fig3B_sinusoid_amplitude():
    title = f'fig3B_sinusoid_3ways_amplitude_conf{CONFIDENCE}'
    if not redraw(title): return

    ff = postlib.fitness_3950_4000
    offset = 0.25
    versions = [([(EXP_NAME,   1, [1])], ff, 0, color.control)]

    for m1, m2, m3 in zip([[9, 1], [9, 2], [2, 3], [9, 4], [9, 5], [9, 6]],
                          [[6, 1], [6, 2], [6, 3], [6, 4], [6, 5], [6, 6]],
                          [[5, 1], [5, 2], [5, 3], [5, 4], [5, 5], [5, 6]]):
        versions.append(([(EXP_NAME, m1[0], [m1[1]])], ff, -offset, color.development))
        versions.append(([(EXP_NAME, m2[0], [m2[1]])], ff,       0, ('#9437ff', '#224d15', '#008f00')))
        versions.append(([(EXP_NAME, m3[0], [m3[1]])], ff,  offset, ('#9437ff', '#941100', '#ff2600')))


    data = postlib.prepare_data_max_fitness(versions, confidence=CONFIDENCE)
    x_labels = ['control']
    for i in range(1, len(versions) - 1, 3):
        exp_name, major_, minors = versions[i][0][0]
        for minor in minors:
            cfg = loadlib.load_cfg(exp_name, major_, minor)
            # if dev_name == 'muscle_stiff':
            #     amplitude = cfg['population']['dev_vars'][dev_name]['index']['subgenerator']['subgenerators'][0]['subgenerators'][0]['amplitude']
            # else:
            amplitude = cfg['population']['dev_vars']['node_mass']['subgenerator']['subgenerators'][0]['subgenerators'][0]['amplitude']
            x_labels.append(str(amplitude))
    x_ticker = list(range(1, len(x_labels) + 1))
    postlib.boxplot_graph(data, postlib.one_delay_triple_counter(), title=title, exp_name=EXP_NAME,
                          baseline=(EXP_NAME, 1, 1), boxwidth=0.2, y_range=(0, Y_MAX),
                          x_labels=x_labels, x_ticker=x_ticker, width=500, height=height_0, redraw=REDRAW)


def fig3C_sinusoid_devlength(dev_name):
    title = f'fig3C_sinusoid_{dev_name}_devlength_conf{CONFIDENCE}'
    if not redraw(title): return

    ff = postlib.fitness_3950_4000
    versions = [([(EXP_NAME, 1, [1])], ff, 0.0, color.control),
                ([(EXP_NAME, 7, [1, 2, 3]) ], ff, 0.0, color.development),
                ([(EXP_NAME, 9, [4]) ], ff, 0.0, color.development),
                ([(EXP_NAME, 7, [4]) ], ff, 0.0, color.development),
               ]

    data = postlib.prepare_data_max_fitness(versions, confidence=CONFIDENCE)
    x_labels = ['control']
    for version in versions[1:]:
        exp_name, major_, minors = version[0][0]
        for minor in minors:
            cfg = loadlib.load_cfg(exp_name, major_, minor)
            # if dev_name == 'muscle_stiff':
            #     amplitude = cfg['population']['dev_vars'][dev_name]['index']['subgenerator']['subgenerators'][0]['subgenerators'][0]['amplitude']
            # else:
            dev_length = cfg['population']['dev_vars'][dev_name]['subgenerator']['subgenerators'][1]['milestones'][-1][0]
            x_labels.append(f'{dev_length}')
    x_ticker = list(range(1, len(x_labels) + 1))
    postlib.boxplot_graph(data, postlib.counter(), title=title, exp_name=EXP_NAME,
                          baseline=(EXP_NAME, 1, 1), boxwidth=0.4, y_range=(0, Y_MAX),
                          x_labels=x_labels, x_ticker=x_ticker, width=500, height=height_0, redraw=REDRAW)


def regenerate_all(redraw=False, optional=False):
    fig3A_sinusoid_mass_pause()
    fig3B_sinusoid_amplitude()
    fig3C_sinusoid_devlength('node_mass')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--redraw', help="regenerate existing boxplots", action='store_true', default=False)
    args = parser.parse_args()
    REDRAW = args.redraw

    regenerate_all(optional=False)
